@@ -22,9 +22,9 @@ const bench = common.createBenchmark(main, {
     'utf8',
     'UTF8',
   ],
-  n: [1e6]
+  n: [1e6],
 }, {
-  flags: ['--expose-internals']
+  flags: ['--expose-internals'],
 });
 
 function main({ encoding, n }) {
