@@ -5,19 +5,19 @@ const Readable = require('stream').Readable;
 
 const bench = common.createBenchmark(main, {
   n: [200e1],
-  type: ['string', 'buffer']
+  type: ['string', 'buffer'],
 });
 
 function main({ n, type }) {
   const s = new Readable();
-  var data = 'a'.repeat(32);
+  let data = 'a'.repeat(32);
   if (type === 'buffer')
     data = Buffer.from(data);
   s._read = function() {};
 
   bench.start();
-  for (var k = 0; k < n; ++k) {
-    for (var i = 0; i < 1e4; ++i)
+  for (let k = 0; k < n; ++k) {
+    for (let i = 0; i < 1e4; ++i)
       s.push(data);
     while (s.read(32));
   }
