@@ -7,24 +7,19 @@
 #include <atomic>
 
 #include "include/v8-platform.h"
-#include "src/ast/ast.h"
 #include "src/base/platform/mutex.h"
 #include "src/base/platform/time.h"
 #include "src/codegen/compiler.h"
 #include "src/common/globals.h"
 #include "src/execution/isolate.h"
 #include "src/flags/flags.h"
-#include "src/handles/global-handles-inl.h"
 #include "src/heap/parked-scope.h"
 #include "src/logging/counters.h"
 #include "src/logging/runtime-call-stats-scope.h"
-#include "src/numbers/hash-seed-inl.h"
 #include "src/objects/instance-type.h"
 #include "src/objects/objects-inl.h"
 #include "src/parsing/parse-info.h"
-#include "src/parsing/parser.h"
-#include "src/roots/roots.h"
-#include "src/sandbox/external-pointer.h"
+#include "src/parsing/scanner.h"
 #include "src/tasks/cancelable-task.h"
 #include "src/tasks/task-utils.h"
 #include "src/zone/zone-list-inl.h"  // crbug.com/v8/8816
@@ -48,9 +43,9 @@ class LazyCompileDispatcher::JobTask : public v8::JobTask {
   size_t GetMaxConcurrency(size_t worker_count) const final {
     size_t n = lazy_compile_dispatcher_->num_jobs_for_background_.load(
         std::memory_order_relaxed);
-    if (FLAG_lazy_compile_dispatcher_max_threads == 0) return n;
+    if (v8_flags.lazy_compile_dispatcher_max_threads == 0) return n;
     return std::min(
-        n, static_cast<size_t>(FLAG_lazy_compile_dispatcher_max_threads));
+        n, static_cast<size_t>(v8_flags.lazy_compile_dispatcher_max_threads));
   }
 
  private:
@@ -74,7 +69,7 @@ LazyCompileDispatcher::LazyCompileDispatcher(Isolate* isolate,
           reinterpret_cast<v8::Isolate*>(isolate))),
       platform_(platform),
       max_stack_size_(max_stack_size),
-      trace_compiler_dispatcher_(FLAG_trace_compiler_dispatcher),
+      trace_compiler_dispatcher_(v8_flags.trace_compiler_dispatcher),
       idle_task_manager_(new CancelableTaskManager()),
       idle_task_scheduled_(false),
       num_jobs_for_background_(0),
@@ -407,7 +402,10 @@ void LazyCompileDispatcher::DoBackgroundWork(JobDelegate* delegate) {
 
   ReusableUnoptimizedCompileState reusable_state(&isolate);
 
-  while (!delegate->ShouldYield()) {
+  while (true) {
+    // Return immediately on yield, avoiding the second loop.
+    if (delegate->ShouldYield()) return;
+
     Job* job = nullptr;
     {
       base::MutexGuard lock(&mutex_);
