@@ -6,7 +6,6 @@
 #define V8_COMPILER_COMPILATION_DEPENDENCIES_H_
 
 #include "src/compiler/js-heap-broker.h"
-#include "src/objects/objects.h"
 #include "src/zone/zone-containers.h"
 
 namespace v8 {
@@ -89,6 +88,7 @@ class V8_EXPORT_PRIVATE CompilationDependencies : public ZoneObject {
   bool DependOnPromiseHookProtector();
   bool DependOnPromiseSpeciesProtector();
   bool DependOnPromiseThenProtector();
+  bool DependOnMegaDOMProtector();
 
   // Record the assumption that {site}'s {ElementsKind} doesn't change.
   void DependOnElementsKind(const AllocationSiteRef& site);
@@ -118,8 +118,13 @@ class V8_EXPORT_PRIVATE CompilationDependencies : public ZoneObject {
   // up to (and including) the {last_prototype}.
   void DependOnStablePrototypeChains(
       ZoneVector<MapRef> const& receiver_maps, WhereToStart start,
-      base::Optional<JSObjectRef> last_prototype =
-          base::Optional<JSObjectRef>());
+      OptionalJSObjectRef last_prototype = OptionalJSObjectRef());
+
+  // For the given map, depend on the stability of (the maps of) all prototypes
+  // up to (and including) the {last_prototype}.
+  void DependOnStablePrototypeChain(
+      MapRef receiver_maps, WhereToStart start,
+      OptionalJSObjectRef last_prototype = OptionalJSObjectRef());
 
   // Like DependOnElementsKind but also applies to all nested allocation sites.
   void DependOnElementsKinds(const AllocationSiteRef& site);
