@@ -4,13 +4,20 @@
 
 #include "src/compiler/loop-analysis.h"
 
+#include "src/base/v8-fallthrough.h"
 #include "src/codegen/tick-counter.h"
+#include "src/compiler/all-nodes.h"
+#include "src/compiler/common-operator.h"
 #include "src/compiler/graph.h"
 #include "src/compiler/node-marker.h"
 #include "src/compiler/node-properties.h"
 #include "src/compiler/node.h"
 #include "src/zone/zone.h"
 
+#if V8_ENABLE_WEBASSEMBLY
+#include "src/wasm/wasm-code-manager.h"
+#endif
+
 namespace v8 {
 namespace internal {
 
@@ -25,7 +32,8 @@ namespace compiler {
 // Temporary information for each node during marking.
 struct NodeInfo {
   Node* node;
-  NodeInfo* next;       // link in chaining loop members
+  NodeInfo* next;  // link in chaining loop members
+  bool backwards_visited;
 };
 
 
@@ -38,7 +46,6 @@ struct TempLoopInfo {
   LoopTree::Loop* loop;
 };
 
-
 // Encapsulation of the loop finding algorithm.
 // -----------------------------------------------------------------------------
 // Conceptually, the contents of a loop are those nodes that are "between" the
@@ -53,6 +60,8 @@ struct TempLoopInfo {
 // 1 bit per loop per node per direction are required during the marking phase.
 // To handle nested loops correctly, the algorithm must filter some reachability
 // marks on edges into/out-of the loop header nodes.
+// Note: this algorithm assumes there are no unreachable loop header nodes
+// (including loop phis).
 class LoopFinderImpl {
  public:
   LoopFinderImpl(Graph* graph, LoopTree* loop_tree, TickCounter* tick_counter,
@@ -61,7 +70,7 @@ class LoopFinderImpl {
         end_(graph->end()),
         queue_(zone),
         queued_(graph, 2),
-        info_(graph->NodeCount(), {nullptr, nullptr}, zone),
+        info_(graph->NodeCount(), {nullptr, nullptr, false}, zone),
         loops_(zone),
         loop_num_(graph->NodeCount(), -1, zone),
         loop_tree_(loop_tree),
@@ -190,9 +199,9 @@ class LoopFinderImpl {
     Queue(end_);
 
     while (!queue_.empty()) {
-      tick_counter_->DoTick();
+      tick_counter_->TickAndMaybeEnterSafepoint();
       Node* node = queue_.front();
-      info(node);
+      info(node).backwards_visited = true;
       queue_.pop_front();
       queued_.Set(node, false);
 
@@ -224,10 +233,18 @@ class LoopFinderImpl {
         Node* input = node->InputAt(i);
         if (IsBackedge(node, i)) {
           // Only propagate the loop mark on backedges.
-          if (SetBackwardMark(input, loop_num)) Queue(input);
+          if (SetBackwardMark(input, loop_num) ||
+              !info(input).backwards_visited) {
+            Queue(input);
+          }
         } else {
           // Entry or normal edge. Propagate all marks except loop_num.
-          if (PropagateBackwardMarks(node, input, loop_num)) Queue(input);
+          // TODO(manoskouk): Add test that needs backwards_visited to function
+          // correctly, probably using wasm loop unrolling when it is available.
+          if (PropagateBackwardMarks(node, input, loop_num) ||
+              !info(input).backwards_visited) {
+            Queue(input);
+          }
         }
       }
     }
@@ -309,7 +326,7 @@ class LoopFinderImpl {
     }
     // Propagate forward on paths that were backward reachable from backedges.
     while (!queue_.empty()) {
-      tick_counter_->DoTick();
+      tick_counter_->TickAndMaybeEnterSafepoint();
       Node* node = queue_.front();
       queue_.pop_front();
       queued_.Set(node, false);
@@ -524,16 +541,207 @@ class LoopFinderImpl {
 LoopTree* LoopFinder::BuildLoopTree(Graph* graph, TickCounter* tick_counter,
                                     Zone* zone) {
   LoopTree* loop_tree =
-      new (graph->zone()) LoopTree(graph->NodeCount(), graph->zone());
+      graph->zone()->New<LoopTree>(graph->NodeCount(), graph->zone());
   LoopFinderImpl finder(graph, loop_tree, tick_counter, zone);
   finder.Run();
-  if (FLAG_trace_turbo_loop) {
+  if (v8_flags.trace_turbo_loop) {
     finder.Print();
   }
   return loop_tree;
 }
 
-Node* LoopTree::HeaderNode(Loop* loop) {
+#if V8_ENABLE_WEBASSEMBLY
+// static
+ZoneUnorderedSet<Node*>* LoopFinder::FindSmallInnermostLoopFromHeader(
+    Node* loop_header, AllNodes& all_nodes, Zone* zone, size_t max_size,
+    Purpose purpose) {
+  auto* visited = zone->New<ZoneUnorderedSet<Node*>>(zone);
+  std::vector<Node*> queue;
+
+  DCHECK_EQ(loop_header->opcode(), IrOpcode::kLoop);
+
+  queue.push_back(loop_header);
+  visited->insert(loop_header);
+
+#define ENQUEUE_USES(use_name, condition)             \
+  for (Node * use_name : node->uses()) {              \
+    if (condition && visited->count(use_name) == 0) { \
+      visited->insert(use_name);                      \
+      queue.push_back(use_name);                      \
+    }                                                 \
+  }
+  bool has_instruction_worth_peeling = false;
+  while (!queue.empty()) {
+    Node* node = queue.back();
+    queue.pop_back();
+    if (node->opcode() == IrOpcode::kEnd) {
+      // We reached the end of the graph. The end node is not part of the loop.
+      visited->erase(node);
+      continue;
+    }
+    if (visited->size() > max_size) return nullptr;
+    switch (node->opcode()) {
+      case IrOpcode::kLoop:
+        // Found nested loop.
+        if (node != loop_header) return nullptr;
+        ENQUEUE_USES(use, true);
+        break;
+      case IrOpcode::kLoopExit:
+        // Found nested loop.
+        if (node->InputAt(1) != loop_header) return nullptr;
+        // LoopExitValue/Effect uses are inside the loop. The rest are not.
+        ENQUEUE_USES(use, (use->opcode() == IrOpcode::kLoopExitEffect ||
+                           use->opcode() == IrOpcode::kLoopExitValue))
+        break;
+      case IrOpcode::kLoopExitEffect:
+      case IrOpcode::kLoopExitValue:
+        if (NodeProperties::GetControlInput(node)->InputAt(1) != loop_header) {
+          // Found nested loop.
+          return nullptr;
+        }
+        // All uses are outside the loop, do nothing.
+        break;
+      // If unrolling, call nodes are considered to have unbounded size,
+      // i.e. >max_size, with the exception of certain wasm builtins.
+      case IrOpcode::kTailCall:
+      case IrOpcode::kJSWasmCall:
+      case IrOpcode::kJSCall:
+        if (purpose == Purpose::kLoopUnrolling) return nullptr;
+        ENQUEUE_USES(use, true)
+        break;
+      case IrOpcode::kCall: {
+        if (purpose == Purpose::kLoopPeeling) {
+          ENQUEUE_USES(use, true);
+          break;
+        }
+        Node* callee = node->InputAt(0);
+        if (callee->opcode() != IrOpcode::kRelocatableInt32Constant &&
+            callee->opcode() != IrOpcode::kRelocatableInt64Constant) {
+          return nullptr;
+        }
+        intptr_t info =
+            OpParameter<RelocatablePtrConstantInfo>(callee->op()).value();
+        using WasmCode = v8::internal::wasm::WasmCode;
+        constexpr intptr_t unrollable_builtins[] = {
+            // Exists in every stack check.
+            WasmCode::kWasmStackGuard,
+            // Fast table operations.
+            WasmCode::kWasmTableGet, WasmCode::kWasmTableSet,
+            WasmCode::kWasmTableGetFuncRef, WasmCode::kWasmTableSetFuncRef,
+            WasmCode::kWasmTableGrow,
+            // Atomics.
+            WasmCode::kWasmAtomicNotify, WasmCode::kWasmI32AtomicWait,
+            WasmCode::kWasmI64AtomicWait,
+            // Exceptions.
+            WasmCode::kWasmAllocateFixedArray, WasmCode::kWasmThrow,
+            WasmCode::kWasmRethrow, WasmCode::kWasmRethrowExplicitContext,
+            // Fast wasm-gc operations.
+            WasmCode::kWasmRefFunc,
+            // While a built-in call, this is the slow path, so it should not
+            // prevent loop unrolling for stringview_wtf16.get_codeunit.
+            WasmCode::kWasmStringViewWtf16GetCodeUnit};
+        if (std::count(std::begin(unrollable_builtins),
+                       std::end(unrollable_builtins), info) == 0) {
+          return nullptr;
+        }
+        ENQUEUE_USES(use, true)
+        break;
+      }
+      case IrOpcode::kWasmStructGet: {
+        // When a chained load occurs in the loop, assume that peeling might
+        // help.
+        // Extending this idea to array.get/array.len has been found to hurt
+        // more than it helps (tested on Sheets, Feb 2023).
+        Node* object = node->InputAt(0);
+        if (object->opcode() == IrOpcode::kWasmStructGet &&
+            visited->find(object) != visited->end()) {
+          has_instruction_worth_peeling = true;
+        }
+        ENQUEUE_USES(use, true);
+        break;
+      }
+      case IrOpcode::kStringPrepareForGetCodeunit:
+        has_instruction_worth_peeling = true;
+        V8_FALLTHROUGH;
+      default:
+        ENQUEUE_USES(use, true)
+        break;
+    }
+  }
+
+  // Check that there is no floating control other than direct nodes to start().
+  // We do this by checking that all non-start control inputs of loop nodes are
+  // also in the loop.
+  // TODO(manoskouk): This is a safety check. Consider making it DEBUG-only when
+  // we are confident there is no incompatible floating control generated in
+  // wasm.
+  for (Node* node : *visited) {
+    // The loop header is allowed to point outside the loop.
+    if (node == loop_header) continue;
+
+    if (!all_nodes.IsLive(node)) continue;
+
+    for (Edge edge : node->input_edges()) {
+      Node* input = edge.to();
+      if (NodeProperties::IsControlEdge(edge) && visited->count(input) == 0 &&
+          input->opcode() != IrOpcode::kStart) {
+        FATAL(
+            "Floating control detected in wasm turbofan graph: Node #%d:%s is "
+            "inside loop headed by #%d, but its control dependency #%d:%s is "
+            "outside",
+            node->id(), node->op()->mnemonic(), loop_header->id(), input->id(),
+            input->op()->mnemonic());
+      }
+    }
+  }
+
+  // Only peel functions containing instructions for which loop peeling is known
+  // to be useful. TODO(7748): Add more instructions to get more benefits out of
+  // loop peeling.
+  if (purpose == Purpose::kLoopPeeling && !has_instruction_worth_peeling) {
+    return nullptr;
+  }
+  return visited;
+}
+#endif  // V8_ENABLE_WEBASSEMBLY
+
+bool LoopFinder::HasMarkedExits(LoopTree* loop_tree,
+                                const LoopTree::Loop* loop) {
+  // Look for returns and if projections that are outside the loop but whose
+  // control input is inside the loop.
+  Node* loop_node = loop_tree->GetLoopControl(loop);
+  for (Node* node : loop_tree->LoopNodes(loop)) {
+    for (Node* use : node->uses()) {
+      if (!loop_tree->Contains(loop, use)) {
+        bool unmarked_exit;
+        switch (node->opcode()) {
+          case IrOpcode::kLoopExit:
+            unmarked_exit = (node->InputAt(1) != loop_node);
+            break;
+          case IrOpcode::kLoopExitValue:
+          case IrOpcode::kLoopExitEffect:
+            unmarked_exit = (node->InputAt(1)->InputAt(1) != loop_node);
+            break;
+          default:
+            unmarked_exit = (use->opcode() != IrOpcode::kTerminate);
+        }
+        if (unmarked_exit) {
+          if (v8_flags.trace_turbo_loop) {
+            PrintF(
+                "Cannot peel loop %i. Loop exit without explicit mark: Node %i "
+                "(%s) is inside loop, but its use %i (%s) is outside.\n",
+                loop_node->id(), node->id(), node->op()->mnemonic(), use->id(),
+                use->op()->mnemonic());
+          }
+          return false;
+        }
+      }
+    }
+  }
+  return true;
+}
+
+Node* LoopTree::HeaderNode(const Loop* loop) {
   Node* first = *HeaderNodes(loop).begin();
   if (first->opcode() == IrOpcode::kLoop) return first;
   DCHECK(IrOpcode::IsPhiOpcode(first->opcode()));
@@ -542,6 +750,26 @@ Node* LoopTree::HeaderNode(Loop* loop) {
   return header;
 }
 
+Node* NodeCopier::map(Node* node, uint32_t copy_index) {
+  DCHECK_LT(copy_index, copy_count_);
+  if (node_map_.Get(node) == 0) return node;
+  return copies_->at(node_map_.Get(node) + copy_index);
+}
+
+void NodeCopier::Insert(Node* original, const NodeVector& new_copies) {
+  DCHECK_EQ(new_copies.size(), copy_count_);
+  node_map_.Set(original, copies_->size() + 1);
+  copies_->push_back(original);
+  copies_->insert(copies_->end(), new_copies.begin(), new_copies.end());
+}
+
+void NodeCopier::Insert(Node* original, Node* copy) {
+  DCHECK_EQ(copy_count_, 1);
+  node_map_.Set(original, copies_->size() + 1);
+  copies_->push_back(original);
+  copies_->push_back(copy);
+}
+
 }  // namespace compiler
 }  // namespace internal
 }  // namespace v8
