@@ -10,14 +10,7 @@ namespace v8 {
 namespace internal {
 namespace compiler {
 
-namespace {
-
-base::LazyInstance<TypeCache>::type kTypeCache = LAZY_INSTANCE_INITIALIZER;
-
-}  // namespace
-
-// static
-TypeCache const& TypeCache::Get() { return kTypeCache.Get(); }
+DEFINE_LAZY_LEAKY_OBJECT_GETTER(const TypeCache, TypeCache::Get)
 
 }  // namespace compiler
 }  // namespace internal
