@@ -3,6 +3,12 @@
 // found in the LICENSE file.
 
 #include "src/d8/d8-console.h"
+
+#include <stdio.h>
+
+#include <fstream>
+
+#include "include/v8-profiler.h"
 #include "src/d8/d8.h"
 #include "src/execution/isolate.h"
 
@@ -30,13 +36,62 @@ void WriteToFile(const char* prefix, FILE* file, Isolate* isolate,
     }
   }
   fprintf(file, "\n");
+  // Flush the file to avoid output to pile up in a buffer. Console output is
+  // often used for timing, so it should appear as soon as the code is executed.
+  fflush(file);
 }
+
+class FileOutputStream : public v8::OutputStream {
+ public:
+  explicit FileOutputStream(const char* filename)
+      : os_(filename, std::ios_base::out | std::ios_base::trunc) {}
+
+  WriteResult WriteAsciiChunk(char* data, int size) override {
+    os_.write(data, size);
+    return kContinue;
+  }
+
+  void EndOfStream() override { os_.close(); }
+
+ private:
+  std::ofstream os_;
+};
+
+static constexpr const char* kCpuProfileOutputFilename = "v8.prof";
+
+class StringOutputStream : public v8::OutputStream {
+ public:
+  WriteResult WriteAsciiChunk(char* data, int size) override {
+    os_.write(data, size);
+    return kContinue;
+  }
+
+  void EndOfStream() override {}
+
+  std::string result() { return os_.str(); }
+
+ private:
+  std::ostringstream os_;
+};
 }  // anonymous namespace
 
 D8Console::D8Console(Isolate* isolate) : isolate_(isolate) {
   default_timer_ = base::TimeTicks::Now();
 }
 
+D8Console::~D8Console() { DCHECK_NULL(profiler_); }
+
+void D8Console::DisposeProfiler() {
+  if (profiler_) {
+    if (profiler_active_) {
+      profiler_->StopProfiling(String::Empty(isolate_));
+      profiler_active_ = false;
+    }
+    profiler_->Dispose();
+    profiler_ = nullptr;
+  }
+}
+
 void D8Console::Assert(const debug::ConsoleCallArguments& args,
                        const v8::debug::ConsoleContext&) {
   // If no arguments given, the "first" argument is undefined which is
@@ -71,9 +126,35 @@ void D8Console::Debug(const debug::ConsoleCallArguments& args,
   WriteToFile("console.debug", stdout, isolate_, args);
 }
 
+void D8Console::Profile(const debug::ConsoleCallArguments& args,
+                        const v8::debug::ConsoleContext&) {
+  if (!profiler_) {
+    profiler_ = CpuProfiler::New(isolate_);
+  }
+  profiler_active_ = true;
+  profiler_->StartProfiling(String::Empty(isolate_), CpuProfilingOptions{});
+}
+
+void D8Console::ProfileEnd(const debug::ConsoleCallArguments& args,
+                           const v8::debug::ConsoleContext&) {
+  if (!profiler_) return;
+  CpuProfile* profile = profiler_->StopProfiling(String::Empty(isolate_));
+  profiler_active_ = false;
+  if (!profile) return;
+  if (Shell::HasOnProfileEndListener(isolate_)) {
+    StringOutputStream out;
+    profile->Serialize(&out);
+    Shell::TriggerOnProfileEndListener(isolate_, out.result());
+  } else {
+    FileOutputStream out(kCpuProfileOutputFilename);
+    profile->Serialize(&out);
+  }
+  profile->Delete();
+}
+
 void D8Console::Time(const debug::ConsoleCallArguments& args,
                      const v8::debug::ConsoleContext&) {
-  if (internal::FLAG_correctness_fuzzer_suppressions) return;
+  if (i::v8_flags.correctness_fuzzer_suppressions) return;
   if (args.Length() == 0) {
     default_timer_ = base::TimeTicks::Now();
   } else {
@@ -95,7 +176,7 @@ void D8Console::Time(const debug::ConsoleCallArguments& args,
 
 void D8Console::TimeEnd(const debug::ConsoleCallArguments& args,
                         const v8::debug::ConsoleContext&) {
-  if (internal::FLAG_correctness_fuzzer_suppressions) return;
+  if (i::v8_flags.correctness_fuzzer_suppressions) return;
   base::TimeDelta delta;
   if (args.Length() == 0) {
     delta = base::TimeTicks::Now() - default_timer_;
@@ -119,7 +200,7 @@ void D8Console::TimeEnd(const debug::ConsoleCallArguments& args,
 
 void D8Console::TimeStamp(const debug::ConsoleCallArguments& args,
                           const v8::debug::ConsoleContext&) {
-  if (internal::FLAG_correctness_fuzzer_suppressions) return;
+  if (i::v8_flags.correctness_fuzzer_suppressions) return;
   base::TimeDelta delta = base::TimeTicks::Now() - default_timer_;
   if (args.Length() == 0) {
     printf("console.timeStamp: default, %f\n", delta.InMillisecondsF());
@@ -136,7 +217,7 @@ void D8Console::TimeStamp(const debug::ConsoleCallArguments& args,
 
 void D8Console::Trace(const debug::ConsoleCallArguments& args,
                       const v8::debug::ConsoleContext&) {
-  if (internal::FLAG_correctness_fuzzer_suppressions) return;
+  if (i::v8_flags.correctness_fuzzer_suppressions) return;
   i::Isolate* i_isolate = reinterpret_cast<i::Isolate*>(isolate_);
   i_isolate->PrintStack(stderr, i::Isolate::kPrintStackConcise);
 }
