@@ -15,6 +15,7 @@ namespace internal {
   V(BigIntTooBig, "BigInt too big")                                            \
   V(CowArrayElementsChanged, "copy-on-write array's elements changed")         \
   V(CouldNotGrowElements, "failed to grow elements store")                     \
+  V(PrepareForOnStackReplacement, "prepare for on stack replacement (OSR)")    \
   V(DeoptimizeNow, "%_DeoptimizeNow")                                          \
   V(DivisionByZero, "division by zero")                                        \
   V(Hole, "hole")                                                              \
@@ -39,6 +40,7 @@ namespace internal {
   V(NaN, "NaN")                                                                \
   V(NoCache, "no cache")                                                       \
   V(NotABigInt, "not a BigInt")                                                \
+  V(NotABigInt64, "not a BigInt64")                                            \
   V(NotAHeapNumber, "not a heap number")                                       \
   V(NotAJavaScriptObject, "not a JavaScript object")                           \
   V(NotAJavaScriptObjectOrNullOrUndefined,                                     \
@@ -51,9 +53,12 @@ namespace internal {
   V(NotAString, "not a String")                                                \
   V(NotASymbol, "not a Symbol")                                                \
   V(NotInt32, "not int32")                                                     \
+  V(NotUint32, "not unsigned int32")                                           \
   V(OutOfBounds, "out of bounds")                                              \
   V(Overflow, "overflow")                                                      \
   V(Smi, "Smi")                                                                \
+  V(StoreToConstant, "Storing to a constant field")                            \
+  V(SuspendGeneratorIsDead, "SuspendGenerator is in a dead branch")            \
   V(TransitionedToMonomorphicIC, "IC transitioned to monomorphic")             \
   V(TransitionedToMegamorphicIC, "IC transitioned to megamorphic")             \
   V(Unknown, "(unknown)")                                                      \
@@ -81,7 +86,7 @@ constexpr DeoptimizeReason kFirstDeoptimizeReason =
     DeoptimizeReason::kArrayBufferWasDetached;
 constexpr DeoptimizeReason kLastDeoptimizeReason =
     DeoptimizeReason::kArrayLengthChanged;
-STATIC_ASSERT(static_cast<int>(kFirstDeoptimizeReason) == 0);
+static_assert(static_cast<int>(kFirstDeoptimizeReason) == 0);
 constexpr int kDeoptimizeReasonCount =
     static_cast<int>(kLastDeoptimizeReason) + 1;
 
@@ -91,6 +96,15 @@ size_t hash_value(DeoptimizeReason reason);
 
 V8_EXPORT_PRIVATE char const* DeoptimizeReasonToString(DeoptimizeReason reason);
 
+constexpr bool IsDeoptimizationWithoutCodeInvalidation(
+    DeoptimizeReason reason) {
+  // Maglev OSRs into Turbofan by first deoptimizing in order to restore the
+  // unoptimized frame layout. Since no actual assumptions in the Maglev code
+  // object are violated, it (and any associated cached optimized code) should
+  // not be invalidated s.t. we may reenter it in the future.
+  return reason == DeoptimizeReason::kPrepareForOnStackReplacement;
+}
+
 }  // namespace internal
 }  // namespace v8
 
