@@ -11,8 +11,7 @@ namespace internal {
 
 namespace {
 
-DEFINE_LAZY_LEAKY_OBJECT_GETTER(base::Thread::LocalStorageKey, GetThreadIdKey,
-                                base::Thread::CreateThreadLocalKey())
+thread_local int thread_id = 0;
 
 std::atomic<int> next_thread_id{1};
 
@@ -20,18 +19,14 @@ std::atomic<int> next_thread_id{1};
 
 // static
 ThreadId ThreadId::TryGetCurrent() {
-  int thread_id = base::Thread::GetThreadLocalInt(*GetThreadIdKey());
   return thread_id == 0 ? Invalid() : ThreadId(thread_id);
 }
 
 // static
 int ThreadId::GetCurrentThreadId() {
-  auto key = *GetThreadIdKey();
-  int thread_id = base::Thread::GetThreadLocalInt(key);
   if (thread_id == 0) {
     thread_id = next_thread_id.fetch_add(1);
     CHECK_LE(1, thread_id);
-    base::Thread::SetThreadLocalInt(key, thread_id);
   }
   return thread_id;
 }
