@@ -13,16 +13,18 @@ const char* BaseSpace::GetSpaceName(AllocationSpace space) {
       return "new_space";
     case OLD_SPACE:
       return "old_space";
-    case MAP_SPACE:
-      return "map_space";
     case CODE_SPACE:
       return "code_space";
+    case SHARED_SPACE:
+      return "shared_space";
     case LO_SPACE:
       return "large_object_space";
     case NEW_LO_SPACE:
       return "new_large_object_space";
     case CODE_LO_SPACE:
       return "code_large_object_space";
+    case SHARED_LO_SPACE:
+      return "shared_large_object_space";
     case RO_SPACE:
       return "read_only_space";
   }
