@@ -5,8 +5,11 @@
 #ifndef V8_HEAP_INVALIDATED_SLOTS_INL_H_
 #define V8_HEAP_INVALIDATED_SLOTS_INL_H_
 
+#include "src/base/logging.h"
 #include "src/heap/invalidated-slots.h"
+#include "src/heap/marking-state-inl.h"
 #include "src/heap/spaces.h"
+#include "src/objects/heap-object.h"
 #include "src/objects/objects-inl.h"
 #include "src/utils/allocation.h"
 
@@ -20,43 +23,55 @@ bool InvalidatedSlotsFilter::IsValid(Address slot) {
   DCHECK_LE(last_slot_, slot);
   last_slot_ = slot;
 #endif
-  if (slot < invalidated_start_) {
+  if (slot < current_.address) {
     return true;
   }
 
-  while (slot >= next_invalidated_start_) {
+  while (slot >= next_.address) {
     NextInvalidatedObject();
   }
 
-  HeapObject invalidated_object = HeapObject::FromAddress(invalidated_start_);
-
-  if (invalidated_size_ == 0) {
-    DCHECK(MarkCompactCollector::IsMapOrForwardedMap(invalidated_object.map()));
-    invalidated_size_ = invalidated_object.Size();
-  }
-
-  int offset = static_cast<int>(slot - invalidated_start_);
+  int offset = static_cast<int>(slot - current_.address);
 
   // OLD_TO_OLD can have slots in map word unlike other remembered sets.
   DCHECK_GE(offset, 0);
   DCHECK_IMPLIES(remembered_set_type_ != OLD_TO_OLD, offset > 0);
 
-  if (offset < invalidated_size_)
-    return offset == 0 ||
-           invalidated_object.IsValidSlot(invalidated_object.map(), offset);
+#if DEBUG
+  if (current_.is_live) {
+    HeapObject object = HeapObject::FromAddress(current_.address);
+    DCHECK_EQ(object.Size(), current_.size);
+  }
+#endif
+
+  if (offset < current_.size) {
+    // Slots in dead invalid objects are all treated as invalid.
+    if (!current_.is_live) return false;
+
+    // Map word is always a valid tagged reference.
+    if (offset == 0) return true;
+
+    // Check whether object has a tagged field at that particular offset.
+    HeapObject invalidated_object = HeapObject::FromAddress(current_.address);
+    DCHECK_IMPLIES(marking_state_,
+                   marking_state_->IsMarked(invalidated_object));
+    DCHECK(MarkCompactCollector::IsMapOrForwarded(invalidated_object.map()));
+    return invalidated_object.IsValidSlot(invalidated_object.map(), offset);
+  }
 
   NextInvalidatedObject();
   return true;
 }
 
 void InvalidatedSlotsFilter::NextInvalidatedObject() {
-  invalidated_start_ = next_invalidated_start_;
-  invalidated_size_ = 0;
+  current_ = next_;
 
   if (iterator_ == iterator_end_) {
-    next_invalidated_start_ = sentinel_;
+    next_ = {sentinel_, 0, false};
   } else {
-    next_invalidated_start_ = iterator_->address();
+    HeapObject object = iterator_->first;
+    bool is_live = marking_state_ ? marking_state_->IsMarked(object) : true;
+    next_ = {object.address(), iterator_->second, is_live};
     iterator_++;
   }
 }
@@ -87,7 +102,7 @@ void InvalidatedSlotsCleanup::Free(Address free_start, Address free_end) {
 
 void InvalidatedSlotsCleanup::NextInvalidatedObject() {
   if (iterator_ != iterator_end_) {
-    invalidated_start_ = iterator_->address();
+    invalidated_start_ = iterator_->first.address();
   } else {
     invalidated_start_ = sentinel_;
   }
