@@ -6,7 +6,6 @@
 
 #include <cstring>
 
-#include "src/base/platform/wrappers.h"
 #include "src/execution/isolate.h"
 #include "src/handles/global-handles.h"
 #include "src/logging/counters.h"
@@ -20,9 +19,9 @@
 #include "src/wasm/wasm-objects-inl.h"
 #endif  // V8_ENABLE_WEBASSEMBLY
 
-#define TRACE_BS(...)                                  \
-  do {                                                 \
-    if (FLAG_trace_backing_store) PrintF(__VA_ARGS__); \
+#define TRACE_BS(...)                                      \
+  do {                                                     \
+    if (v8_flags.trace_backing_store) PrintF(__VA_ARGS__); \
   } while (false)
 
 namespace v8 {
@@ -55,21 +54,6 @@ enum class AllocationStatus {
   kOtherFailure  // Failed for an unknown reason
 };
 
-// Attempts to allocate memory inside the sandbox currently fall back to
-// allocating memory outside of the sandbox if necessary. Once this fallback is
-// no longer allowed/possible, these cases will become allocation failures
-// instead. To track the frequency of such events, the outcome of memory
-// allocation attempts inside the sandbox is reported to UMA.
-//
-// See caged_memory_allocation_outcome in counters-definitions.h
-// This class and the entry in counters-definitions.h use the term "cage"
-// instead of "sandbox" for historical reasons.
-enum class CagedMemoryAllocationOutcome {
-  kSuccess,      // Allocation succeeded inside the cage
-  kOutsideCage,  // Allocation failed inside the cage but succeeded outside
-  kFailure,      // Allocation failed inside and outside of the cage
-};
-
 base::AddressRegion GetReservedRegion(bool has_guard_regions,
                                       void* buffer_start,
                                       size_t byte_capacity) {
@@ -96,7 +80,11 @@ base::AddressRegion GetReservedRegion(bool has_guard_regions,
 
 size_t GetReservationSize(bool has_guard_regions, size_t byte_capacity) {
 #if V8_TARGET_ARCH_64_BIT && V8_ENABLE_WEBASSEMBLY
-  if (has_guard_regions) return kFullGuardSize;
+  if (has_guard_regions) {
+    static_assert(kFullGuardSize > size_t{4} * GB);
+    DCHECK_LE(byte_capacity, size_t{4} * GB);
+    return kFullGuardSize;
+  }
 #else
   DCHECK(!has_guard_regions);
 #endif
@@ -109,30 +97,12 @@ void RecordStatus(Isolate* isolate, AllocationStatus status) {
       static_cast<int>(status));
 }
 
-// When the sandbox is active, this function records the outcome of attempts to
-// allocate memory inside the sandbox which fall back to allocating memory
-// outside of the sandbox. Passing a value of nullptr for the result indicates
-// that the memory could not be allocated at all.
-void RecordSandboxMemoryAllocationResult(Isolate* isolate, void* result) {
-  // This metric is only meaningful when the sandbox is active.
-#ifdef V8_SANDBOX
-  if (GetProcessWideSandbox()->is_initialized()) {
-    CagedMemoryAllocationOutcome outcome;
-    if (result) {
-      bool allocation_in_cage = GetProcessWideSandbox()->Contains(result);
-      outcome = allocation_in_cage ? CagedMemoryAllocationOutcome::kSuccess
-                                   : CagedMemoryAllocationOutcome::kOutsideCage;
-    } else {
-      outcome = CagedMemoryAllocationOutcome::kFailure;
-    }
-    isolate->counters()->caged_memory_allocation_outcome()->AddSample(
-        static_cast<int>(outcome));
-  }
-#endif
-}
-
 inline void DebugCheckZero(void* start, size_t byte_length) {
-#if DEBUG
+#ifdef DEBUG
+#ifdef V8_IS_TSAN
+  // TSan in debug mode is particularly slow. Skip this check for buffers >64MB.
+  if (byte_length > 64 * MB) return;
+#endif  // TSan debug build
   // Double check memory is zero-initialized. Despite being DEBUG-only,
   // this function is somewhat optimized for the benefit of test suite
   // execution times (some tests allocate several gigabytes).
@@ -169,6 +139,21 @@ void BackingStore::Clear() {
   type_specific_data_.v8_api_array_buffer_allocator = nullptr;
 }
 
+void BackingStore::FreeResizableMemory() {
+  DCHECK(free_on_destruct_);
+  DCHECK(!custom_deleter_);
+  DCHECK(is_resizable_by_js_ || is_wasm_memory_);
+  auto region =
+      GetReservedRegion(has_guard_regions_, buffer_start_, byte_capacity_);
+
+  PageAllocator* page_allocator = GetArrayBufferPageAllocator();
+  if (!region.is_empty()) {
+    FreePages(page_allocator, reinterpret_cast<void*>(region.begin()),
+              region.size());
+  }
+  Clear();
+}
+
 BackingStore::BackingStore(void* buffer_start, size_t byte_length,
                            size_t max_byte_length, size_t byte_capacity,
                            SharedFlag shared, ResizableFlag resizable,
@@ -181,7 +166,7 @@ BackingStore::BackingStore(void* buffer_start, size_t byte_length,
       byte_capacity_(byte_capacity),
       id_(next_backing_store_id_.fetch_add(1)),
       is_shared_(shared == SharedFlag::kShared),
-      is_resizable_(resizable == ResizableFlag::kResizable),
+      is_resizable_by_js_(resizable == ResizableFlag::kResizable),
       is_wasm_memory_(is_wasm_memory),
       holds_shared_ptr_to_allocator_(false),
       free_on_destruct_(free_on_destruct),
@@ -190,10 +175,10 @@ BackingStore::BackingStore(void* buffer_start, size_t byte_length,
       custom_deleter_(custom_deleter),
       empty_deleter_(empty_deleter) {
   // TODO(v8:11111): RAB / GSAB - Wasm integration.
-  DCHECK_IMPLIES(is_wasm_memory_, !is_resizable_);
-  DCHECK_IMPLIES(is_resizable_, !custom_deleter_);
-  DCHECK_IMPLIES(is_resizable_, free_on_destruct_);
-  DCHECK_IMPLIES(!is_wasm_memory && !is_resizable_,
+  DCHECK_IMPLIES(is_wasm_memory_, !is_resizable_by_js_);
+  DCHECK_IMPLIES(is_resizable_by_js_, !custom_deleter_);
+  DCHECK_IMPLIES(is_resizable_by_js_, free_on_destruct_);
+  DCHECK_IMPLIES(!is_wasm_memory && !is_resizable_by_js_,
                  byte_length_ == max_byte_length_);
   DCHECK_GE(max_byte_length_, byte_length_);
   DCHECK_GE(byte_capacity_, max_byte_length_);
@@ -207,24 +192,10 @@ BackingStore::~BackingStore() {
     return;
   }
 
-  PageAllocator* page_allocator = GetPlatformPageAllocator();
-  // TODO(saelo) here and elsewhere in this file, replace with
-  // GetArrayBufferPageAllocator once the fallback to the platform page
-  // allocator is no longer allowed.
-#ifdef V8_SANDBOX
-  if (GetProcessWideSandbox()->Contains(buffer_start_)) {
-    page_allocator = GetSandboxPageAllocator();
-  } else {
-    DCHECK(kAllowBackingStoresOutsideSandbox);
-  }
-#endif
-
 #if V8_ENABLE_WEBASSEMBLY
   if (is_wasm_memory_) {
     // TODO(v8:11111): RAB / GSAB - Wasm integration.
-    DCHECK(!is_resizable_);
-    DCHECK(free_on_destruct_);
-    DCHECK(!custom_deleter_);
+    DCHECK(!is_resizable_by_js_);
     size_t reservation_size =
         GetReservationSize(has_guard_regions_, byte_capacity_);
     TRACE_BS(
@@ -236,31 +207,14 @@ BackingStore::~BackingStore() {
       delete shared_data;
       type_specific_data_.shared_wasm_memory_data = nullptr;
     }
-
     // Wasm memories are always allocated through the page allocator.
-    auto region =
-        GetReservedRegion(has_guard_regions_, buffer_start_, byte_capacity_);
-
-    if (!region.is_empty()) {
-      FreePages(page_allocator, reinterpret_cast<void*>(region.begin()),
-                region.size());
-    }
-    Clear();
+    FreeResizableMemory();
     return;
   }
 #endif  // V8_ENABLE_WEBASSEMBLY
 
-  if (is_resizable_) {
-    DCHECK(free_on_destruct_);
-    DCHECK(!custom_deleter_);
-    auto region =
-        GetReservedRegion(has_guard_regions_, buffer_start_, byte_capacity_);
-
-    if (!region.is_empty()) {
-      FreePages(page_allocator, reinterpret_cast<void*>(region.begin()),
-                region.size());
-    }
-    Clear();
+  if (is_resizable_by_js_) {
+    FreeResizableMemory();
     return;
   }
   if (custom_deleter_) {
@@ -309,7 +263,7 @@ std::unique_ptr<BackingStore> BackingStore::Allocate(
         constexpr bool
             kDebugCheckZeroDisabledDueToNodeNotImplementingZeroInitAPI = true;
         if ((!(kDebugCheckZeroDisabledDueToNodeNotImplementingZeroInitAPI)) &&
-            !FLAG_mock_arraybuffer_allocator) {
+            !v8_flags.mock_arraybuffer_allocator) {
           DebugCheckZero(buffer_start, byte_length);
         }
       }
@@ -324,6 +278,14 @@ std::unique_ptr<BackingStore> BackingStore::Allocate(
       counters->array_buffer_new_size_failures()->AddSample(mb_length);
       return {};
     }
+#ifdef V8_ENABLE_SANDBOX
+    // Check to catch use of a non-sandbox-compatible ArrayBufferAllocator.
+    CHECK_WITH_MSG(GetProcessWideSandbox()->Contains(buffer_start),
+                   "When the V8 Sandbox is enabled, ArrayBuffer backing stores "
+                   "must be allocated inside the sandbox address space. Please "
+                   "use an appropriate ArrayBuffer::Allocator to allocate "
+                   "these buffers, or disable the sandbox.");
+#endif
   }
 
   auto result = new BackingStore(buffer_start,                  // start
@@ -356,33 +318,10 @@ void BackingStore::SetAllocatorFromIsolate(Isolate* isolate) {
   }
 }
 
-#if V8_ENABLE_WEBASSEMBLY
-// Allocate a backing store for a Wasm memory. Always use the page allocator
-// and add guard regions.
-std::unique_ptr<BackingStore> BackingStore::TryAllocateWasmMemory(
-    Isolate* isolate, size_t initial_pages, size_t maximum_pages,
-    SharedFlag shared) {
-  // Compute size of reserved memory.
-  size_t engine_max_pages = wasm::max_mem_pages();
-  maximum_pages = std::min(engine_max_pages, maximum_pages);
-
-  auto result = TryAllocateAndPartiallyCommitMemory(
-      isolate, initial_pages * wasm::kWasmPageSize,
-      maximum_pages * wasm::kWasmPageSize, wasm::kWasmPageSize, initial_pages,
-      maximum_pages, true, shared);
-  // Shared Wasm memories need an anchor for the memory object list.
-  if (result && shared == SharedFlag::kShared) {
-    result->type_specific_data_.shared_wasm_memory_data =
-        new SharedWasmMemoryData();
-  }
-  return result;
-}
-#endif  // V8_ENABLE_WEBASSEMBLY
-
 std::unique_ptr<BackingStore> BackingStore::TryAllocateAndPartiallyCommitMemory(
     Isolate* isolate, size_t byte_length, size_t max_byte_length,
     size_t page_size, size_t initial_pages, size_t maximum_pages,
-    bool is_wasm_memory, SharedFlag shared) {
+    WasmMemoryFlag wasm_memory, SharedFlag shared) {
   // Enforce engine limitation on the maximum number of pages.
   if (maximum_pages > std::numeric_limits<size_t>::max() / page_size) {
     return nullptr;
@@ -394,10 +333,11 @@ std::unique_ptr<BackingStore> BackingStore::TryAllocateAndPartiallyCommitMemory(
   TRACE_BS("BSw:try   %zu pages, %zu max\n", initial_pages, maximum_pages);
 
 #if V8_ENABLE_WEBASSEMBLY
-  bool guards = is_wasm_memory && trap_handler::IsTrapHandlerEnabled();
+  bool guards = wasm_memory == WasmMemoryFlag::kWasmMemory32 &&
+                trap_handler::IsTrapHandlerEnabled();
 #else
-  CHECK(!is_wasm_memory);
-  bool guards = false;
+  CHECK_EQ(WasmMemoryFlag::kNotWasm, wasm_memory);
+  constexpr bool guards = false;
 #endif  // V8_ENABLE_WEBASSEMBLY
 
   // For accounting purposes, whether a GC was necessary.
@@ -411,8 +351,10 @@ std::unique_ptr<BackingStore> BackingStore::TryAllocateAndPartiallyCommitMemory(
       // Collect garbage and retry.
       did_retry = true;
       // TODO(wasm): try Heap::EagerlyFreeExternalMemory() first?
-      isolate->heap()->MemoryPressureNotification(
-          MemoryPressureLevel::kCritical, true);
+      if (isolate != nullptr) {
+        isolate->heap()->MemoryPressureNotification(
+            MemoryPressureLevel::kCritical, true);
+      }
     }
     return false;
   };
@@ -424,30 +366,17 @@ std::unique_ptr<BackingStore> BackingStore::TryAllocateAndPartiallyCommitMemory(
   // Allocate pages (inaccessible by default).
   //--------------------------------------------------------------------------
   void* allocation_base = nullptr;
-  PageAllocator* page_allocator = GetPlatformPageAllocator();
+  PageAllocator* page_allocator = GetArrayBufferPageAllocator();
   auto allocate_pages = [&] {
-#ifdef V8_SANDBOX
-    page_allocator = GetSandboxPageAllocator();
-    allocation_base = AllocatePages(page_allocator, nullptr, reservation_size,
-                                    page_size, PageAllocator::kNoAccess);
-    if (allocation_base) return true;
-    // We currently still allow falling back to the platform page allocator if
-    // the sandbox page allocator fails. This will eventually be removed.
-    // TODO(chromium:1218005) once we forbid the fallback, we should have a
-    // single API, e.g. GetArrayBufferPageAllocator(), that returns the correct
-    // page allocator to use here depending on whether the sandbox is enabled
-    // or not.
-    if (!kAllowBackingStoresOutsideSandbox) return false;
-    page_allocator = GetPlatformPageAllocator();
-#endif
     allocation_base = AllocatePages(page_allocator, nullptr, reservation_size,
                                     page_size, PageAllocator::kNoAccess);
     return allocation_base != nullptr;
   };
   if (!gc_retry(allocate_pages)) {
     // Page allocator could not reserve enough pages.
-    RecordStatus(isolate, AllocationStatus::kOtherFailure);
-    RecordSandboxMemoryAllocationResult(isolate, nullptr);
+    if (isolate != nullptr) {
+      RecordStatus(isolate, AllocationStatus::kOtherFailure);
+    }
     TRACE_BS("BSw:try   failed to allocate pages\n");
     return {};
   }
@@ -482,10 +411,12 @@ std::unique_ptr<BackingStore> BackingStore::TryAllocateAndPartiallyCommitMemory(
 
   DebugCheckZero(buffer_start, byte_length);  // touch the bytes.
 
-  RecordStatus(isolate, did_retry ? AllocationStatus::kSuccessAfterRetry
-                                  : AllocationStatus::kSuccess);
-  RecordSandboxMemoryAllocationResult(isolate, allocation_base);
+  if (isolate != nullptr) {
+    RecordStatus(isolate, did_retry ? AllocationStatus::kSuccessAfterRetry
+                                    : AllocationStatus::kSuccess);
+  }
 
+  const bool is_wasm_memory = wasm_memory != WasmMemoryFlag::kNotWasm;
   ResizableFlag resizable =
       is_wasm_memory ? ResizableFlag::kNotResizable : ResizableFlag::kResizable;
 
@@ -514,41 +445,49 @@ std::unique_ptr<BackingStore> BackingStore::TryAllocateAndPartiallyCommitMemory(
 // and add guard regions.
 std::unique_ptr<BackingStore> BackingStore::AllocateWasmMemory(
     Isolate* isolate, size_t initial_pages, size_t maximum_pages,
-    SharedFlag shared) {
+    WasmMemoryFlag wasm_memory, SharedFlag shared) {
   // Wasm pages must be a multiple of the allocation page size.
   DCHECK_EQ(0, wasm::kWasmPageSize % AllocatePageSize());
+  DCHECK_LE(initial_pages, maximum_pages);
 
-  // Enforce engine limitation on the maximum number of pages.
-  if (initial_pages > wasm::max_mem_pages()) return nullptr;
+  DCHECK(wasm_memory == WasmMemoryFlag::kWasmMemory32 ||
+         wasm_memory == WasmMemoryFlag::kWasmMemory64);
 
-  auto backing_store =
-      TryAllocateWasmMemory(isolate, initial_pages, maximum_pages, shared);
-  if (maximum_pages == initial_pages) {
-    // If initial pages, and maximum are equal, nothing more to do return early.
-    return backing_store;
-  }
+  auto TryAllocate = [isolate, initial_pages, wasm_memory,
+                      shared](size_t maximum_pages) {
+    auto result = TryAllocateAndPartiallyCommitMemory(
+        isolate, initial_pages * wasm::kWasmPageSize,
+        maximum_pages * wasm::kWasmPageSize, wasm::kWasmPageSize, initial_pages,
+        maximum_pages, wasm_memory, shared);
+    if (result && shared == SharedFlag::kShared) {
+      result->type_specific_data_.shared_wasm_memory_data =
+          new SharedWasmMemoryData();
+    }
+    return result;
+  };
+  auto backing_store = TryAllocate(maximum_pages);
+  if (!backing_store && maximum_pages - initial_pages >= 4) {
+    // Retry with smaller maximum pages at each retry.
+    auto delta = (maximum_pages - initial_pages) / 4;
+    size_t sizes[] = {maximum_pages - delta, maximum_pages - 2 * delta,
+                      maximum_pages - 3 * delta, initial_pages};
 
-  // Retry with smaller maximum pages at each retry.
-  const int kAllocationTries = 3;
-  auto delta = (maximum_pages - initial_pages) / (kAllocationTries + 1);
-  size_t sizes[] = {maximum_pages - delta, maximum_pages - 2 * delta,
-                    maximum_pages - 3 * delta, initial_pages};
-
-  for (size_t i = 0; i < arraysize(sizes) && !backing_store; i++) {
-    backing_store =
-        TryAllocateWasmMemory(isolate, initial_pages, sizes[i], shared);
+    for (size_t reduced_maximum_pages : sizes) {
+      backing_store = TryAllocate(reduced_maximum_pages);
+      if (backing_store) break;
+    }
   }
   return backing_store;
 }
 
-std::unique_ptr<BackingStore> BackingStore::CopyWasmMemory(Isolate* isolate,
-                                                           size_t new_pages,
-                                                           size_t max_pages) {
+std::unique_ptr<BackingStore> BackingStore::CopyWasmMemory(
+    Isolate* isolate, size_t new_pages, size_t max_pages,
+    WasmMemoryFlag wasm_memory) {
   // Note that we could allocate uninitialized to save initialization cost here,
   // but since Wasm memories are allocated by the page allocator, the zeroing
   // cost is already built-in.
   auto new_backing_store = BackingStore::AllocateWasmMemory(
-      isolate, new_pages, max_pages,
+      isolate, new_pages, max_pages, wasm_memory,
       is_shared() ? SharedFlag::kShared : SharedFlag::kNotShared);
 
   if (!new_backing_store ||
@@ -666,19 +605,47 @@ void BackingStore::UpdateSharedWasmMemoryObjects(Isolate* isolate) {
 
 // Commit already reserved memory (for RAB backing stores (not shared)).
 BackingStore::ResizeOrGrowResult BackingStore::ResizeInPlace(
-    Isolate* isolate, size_t new_byte_length, size_t new_committed_length) {
+    Isolate* isolate, size_t new_byte_length) {
+  size_t page_size = AllocatePageSize();
+  size_t new_committed_pages;
+  bool round_return_value =
+      RoundUpToPageSize(new_byte_length, page_size,
+                        JSArrayBuffer::kMaxByteLength, &new_committed_pages);
+  CHECK(round_return_value);
+
+  size_t new_committed_length = new_committed_pages * page_size;
   DCHECK_LE(new_byte_length, new_committed_length);
   DCHECK(!is_shared());
 
   if (new_byte_length < byte_length_) {
-    // TOOO(v8:11111): Figure out a strategy for shrinking - when do we
-    // un-commit the memory?
-
     // Zero the memory so that in case the buffer is grown later, we have
-    // zeroed the contents already.
+    // zeroed the contents already. This is especially needed for the portion of
+    // the memory we're not going to decommit below (since it belongs to a
+    // committed page). In addition, we don't rely on all platforms always
+    // zeroing decommitted-then-recommitted memory, but zero the memory
+    // explicitly here.
     memset(reinterpret_cast<byte*>(buffer_start_) + new_byte_length, 0,
            byte_length_ - new_byte_length);
 
+    // Check if we can un-commit some pages.
+    size_t old_committed_pages;
+    round_return_value =
+        RoundUpToPageSize(byte_length_, page_size,
+                          JSArrayBuffer::kMaxByteLength, &old_committed_pages);
+    CHECK(round_return_value);
+    DCHECK_LE(new_committed_pages, old_committed_pages);
+
+    if (new_committed_pages < old_committed_pages) {
+      size_t old_committed_length = old_committed_pages * page_size;
+      if (!i::SetPermissions(
+              GetPlatformPageAllocator(),
+              reinterpret_cast<byte*>(buffer_start_) + new_committed_length,
+              old_committed_length - new_committed_length,
+              PageAllocator::kNoAccess)) {
+        return kFailure;
+      }
+    }
+
     // Changing the byte length wouldn't strictly speaking be needed, since
     // the JSArrayBuffer already stores the updated length. This is to keep
     // the BackingStore and JSArrayBuffer in sync.
@@ -707,7 +674,15 @@ BackingStore::ResizeOrGrowResult BackingStore::ResizeInPlace(
 
 // Commit already reserved memory (for GSAB backing stores (shared)).
 BackingStore::ResizeOrGrowResult BackingStore::GrowInPlace(
-    Isolate* isolate, size_t new_byte_length, size_t new_committed_length) {
+    Isolate* isolate, size_t new_byte_length) {
+  size_t page_size = AllocatePageSize();
+  size_t new_committed_pages;
+  bool round_return_value =
+      RoundUpToPageSize(new_byte_length, page_size,
+                        JSArrayBuffer::kMaxByteLength, &new_committed_pages);
+  CHECK(round_return_value);
+
+  size_t new_committed_length = new_committed_pages * page_size;
   DCHECK_LE(new_byte_length, new_committed_length);
   DCHECK(is_shared());
   // See comment in GrowWasmMemoryInPlace.
@@ -807,8 +782,7 @@ std::unique_ptr<BackingStore> BackingStore::EmptyBackingStore(
 }
 
 bool BackingStore::Reallocate(Isolate* isolate, size_t new_byte_length) {
-  CHECK(!is_wasm_memory_ && !custom_deleter_ && !globally_registered_ &&
-        free_on_destruct_ && !is_resizable_);
+  CHECK(CanReallocate());
   auto allocator = get_v8_api_array_buffer_allocator();
   CHECK_EQ(isolate->array_buffer_allocator(), allocator);
   CHECK_EQ(byte_length_, byte_capacity_);
