@@ -32,9 +32,7 @@ RELAXED_INT16_ACCESSORS(DescriptorArray, number_of_all_descriptors,
                         kNumberOfAllDescriptorsOffset)
 RELAXED_INT16_ACCESSORS(DescriptorArray, number_of_descriptors,
                         kNumberOfDescriptorsOffset)
-RELAXED_INT16_ACCESSORS(DescriptorArray, raw_number_of_marked_descriptors,
-                        kRawNumberOfMarkedDescriptorsOffset)
-RELAXED_INT16_ACCESSORS(DescriptorArray, filler16bits, kFiller16BitsOffset)
+RELAXED_UINT32_ACCESSORS(DescriptorArray, raw_gc_state, kRawGcStateOffset)
 
 inline int16_t DescriptorArray::number_of_slack_descriptors() const {
   return number_of_all_descriptors() - number_of_descriptors();
@@ -44,14 +42,6 @@ inline int DescriptorArray::number_of_entries() const {
   return number_of_descriptors();
 }
 
-inline int16_t DescriptorArray::CompareAndSwapRawNumberOfMarkedDescriptors(
-    int16_t expected, int16_t value) {
-  return base::Relaxed_CompareAndSwap(
-      reinterpret_cast<base::Atomic16*>(
-          FIELD_ADDR(*this, kRawNumberOfMarkedDescriptorsOffset)),
-      expected, value);
-}
-
 void DescriptorArray::CopyEnumCacheFrom(DescriptorArray array) {
   set_enum_cache(array.enum_cache());
 }
@@ -71,6 +61,25 @@ InternalIndex DescriptorArray::Search(Name name, Map map,
   return Search(name, number_of_own_descriptors, concurrent_search);
 }
 
+InternalIndex DescriptorArray::Search(int field_index, int valid_descriptors) {
+  for (int desc_index = field_index; desc_index < valid_descriptors;
+       ++desc_index) {
+    PropertyDetails details = GetDetails(InternalIndex(desc_index));
+    if (details.location() != PropertyLocation::kField) continue;
+    if (field_index == details.field_index()) {
+      return InternalIndex(desc_index);
+    }
+    DCHECK_LT(details.field_index(), field_index);
+  }
+  return InternalIndex::NotFound();
+}
+
+InternalIndex DescriptorArray::Search(int field_index, Map map) {
+  int number_of_own_descriptors = map.NumberOfOwnDescriptors();
+  if (number_of_own_descriptors == 0) return InternalIndex::NotFound();
+  return Search(field_index, number_of_own_descriptors);
+}
+
 InternalIndex DescriptorArray::SearchWithCache(Isolate* isolate, Name name,
                                                Map map) {
   DCHECK(name.IsUniqueName());
@@ -224,17 +233,24 @@ void DescriptorArray::Append(Descriptor* desc) {
   set_number_of_descriptors(descriptor_number + 1);
   Set(InternalIndex(descriptor_number), desc);
 
-  uint32_t hash = desc->GetKey()->hash();
+  uint32_t desc_hash = desc->GetKey()->hash();
+  // Hash value can't be zero, see String::ComputeAndSetHash()
+  uint32_t collision_hash = 0;
 
   int insertion;
 
   for (insertion = descriptor_number; insertion > 0; --insertion) {
     Name key = GetSortedKey(insertion - 1);
-    if (key.hash() <= hash) break;
+    collision_hash = key.hash();
+    if (collision_hash <= desc_hash) break;
     SetSortedKey(insertion, GetSortedKeyIndex(insertion - 1));
   }
 
   SetSortedKey(insertion, descriptor_number);
+
+  if (V8_LIKELY(collision_hash != desc_hash)) return;
+
+  CheckNameCollisionDuringInsertion(desc, desc_hash, insertion);
 }
 
 void DescriptorArray::SwapSortedKeys(int first, int second) {
@@ -243,6 +259,82 @@ void DescriptorArray::SwapSortedKeys(int first, int second) {
   SetSortedKey(second, first_key);
 }
 
+// static
+bool DescriptorArrayMarkingState::TryUpdateIndicesToMark(
+    unsigned gc_epoch, DescriptorArray array, DescriptorIndex index_to_mark) {
+  const auto current_epoch = gc_epoch & Epoch::kMask;
+  while (true) {
+    const RawGCStateType raw_gc_state = array.raw_gc_state(kRelaxedLoad);
+    const auto epoch_from_state = Epoch::decode(raw_gc_state);
+    RawGCStateType new_raw_gc_state = 0;
+    if (current_epoch != epoch_from_state) {
+      // If the epochs do not match, then either the raw_gc_state is zero
+      // (freshly allocated descriptor array) or the epoch from value lags
+      // by 1.
+      DCHECK_IMPLIES(raw_gc_state != 0,
+                     Epoch::decode(epoch_from_state + 1) == current_epoch);
+      new_raw_gc_state = NewState(current_epoch, 0, index_to_mark);
+    } else {
+      const DescriptorIndex already_marked = Marked::decode(raw_gc_state);
+      const DescriptorIndex delta = Delta::decode(raw_gc_state);
+      if ((already_marked + delta) >= index_to_mark) {
+        return false;
+      }
+      new_raw_gc_state = NewState(current_epoch, already_marked,
+                                  index_to_mark - already_marked);
+    }
+    if (SwapState(array, raw_gc_state, new_raw_gc_state)) {
+      return true;
+    }
+  }
+}
+
+// static
+std::pair<DescriptorArrayMarkingState::DescriptorIndex,
+          DescriptorArrayMarkingState::DescriptorIndex>
+DescriptorArrayMarkingState::AcquireDescriptorRangeToMark(
+    unsigned gc_epoch, DescriptorArray array) {
+  const auto current_epoch = gc_epoch & Epoch::kMask;
+  while (true) {
+    const RawGCStateType raw_gc_state = array.raw_gc_state(kRelaxedLoad);
+    const DescriptorIndex marked = Marked::decode(raw_gc_state);
+    const DescriptorIndex delta = Delta::decode(raw_gc_state);
+    // We may encounter an array here that was merely pushed to the marker. In
+    // such a case, we process all descriptors (if we succeed). The cases to
+    // check are:
+    // 1. Epoch mismatch: Happens when descriptors survive a GC cycle.
+    // 2. Epoch matches but marked/delta is 0: Can happen when descriptors are
+    //    newly allocated in the current cycle.
+    if (current_epoch != Epoch::decode(raw_gc_state) || (marked + delta) == 0) {
+      // In case number of descriptors is 0 and we reach the array through roots
+      // marking, mark also slack to get a proper transition from 0 marked to X
+      // marked. Otherwise, we would need to treat the state [0,0[ for marked
+      // and delta as valid state which leads to double-accounting through the
+      // marking barrier (when nof>1 in the barrier).
+      const int16_t number_of_descriptors =
+          array.number_of_descriptors() ? array.number_of_descriptors()
+                                        : array.number_of_all_descriptors();
+      DCHECK_GT(number_of_descriptors, 0);
+      if (SwapState(array, raw_gc_state,
+                    NewState(current_epoch, number_of_descriptors, 0))) {
+        return {0, number_of_descriptors};
+      }
+      continue;
+    }
+
+    // The delta is 0, so everything has been processed. Return the marked
+    // indices.
+    if (delta == 0) {
+      return {marked, marked};
+    }
+
+    if (SwapState(array, raw_gc_state,
+                  NewState(current_epoch, marked + delta, 0))) {
+      return {marked, marked + delta};
+    }
+  }
+}
+
 }  // namespace internal
 }  // namespace v8
 
