@@ -103,6 +103,10 @@ enum ElementsKind : uint8_t {
   PACKED_FROZEN_ELEMENTS,
   HOLEY_FROZEN_ELEMENTS,
 
+  // SharedArray elements kind. A FAST_SEALED_ELEMENTS variation useful to
+  // code specific paths for SharedArrays.
+  SHARED_ARRAY_ELEMENTS,
+
   // The "slow" kind.
   DICTIONARY_ELEMENTS,
 
@@ -139,7 +143,7 @@ enum ElementsKind : uint8_t {
   LAST_RAB_GSAB_FIXED_TYPED_ARRAY_ELEMENTS_KIND = RAB_GSAB_BIGINT64_ELEMENTS,
   TERMINAL_FAST_ELEMENTS_KIND = HOLEY_ELEMENTS,
   FIRST_ANY_NONEXTENSIBLE_ELEMENTS_KIND = PACKED_NONEXTENSIBLE_ELEMENTS,
-  LAST_ANY_NONEXTENSIBLE_ELEMENTS_KIND = HOLEY_FROZEN_ELEMENTS,
+  LAST_ANY_NONEXTENSIBLE_ELEMENTS_KIND = SHARED_ARRAY_ELEMENTS,
 
 // Alias for kSystemPointerSize-sized elements
 #ifdef V8_COMPRESS_POINTERS
@@ -158,15 +162,71 @@ constexpr int kFastElementsKindPackedToHoley =
     HOLEY_SMI_ELEMENTS - PACKED_SMI_ELEMENTS;
 
 constexpr int kElementsKindBits = 6;
-STATIC_ASSERT((1 << kElementsKindBits) > LAST_ELEMENTS_KIND);
-STATIC_ASSERT((1 << (kElementsKindBits - 1)) <= LAST_ELEMENTS_KIND);
+static_assert((1 << kElementsKindBits) > LAST_ELEMENTS_KIND);
+static_assert((1 << (kElementsKindBits - 1)) <= LAST_ELEMENTS_KIND);
 
 constexpr int kFastElementsKindBits = 3;
-STATIC_ASSERT((1 << kFastElementsKindBits) > LAST_FAST_ELEMENTS_KIND);
-STATIC_ASSERT((1 << (kFastElementsKindBits - 1)) <= LAST_FAST_ELEMENTS_KIND);
+static_assert((1 << kFastElementsKindBits) > LAST_FAST_ELEMENTS_KIND);
+static_assert((1 << (kFastElementsKindBits - 1)) <= LAST_FAST_ELEMENTS_KIND);
 
-V8_EXPORT_PRIVATE int ElementsKindToShiftSize(ElementsKind elements_kind);
-V8_EXPORT_PRIVATE int ElementsKindToByteSize(ElementsKind elements_kind);
+const uint8_t* TypedArrayAndRabGsabTypedArrayElementsKindShifts();
+const uint8_t* TypedArrayAndRabGsabTypedArrayElementsKindSizes();
+inline constexpr int ElementsKindToShiftSize(ElementsKind elements_kind) {
+  switch (elements_kind) {
+    case UINT8_ELEMENTS:
+    case INT8_ELEMENTS:
+    case UINT8_CLAMPED_ELEMENTS:
+    case RAB_GSAB_UINT8_ELEMENTS:
+    case RAB_GSAB_INT8_ELEMENTS:
+    case RAB_GSAB_UINT8_CLAMPED_ELEMENTS:
+      return 0;
+    case UINT16_ELEMENTS:
+    case INT16_ELEMENTS:
+    case RAB_GSAB_UINT16_ELEMENTS:
+    case RAB_GSAB_INT16_ELEMENTS:
+      return 1;
+    case UINT32_ELEMENTS:
+    case INT32_ELEMENTS:
+    case FLOAT32_ELEMENTS:
+    case RAB_GSAB_UINT32_ELEMENTS:
+    case RAB_GSAB_INT32_ELEMENTS:
+    case RAB_GSAB_FLOAT32_ELEMENTS:
+      return 2;
+    case PACKED_DOUBLE_ELEMENTS:
+    case HOLEY_DOUBLE_ELEMENTS:
+    case FLOAT64_ELEMENTS:
+    case BIGINT64_ELEMENTS:
+    case BIGUINT64_ELEMENTS:
+    case RAB_GSAB_FLOAT64_ELEMENTS:
+    case RAB_GSAB_BIGINT64_ELEMENTS:
+    case RAB_GSAB_BIGUINT64_ELEMENTS:
+      return 3;
+    case PACKED_SMI_ELEMENTS:
+    case PACKED_ELEMENTS:
+    case PACKED_FROZEN_ELEMENTS:
+    case PACKED_SEALED_ELEMENTS:
+    case PACKED_NONEXTENSIBLE_ELEMENTS:
+    case HOLEY_SMI_ELEMENTS:
+    case HOLEY_ELEMENTS:
+    case HOLEY_FROZEN_ELEMENTS:
+    case HOLEY_SEALED_ELEMENTS:
+    case HOLEY_NONEXTENSIBLE_ELEMENTS:
+    case DICTIONARY_ELEMENTS:
+    case FAST_SLOPPY_ARGUMENTS_ELEMENTS:
+    case SLOW_SLOPPY_ARGUMENTS_ELEMENTS:
+    case FAST_STRING_WRAPPER_ELEMENTS:
+    case SLOW_STRING_WRAPPER_ELEMENTS:
+    case SHARED_ARRAY_ELEMENTS:
+      return kTaggedSizeLog2;
+    case WASM_ARRAY_ELEMENTS:
+    case NO_ELEMENTS:
+      CONSTEXPR_UNREACHABLE();
+  }
+  CONSTEXPR_UNREACHABLE();
+}
+inline constexpr int ElementsKindToByteSize(ElementsKind elements_kind) {
+  return 1 << ElementsKindToShiftSize(elements_kind);
+}
 int GetDefaultHeaderSizeForElementsKind(ElementsKind elements_kind);
 const char* ElementsKindToString(ElementsKind kind);
 
@@ -225,23 +285,42 @@ inline bool IsFloatTypedArrayElementsKind(ElementsKind kind) {
          kind == RAB_GSAB_FLOAT32_ELEMENTS || kind == RAB_GSAB_FLOAT64_ELEMENTS;
 }
 
+inline bool IsSignedIntTypedArrayElementsKind(ElementsKind kind) {
+  return kind == INT8_ELEMENTS || kind == RAB_GSAB_INT8_ELEMENTS ||
+         kind == INT16_ELEMENTS || kind == RAB_GSAB_INT16_ELEMENTS ||
+         kind == INT32_ELEMENTS || kind == RAB_GSAB_INT32_ELEMENTS;
+}
+
+inline bool IsUnsignedIntTypedArrayElementsKind(ElementsKind kind) {
+  return kind == UINT8_CLAMPED_ELEMENTS ||
+         kind == RAB_GSAB_UINT8_CLAMPED_ELEMENTS || kind == UINT8_ELEMENTS ||
+         kind == RAB_GSAB_UINT8_ELEMENTS || kind == UINT16_ELEMENTS ||
+         kind == RAB_GSAB_UINT16_ELEMENTS || kind == UINT32_ELEMENTS ||
+         kind == RAB_GSAB_UINT32_ELEMENTS;
+}
+
 inline bool IsWasmArrayElementsKind(ElementsKind kind) {
   return kind == WASM_ARRAY_ELEMENTS;
 }
 
+inline bool IsSharedArrayElementsKind(ElementsKind kind) {
+  return kind == SHARED_ARRAY_ELEMENTS;
+}
+
 inline bool IsTerminalElementsKind(ElementsKind kind) {
   return kind == TERMINAL_FAST_ELEMENTS_KIND ||
-         IsTypedArrayElementsKind(kind) ||
+         IsTypedArrayOrRabGsabTypedArrayElementsKind(kind) ||
          IsRabGsabTypedArrayElementsKind(kind);
 }
 
 inline bool IsFastElementsKind(ElementsKind kind) {
-  STATIC_ASSERT(FIRST_FAST_ELEMENTS_KIND == 0);
+  static_assert(FIRST_FAST_ELEMENTS_KIND == 0);
   return kind <= LAST_FAST_ELEMENTS_KIND;
 }
 
 inline bool IsTransitionElementsKind(ElementsKind kind) {
-  return IsFastElementsKind(kind) || IsTypedArrayElementsKind(kind) ||
+  return IsFastElementsKind(kind) ||
+         IsTypedArrayOrRabGsabTypedArrayElementsKind(kind) ||
          kind == FAST_SLOPPY_ARGUMENTS_ELEMENTS ||
          kind == FAST_STRING_WRAPPER_ELEMENTS;
 }
@@ -266,29 +345,31 @@ inline bool IsAnyNonextensibleElementsKindUnchecked(ElementsKind kind) {
 
 inline bool IsAnyNonextensibleElementsKind(ElementsKind kind) {
   DCHECK_IMPLIES(IsAnyNonextensibleElementsKindUnchecked(kind),
-                 FLAG_enable_sealed_frozen_elements_kind);
+                 v8_flags.enable_sealed_frozen_elements_kind);
   return IsAnyNonextensibleElementsKindUnchecked(kind);
 }
 
 inline bool IsNonextensibleElementsKind(ElementsKind kind) {
   DCHECK_IMPLIES(base::IsInRange(kind, PACKED_NONEXTENSIBLE_ELEMENTS,
                                  HOLEY_NONEXTENSIBLE_ELEMENTS),
-                 FLAG_enable_sealed_frozen_elements_kind);
+                 v8_flags.enable_sealed_frozen_elements_kind);
   return base::IsInRange(kind, PACKED_NONEXTENSIBLE_ELEMENTS,
                          HOLEY_NONEXTENSIBLE_ELEMENTS);
 }
 
 inline bool IsSealedElementsKind(ElementsKind kind) {
   DCHECK_IMPLIES(
-      base::IsInRange(kind, PACKED_SEALED_ELEMENTS, HOLEY_SEALED_ELEMENTS),
-      FLAG_enable_sealed_frozen_elements_kind);
-  return base::IsInRange(kind, PACKED_SEALED_ELEMENTS, HOLEY_SEALED_ELEMENTS);
+      base::IsInRange(kind, PACKED_SEALED_ELEMENTS, HOLEY_SEALED_ELEMENTS) ||
+          IsSharedArrayElementsKind(kind),
+      v8_flags.enable_sealed_frozen_elements_kind);
+  return IsSharedArrayElementsKind(kind) ||
+         base::IsInRange(kind, PACKED_SEALED_ELEMENTS, HOLEY_SEALED_ELEMENTS);
 }
 
 inline bool IsFrozenElementsKind(ElementsKind kind) {
   DCHECK_IMPLIES(
       base::IsInRange(kind, PACKED_FROZEN_ELEMENTS, HOLEY_FROZEN_ELEMENTS),
-      FLAG_enable_sealed_frozen_elements_kind);
+      v8_flags.enable_sealed_frozen_elements_kind);
   return base::IsInRange(kind, PACKED_FROZEN_ELEMENTS, HOLEY_FROZEN_ELEMENTS);
 }
 
@@ -312,7 +393,7 @@ inline bool IsAnyHoleyNonextensibleElementsKind(ElementsKind kind) {
   DCHECK_IMPLIES(kind == HOLEY_NONEXTENSIBLE_ELEMENTS ||
                      kind == HOLEY_SEALED_ELEMENTS ||
                      kind == HOLEY_FROZEN_ELEMENTS,
-                 FLAG_enable_sealed_frozen_elements_kind);
+                 v8_flags.enable_sealed_frozen_elements_kind);
   return kind == HOLEY_NONEXTENSIBLE_ELEMENTS ||
          kind == HOLEY_SEALED_ELEMENTS || kind == HOLEY_FROZEN_ELEMENTS;
 }
