@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # Copyright 2013 the V8 project authors. All rights reserved.
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -26,12 +26,9 @@
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-# for py2/py3 compatibility
-from __future__ import print_function
-
 import argparse
 import datetime
-import httplib
+from distutils.version import LooseVersion
 import glob
 import imp
 import json
@@ -43,12 +40,14 @@ import sys
 import textwrap
 import time
 import urllib
-import urllib2
 
 from git_recipes import GitRecipesMixin
 from git_recipes import GitFailedException
 
-CHANGELOG_FILE = "ChangeLog"
+import http.client as httplib
+import urllib.request as urllib2
+
+
 DAY_IN_SECONDS = 24 * 60 * 60
 PUSH_MSG_GIT_RE = re.compile(r".* \(based on (?P<git_rev>[a-fA-F0-9]+)\)$")
 PUSH_MSG_NEW_RE = re.compile(r"^Version \d+\.\d+\.\d+$")
@@ -93,109 +92,6 @@ def MSub(rexp, replacement, text):
   return re.sub(rexp, replacement, text, flags=re.MULTILINE)
 
 
-def Fill80(line):
-  # Replace tabs and remove surrounding space.
-  line = re.sub(r"\t", r"        ", line.strip())
-
-  # Format with 8 characters indentation and line width 80.
-  return textwrap.fill(line, width=80, initial_indent="        ",
-                       subsequent_indent="        ")
-
-
-def MakeComment(text):
-  return MSub(r"^( ?)", "#", text)
-
-
-def StripComments(text):
-  # Use split not splitlines to keep terminal newlines.
-  return "\n".join(filter(lambda x: not x.startswith("#"), text.split("\n")))
-
-
-def MakeChangeLogBody(commit_messages, auto_format=False):
-  result = ""
-  added_titles = set()
-  for (title, body, author) in commit_messages:
-    # TODO(machenbach): Better check for reverts. A revert should remove the
-    # original CL from the actual log entry.
-    title = title.strip()
-    if auto_format:
-      # Only add commits that set the LOG flag correctly.
-      log_exp = r"^[ \t]*LOG[ \t]*=[ \t]*(?:(?:Y(?:ES)?)|TRUE)"
-      if not re.search(log_exp, body, flags=re.I | re.M):
-        continue
-      # Never include reverts.
-      if title.startswith("Revert "):
-        continue
-      # Don't include duplicates.
-      if title in added_titles:
-        continue
-
-    # Add and format the commit's title and bug reference. Move dot to the end.
-    added_titles.add(title)
-    raw_title = re.sub(r"(\.|\?|!)$", "", title)
-    bug_reference = MakeChangeLogBugReference(body)
-    space = " " if bug_reference else ""
-    result += "%s\n" % Fill80("%s%s%s." % (raw_title, space, bug_reference))
-
-    # Append the commit's author for reference if not in auto-format mode.
-    if not auto_format:
-      result += "%s\n" % Fill80("(%s)" % author.strip())
-
-    result += "\n"
-  return result
-
-
-def MakeChangeLogBugReference(body):
-  """Grep for "BUG=xxxx" lines in the commit message and convert them to
-  "(issue xxxx)".
-  """
-  crbugs = []
-  v8bugs = []
-
-  def AddIssues(text):
-    ref = re.match(r"^BUG[ \t]*=[ \t]*(.+)$", text.strip())
-    if not ref:
-      return
-    for bug in ref.group(1).split(","):
-      bug = bug.strip()
-      match = re.match(r"^v8:(\d+)$", bug)
-      if match: v8bugs.append(int(match.group(1)))
-      else:
-        match = re.match(r"^(?:chromium:)?(\d+)$", bug)
-        if match: crbugs.append(int(match.group(1)))
-
-  # Add issues to crbugs and v8bugs.
-  map(AddIssues, body.splitlines())
-
-  # Filter duplicates, sort, stringify.
-  crbugs = map(str, sorted(set(crbugs)))
-  v8bugs = map(str, sorted(set(v8bugs)))
-
-  bug_groups = []
-  def FormatIssues(prefix, bugs):
-    if len(bugs) > 0:
-      plural = "s" if len(bugs) > 1 else ""
-      bug_groups.append("%sissue%s %s" % (prefix, plural, ", ".join(bugs)))
-
-  FormatIssues("", v8bugs)
-  FormatIssues("Chromium ", crbugs)
-
-  if len(bug_groups) > 0:
-    return "(%s)" % ", ".join(bug_groups)
-  else:
-    return ""
-
-
-def SortingKey(version):
-  """Key for sorting version number strings: '3.11' > '3.2.1.1'"""
-  version_keys = map(int, version.split("."))
-  # Fill up to full version numbers to normalize comparison.
-  while len(version_keys) < 4:  # pragma: no cover
-    version_keys.append(0)
-  # Fill digits.
-  return ".".join(map("{0:04d}".format, version_keys))
-
-
 # Some commands don't like the pipe, e.g. calling vi from within the script or
 # from subscripts like git cl upload.
 def Command(cmd, args="", prefix="", pipe=True, cwd=None):
@@ -207,7 +103,7 @@ def Command(cmd, args="", prefix="", pipe=True, cwd=None):
   sys.stdout.flush()
   try:
     if pipe:
-      return subprocess.check_output(cmd_line, shell=True, cwd=cwd)
+      return subprocess.check_output(cmd_line, shell=True, cwd=cwd).decode('utf-8')
     else:
       return subprocess.check_call(cmd_line, shell=True, cwd=cwd)
   except subprocess.CalledProcessError:
@@ -218,15 +114,15 @@ def Command(cmd, args="", prefix="", pipe=True, cwd=None):
 
 
 def SanitizeVersionTag(tag):
-    version_without_prefix = re.compile(r"^\d+\.\d+\.\d+(?:\.\d+)?$")
-    version_with_prefix = re.compile(r"^tags\/\d+\.\d+\.\d+(?:\.\d+)?$")
+  version_without_prefix = re.compile(r"^\d+\.\d+\.\d+(?:\.\d+)?$")
+  version_with_prefix = re.compile(r"^tags\/\d+\.\d+\.\d+(?:\.\d+)?$")
 
-    if version_without_prefix.match(tag):
-      return tag
-    elif version_with_prefix.match(tag):
-        return tag[len("tags/"):]
-    else:
-      return None
+  if version_without_prefix.match(tag):
+    return tag
+  elif version_with_prefix.match(tag):
+    return tag[len("tags/"):]
+  else:
+    return None
 
 
 def NormalizeVersionTags(version_tags):
@@ -250,6 +146,7 @@ class SideEffectHandler(object):  # pragma: no cover
     return Command(cmd, args, prefix, pipe, cwd=cwd)
 
   def ReadLine(self):
+    sys.stdout.flush()
     return sys.stdin.readline().strip()
 
   def ReadURL(self, url, params=None):
@@ -282,9 +179,6 @@ class SideEffectHandler(object):  # pragma: no cover
   def Sleep(self, seconds):
     time.sleep(seconds)
 
-  def GetDate(self):
-    return datetime.date.today().strftime("%Y-%m-%d")
-
   def GetUTCStamp(self):
     return time.mktime(datetime.datetime.utcnow().timetuple())
 
@@ -311,13 +205,13 @@ class VCInterface(object):
   def GetBranches(self):
     raise NotImplementedError()
 
-  def MasterBranch(self):
+  def MainBranch(self):
     raise NotImplementedError()
 
   def CandidateBranch(self):
     raise NotImplementedError()
 
-  def RemoteMasterBranch(self):
+  def RemoteMainBranch(self):
     raise NotImplementedError()
 
   def RemoteCandidateBranch(self):
@@ -345,7 +239,7 @@ class GitInterface(VCInterface):
     self.step.Git("fetch")
 
   def GetTags(self):
-     return self.step.Git("tag").strip().splitlines()
+    return self.step.Git("tag").strip().splitlines()
 
   def GetBranches(self):
     # Get relevant remote branches, e.g. "branch-heads/3.25".
@@ -353,16 +247,16 @@ class GitInterface(VCInterface):
         lambda s: re.match(r"^branch\-heads/\d+\.\d+$", s),
         self.step.GitRemotes())
     # Remove 'branch-heads/' prefix.
-    return map(lambda s: s[13:], branches)
+    return [b[13:] for b in branches]
 
-  def MasterBranch(self):
-    return "master"
+  def MainBranch(self):
+    return "main"
 
   def CandidateBranch(self):
     return "candidates"
 
-  def RemoteMasterBranch(self):
-    return "origin/master"
+  def RemoteMainBranch(self):
+    return "origin/main"
 
   def RemoteCandidateBranch(self):
     return "origin/candidates"
@@ -372,7 +266,7 @@ class GitInterface(VCInterface):
     # want.
     if name.startswith('refs/'):
       return name
-    if name in ["candidates", "master"]:
+    if name in ["candidates", "main"]:
       return "refs/remotes/origin/%s" % name
     try:
       # Check if branch is in heads.
@@ -527,9 +421,6 @@ class Step(GitRecipesMixin):
     cmd = lambda: self._side_effect_handler.ReadURL(url, params)
     return self.Retry(cmd, retry_on, wait_plan)
 
-  def GetDate(self):
-    return self._side_effect_handler.GetDate()
-
   def Die(self, msg=""):
     if msg != "":
       print("Error: %s" % msg)
@@ -574,8 +465,8 @@ class Step(GitRecipesMixin):
     if not self.GitIsWorkdirClean():  # pragma: no cover
       self.Die("Workspace is not clean. Please commit or undo your changes.")
 
-    # Checkout master in case the script was left on a work branch.
-    self.GitCheckout('origin/master')
+    # Checkout main in case the script was left on a work branch.
+    self.GitCheckout('origin/main')
 
     # Fetch unfetched revisions.
     self.vc.Fetch()
@@ -585,7 +476,7 @@ class Step(GitRecipesMixin):
     self.DeleteBranch(self._config["BRANCHNAME"])
 
   def CommonCleanup(self):
-    self.GitCheckout('origin/master')
+    self.GitCheckout('origin/main')
     self.GitDeleteBranch(self._config["BRANCHNAME"])
 
     # Clean up all temporary files.
@@ -611,7 +502,7 @@ class Step(GitRecipesMixin):
   def WaitForLGTM(self):
     print ("Please wait for an LGTM, then type \"LGTM<Return>\" to commit "
            "your change. (If you need to iterate on the patch or double check "
-           "that it's sane, do so in another shell, but remember to not "
+           "that it's sensible, do so in another shell, but remember to not "
            "change the headline of the uploaded CL.")
     answer = ""
     while answer != "LGTM":
@@ -642,22 +533,13 @@ class Step(GitRecipesMixin):
       self.WaitForResolvingConflicts(patch_file)
 
   def GetVersionTag(self, revision):
-    tag = self.Git("describe --tags %s" % revision).strip()
-    return SanitizeVersionTag(tag)
+    tags = self.Git(f"tag --points-at {revision}").strip().split('\n')
+    for tag in tags:
+      sanitized_tag = SanitizeVersionTag(tag)
+      if sanitized_tag:
+        return sanitized_tag
 
-  def GetRecentReleases(self, max_age):
-    # Make sure tags are fetched.
-    self.Git("fetch origin +refs/tags/*:refs/tags/*")
-
-    # Current timestamp.
-    time_now = int(self._side_effect_handler.GetUTCStamp())
-
-    # List every tag from a given period.
-    revisions = self.Git("rev-list --max-age=%d --tags" %
-                         int(time_now - max_age)).strip()
-
-    # Filter out revisions who's tag is off by one or more commits.
-    return filter(lambda r: self.GetVersionTag(r), revisions.splitlines())
+    return None
 
   def GetLatestVersion(self):
     # Use cached version if available.
@@ -671,7 +553,7 @@ class Step(GitRecipesMixin):
     only_version_tags = NormalizeVersionTags(all_tags)
 
     version = sorted(only_version_tags,
-                     key=SortingKey, reverse=True)[0]
+                     key=LooseVersion, reverse=True)[0]
     self["latest_version"] = version
     return version
 
@@ -705,13 +587,13 @@ class Step(GitRecipesMixin):
     if match:
       # Legacy: In the old process there's one level of indirection. The
       # version is on the candidates branch and points to the real release
-      # base on master through the commit message.
+      # base on main through the commit message.
       return match.group("git_rev")
     match = PUSH_MSG_NEW_RE.match(title)
     if match:
-      # This is a new-style v8 version branched from master. The commit
+      # This is a new-style v8 version branched from main. The commit
       # "latest_hash" is the version-file change. Its parent is the release
-      # base on master.
+      # base on main.
       return self.GitLog(n=1, format="%H", git_hash="%s^" % latest_hash)
 
     self.Die("Unknown latest release: %s" % latest_hash)
@@ -789,23 +671,27 @@ class UploadStep(Step):
 
     self.GitUpload(reviewer, self._options.force_upload,
                    bypass_hooks=self._options.bypass_upload_hooks,
-                   cc=self._options.cc, tbr_reviewer=tbr_reviewer)
+                   tbr_reviewer=tbr_reviewer)
 
 
 def MakeStep(step_class=Step, number=0, state=None, config=None,
              options=None, side_effect_handler=DEFAULT_SIDE_EFFECT_HANDLER):
-    # Allow to pass in empty dictionaries.
-    state = state if state is not None else {}
-    config = config if config is not None else {}
+  # Allow to pass in empty dictionaries.
+  state = state if state is not None else {}
+  config = config if config is not None else {}
 
-    try:
-      message = step_class.MESSAGE
-    except AttributeError:
-      message = step_class.__name__
+  try:
+    message = step_class.MESSAGE
+  except AttributeError:
+    message = step_class.__name__
 
-    return step_class(message, number=number, config=config,
-                      state=state, options=options,
-                      handler=side_effect_handler)
+  return step_class(
+      message,
+      number=number,
+      config=config,
+      state=state,
+      options=options,
+      handler=side_effect_handler)
 
 
 class ScriptsBase(object):
