@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # Copyright 2013 the V8 project authors. All rights reserved.
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are
@@ -26,9 +26,7 @@
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-# for py2/py3 compatibility
-from __future__ import print_function
-
+import json
 import os
 import shutil
 import tempfile
@@ -41,25 +39,19 @@ import auto_roll
 import common_includes
 from common_includes import *
 import create_release
-from create_release import CreateRelease
+from create_release import *
 import merge_to_branch
 from merge_to_branch import MergeToBranch
-import push_to_candidates
-from push_to_candidates import *
-from auto_tag import AutoTag
 import roll_merge
 from roll_merge import RollMerge
 
 TEST_CONFIG = {
   "DEFAULT_CWD": None,
   "BRANCHNAME": "test-prepare-push",
-  "CANDIDATESBRANCH": "test-candidates-push",
-  "PERSISTFILE_BASENAME": "/tmp/test-v8-push-to-candidates-tempfile",
-  "CHANGELOG_ENTRY_FILE":
-      "/tmp/test-v8-push-to-candidates-tempfile-changelog-entry",
-  "PATCH_FILE": "/tmp/test-v8-push-to-candidates-tempfile-patch",
-  "COMMITMSG_FILE": "/tmp/test-v8-push-to-candidates-tempfile-commitmsg",
-  "CHROMIUM": "/tmp/test-v8-push-to-candidates-tempfile-chromium",
+  "PERSISTFILE_BASENAME": "/tmp/test-create-releases-tempfile",
+  "PATCH_FILE": "/tmp/test-v8-create-releases-tempfile-tempfile-patch",
+  "COMMITMSG_FILE": "/tmp/test-v8-create-releases-tempfile-commitmsg",
+  "CHROMIUM": "/tmp/test-create-releases-tempfile-chromium",
   "SETTINGS_LOCATION": None,
   "ALREADY_MERGING_SENTINEL_FILE":
       "/tmp/test-merge-to-branch-tempfile-already-merging",
@@ -98,141 +90,9 @@ class ToplevelTest(unittest.TestCase):
                 ]
     self.assertEquals(expected, NormalizeVersionTags(input))
 
-  def testMakeComment(self):
-    self.assertEquals("#   Line 1\n#   Line 2\n#",
-                      MakeComment("    Line 1\n    Line 2\n"))
-    self.assertEquals("#Line 1\n#Line 2",
-                      MakeComment("Line 1\n Line 2"))
-
-  def testStripComments(self):
-    self.assertEquals("    Line 1\n    Line 3\n",
-        StripComments("    Line 1\n#   Line 2\n    Line 3\n#\n"))
-    self.assertEquals("\nLine 2 ### Test\n #",
-        StripComments("###\n# \n\n#  Line 1\nLine 2 ### Test\n #"))
-
-  def testMakeChangeLogBodySimple(self):
-    commits = [
-          ["Title text 1",
-           "Title text 1\n\nBUG=\n",
-           "author1@chromium.org"],
-          ["Title text 2.",
-           "Title text 2\n\nBUG=1234\n",
-           "author2@chromium.org"],
-        ]
-    self.assertEquals("        Title text 1.\n"
-                      "        (author1@chromium.org)\n\n"
-                      "        Title text 2 (Chromium issue 1234).\n"
-                      "        (author2@chromium.org)\n\n",
-                      MakeChangeLogBody(commits))
-
-  def testMakeChangeLogBodyEmpty(self):
-    self.assertEquals("", MakeChangeLogBody([]))
-
-  def testMakeChangeLogBodyAutoFormat(self):
-    commits = [
-          ["Title text 1!",
-           "Title text 1\nLOG=y\nBUG=\n",
-           "author1@chromium.org"],
-          ["Title text 2",
-           "Title text 2\n\nBUG=1234\n",
-           "author2@chromium.org"],
-          ["Title text 3",
-           "Title text 3\n\nBUG=1234\nLOG = Yes\n",
-           "author3@chromium.org"],
-          ["Title text 3",
-           "Title text 4\n\nBUG=1234\nLOG=\n",
-           "author4@chromium.org"],
-        ]
-    self.assertEquals("        Title text 1.\n\n"
-                      "        Title text 3 (Chromium issue 1234).\n\n",
-                      MakeChangeLogBody(commits, True))
-
-  def testRegressWrongLogEntryOnTrue(self):
-    body = """
-Check elimination: Learn from if(CompareMap(x)) on true branch.
-
-BUG=
-R=verwaest@chromium.org
-
-Committed: https://code.google.com/p/v8/source/detail?r=18210
-"""
-    self.assertEquals("", MakeChangeLogBody([["title", body, "author"]], True))
-
-  def testMakeChangeLogBugReferenceEmpty(self):
-    self.assertEquals("", MakeChangeLogBugReference(""))
-    self.assertEquals("", MakeChangeLogBugReference("LOG="))
-    self.assertEquals("", MakeChangeLogBugReference(" BUG ="))
-    self.assertEquals("", MakeChangeLogBugReference("BUG=none\t"))
-
-  def testMakeChangeLogBugReferenceSimple(self):
-    self.assertEquals("(issue 987654)",
-                      MakeChangeLogBugReference("BUG = v8:987654"))
-    self.assertEquals("(Chromium issue 987654)",
-                      MakeChangeLogBugReference("BUG=987654 "))
-
-  def testMakeChangeLogBugReferenceFromBody(self):
-    self.assertEquals("(Chromium issue 1234567)",
-                      MakeChangeLogBugReference("Title\n\nTBR=\nBUG=\n"
-                                                " BUG=\tchromium:1234567\t\n"
-                                                "R=somebody\n"))
-
-  def testMakeChangeLogBugReferenceMultiple(self):
-    # All issues should be sorted and grouped. Multiple references to the same
-    # issue should be filtered.
-    self.assertEquals("(issues 123, 234, Chromium issue 345)",
-                      MakeChangeLogBugReference("Title\n\n"
-                                                "BUG=v8:234\n"
-                                                "  BUG\t= 345, \tv8:234,\n"
-                                                "BUG=v8:123\n"
-                                                "R=somebody\n"))
-    self.assertEquals("(Chromium issues 123, 234)",
-                      MakeChangeLogBugReference("Title\n\n"
-                                                "BUG=234,,chromium:123 \n"
-                                                "R=somebody\n"))
-    self.assertEquals("(Chromium issues 123, 234)",
-                      MakeChangeLogBugReference("Title\n\n"
-                                                "BUG=chromium:234, , 123\n"
-                                                "R=somebody\n"))
-    self.assertEquals("(issues 345, 456)",
-                      MakeChangeLogBugReference("Title\n\n"
-                                                "\t\tBUG=v8:345,v8:456\n"
-                                                "R=somebody\n"))
-    self.assertEquals("(issue 123, Chromium issues 345, 456)",
-                      MakeChangeLogBugReference("Title\n\n"
-                                                "BUG=chromium:456\n"
-                                                "BUG = none\n"
-                                                "R=somebody\n"
-                                                "BUG=456,v8:123, 345"))
-
-  # TODO(machenbach): These test don't make much sense when the formatting is
-  # done later.
-  def testMakeChangeLogBugReferenceLong(self):
-    # -----------------00--------10--------20--------30--------
-    self.assertEquals("(issues 234, 1234567890, 1234567"
-                      "8901234567890, Chromium issues 12345678,"
-                      " 123456789)",
-                      MakeChangeLogBugReference("BUG=v8:234\n"
-                                                "BUG=v8:1234567890\n"
-                                                "BUG=v8:12345678901234567890\n"
-                                                "BUG=123456789\n"
-                                                "BUG=12345678\n"))
-    # -----------------00--------10--------20--------30--------
-    self.assertEquals("(issues 234, 1234567890, 1234567"
-                      "8901234567890, Chromium issues"
-                      " 123456789, 1234567890)",
-                      MakeChangeLogBugReference("BUG=v8:234\n"
-                                                "BUG=v8:12345678901234567890\n"
-                                                "BUG=v8:1234567890\n"
-                                                "BUG=123456789\n"
-                                                "BUG=1234567890\n"))
-    # -----------------00--------10--------20--------30--------
-    self.assertEquals("(Chromium issues 234, 1234567890"
-                      ", 12345678901234567, "
-                      "1234567890123456789)",
-                      MakeChangeLogBugReference("BUG=234\n"
-                                                "BUG=12345678901234567\n"
-                                                "BUG=1234567890123456789\n"
-                                                "BUG=1234567890\n"))
+  def testCommand(self):
+    """Ensure json can decode the output of commands."""
+    json.dumps(Command('ls', pipe=True))
 
 
 def Cmd(*args, **kwargs):
@@ -278,6 +138,7 @@ class SimpleMock(object):
 
   def Call(self, name, *args, **kwargs):  # pragma: no cover
     self._index += 1
+
     try:
       expected_call = self._recipe[self._index]
     except IndexError:
@@ -380,9 +241,9 @@ class ScriptTest(unittest.TestCase):
                     config=TEST_CONFIG, side_effect_handler=self,
                     options=options)
 
-  def RunStep(self, script=PushToCandidates, step_class=Step, args=None):
+  def RunStep(self, script=CreateRelease, step_class=Step, args=None):
     """Convenience wrapper."""
-    args = args if args is not None else ["-m"]
+    args = args if args is not None else ["-m", "-a=author", "-r=reviewer", ]
     return script(TEST_CONFIG, self, self._state).RunSteps([step_class], args)
 
   def Call(self, fun, *args, **kwargs):
@@ -405,9 +266,6 @@ class ScriptTest(unittest.TestCase):
   def Sleep(self, seconds):
     pass
 
-  def GetDate(self):
-    return "1999-07-31"
-
   def GetUTCStamp(self):
     return "1000000"
 
@@ -443,7 +301,7 @@ class ScriptTest(unittest.TestCase):
   def testCommonPrepareDefault(self):
     self.Expect([
       Cmd("git status -s -uno", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git fetch", ""),
       Cmd("git branch", "  branch1\n* %s" % TEST_CONFIG["BRANCHNAME"]),
       RL("Y"),
@@ -455,7 +313,7 @@ class ScriptTest(unittest.TestCase):
   def testCommonPrepareNoConfirm(self):
     self.Expect([
       Cmd("git status -s -uno", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git fetch", ""),
       Cmd("git branch", "  branch1\n* %s" % TEST_CONFIG["BRANCHNAME"]),
       RL("n"),
@@ -466,7 +324,7 @@ class ScriptTest(unittest.TestCase):
   def testCommonPrepareDeleteBranchFailure(self):
     self.Expect([
       Cmd("git status -s -uno", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git fetch", ""),
       Cmd("git branch", "  branch1\n* %s" % TEST_CONFIG["BRANCHNAME"]),
       RL("Y"),
@@ -486,13 +344,13 @@ class ScriptTest(unittest.TestCase):
   def testTagTimeout(self):
     self.Expect([
       Cmd("git fetch", ""),
-      Cmd("git log -1 --format=%H --grep=\"Title\" origin/candidates", ""),
+      Cmd("git log -1 --format=%H --grep=\"Title\" origin/tag_name", ""),
       Cmd("git fetch", ""),
-      Cmd("git log -1 --format=%H --grep=\"Title\" origin/candidates", ""),
+      Cmd("git log -1 --format=%H --grep=\"Title\" origin/tag_name", ""),
       Cmd("git fetch", ""),
-      Cmd("git log -1 --format=%H --grep=\"Title\" origin/candidates", ""),
+      Cmd("git log -1 --format=%H --grep=\"Title\" origin/tag_name", ""),
       Cmd("git fetch", ""),
-      Cmd("git log -1 --format=%H --grep=\"Title\" origin/candidates", ""),
+      Cmd("git log -1 --format=%H --grep=\"Title\" origin/tag_name", ""),
     ])
     args = ["--branch", "candidates", "ab12345"]
     self._state["version"] = "tag_name"
@@ -530,84 +388,6 @@ class ScriptTest(unittest.TestCase):
                           r"\g<space>3",
                           "//\n#define V8_BUILD_NUMBER  321\n"))
 
-  def testPreparePushRevision(self):
-    # Tests the default push hash used when the --revision option is not set.
-    self.Expect([
-      Cmd("git log -1 --format=%H HEAD", "push_hash")
-    ])
-
-    self.RunStep(PushToCandidates, PreparePushRevision)
-    self.assertEquals("push_hash", self._state["push_hash"])
-
-  def testPrepareChangeLog(self):
-    self.WriteFakeVersionFile()
-    TEST_CONFIG["CHANGELOG_ENTRY_FILE"] = self.MakeEmptyTempFile()
-
-    self.Expect([
-      Cmd("git log --format=%H 1234..push_hash", "rev1\nrev2\nrev3\nrev4"),
-      Cmd("git log -1 --format=%s rev1", "Title text 1"),
-      Cmd("git log -1 --format=%B rev1", "Title\n\nBUG=\nLOG=y\n"),
-      Cmd("git log -1 --format=%an rev1", "author1@chromium.org"),
-      Cmd("git log -1 --format=%s rev2", "Title text 2."),
-      Cmd("git log -1 --format=%B rev2", "Title\n\nBUG=123\nLOG= \n"),
-      Cmd("git log -1 --format=%an rev2", "author2@chromium.org"),
-      Cmd("git log -1 --format=%s rev3", "Title text 3"),
-      Cmd("git log -1 --format=%B rev3", "Title\n\nBUG=321\nLOG=true\n"),
-      Cmd("git log -1 --format=%an rev3", "author3@chromium.org"),
-      Cmd("git log -1 --format=%s rev4", "Title text 4"),
-      Cmd("git log -1 --format=%B rev4", "Title\n\nBUG=456\nLOG=N"),
-      Cmd("git log -1 --format=%an rev4", "author4@chromium.org"),
-    ])
-
-    self._state["last_push_master"] = "1234"
-    self._state["push_hash"] = "push_hash"
-    self._state["version"] = "3.22.5"
-    self.RunStep(PushToCandidates, PrepareChangeLog)
-
-    actual_cl = FileToText(TEST_CONFIG["CHANGELOG_ENTRY_FILE"])
-
-    expected_cl = """1999-07-31: Version 3.22.5
-
-        Title text 1.
-
-        Title text 3 (Chromium issue 321).
-
-        Performance and stability improvements on all platforms.
-#
-# The change log above is auto-generated. Please review if all relevant
-# commit messages from the list below are included.
-# All lines starting with # will be stripped.
-#
-#       Title text 1.
-#       (author1@chromium.org)
-#
-#       Title text 2 (Chromium issue 123).
-#       (author2@chromium.org)
-#
-#       Title text 3 (Chromium issue 321).
-#       (author3@chromium.org)
-#
-#       Title text 4 (Chromium issue 456).
-#       (author4@chromium.org)
-#
-#"""
-
-    self.assertEquals(expected_cl, actual_cl)
-
-  def testEditChangeLog(self):
-    TEST_CONFIG["CHANGELOG_ENTRY_FILE"] = self.MakeEmptyTempFile()
-    TextToFile("  New  \n\tLines  \n", TEST_CONFIG["CHANGELOG_ENTRY_FILE"])
-    os.environ["EDITOR"] = "vi"
-    self.Expect([
-      RL(""),  # Open editor.
-      Cmd("vi %s" % TEST_CONFIG["CHANGELOG_ENTRY_FILE"], ""),
-    ])
-
-    self.RunStep(PushToCandidates, EditChangeLog)
-
-    self.assertEquals("New\n        Lines",
-                      FileToText(TEST_CONFIG["CHANGELOG_ENTRY_FILE"]))
-
   TAGS = """
 4425.0
 0.0.0.0
@@ -616,74 +396,23 @@ class ScriptTest(unittest.TestCase):
 test_tag
 """
 
-  # Version as tag: 3.22.4.0. Version on master: 3.22.6.
+  # Version as tag: 3.22.4.0. Version on main: 3.22.6.
   # Make sure that the latest version is 3.22.6.0.
   def testIncrementVersion(self):
     self.Expect([
       Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
       Cmd("git tag", self.TAGS),
-      Cmd("git checkout -f origin/master -- include/v8-version.h",
+      Cmd("git checkout -f origin/main -- include/v8-version.h",
           "", cb=lambda: self.WriteFakeVersionFile(3, 22, 6)),
     ])
 
-    self.RunStep(PushToCandidates, IncrementVersion)
+    self.RunStep(CreateRelease, IncrementVersion)
 
     self.assertEquals("3", self._state["new_major"])
     self.assertEquals("22", self._state["new_minor"])
     self.assertEquals("7", self._state["new_build"])
     self.assertEquals("0", self._state["new_patch"])
 
-  def _TestSquashCommits(self, change_log, expected_msg):
-    TEST_CONFIG["CHANGELOG_ENTRY_FILE"] = self.MakeEmptyTempFile()
-    with open(TEST_CONFIG["CHANGELOG_ENTRY_FILE"], "w") as f:
-      f.write(change_log)
-
-    self.Expect([
-      Cmd("git diff origin/candidates hash1", "patch content"),
-    ])
-
-    self._state["push_hash"] = "hash1"
-    self._state["date"] = "1999-11-11"
-
-    self.RunStep(PushToCandidates, SquashCommits)
-    self.assertEquals(FileToText(TEST_CONFIG["COMMITMSG_FILE"]), expected_msg)
-
-    patch = FileToText(TEST_CONFIG["PATCH_FILE"])
-    self.assertTrue(re.search(r"patch content", patch))
-
-  def testSquashCommitsUnformatted(self):
-    change_log = """1999-11-11: Version 3.22.5
-
-        Log text 1.
-        Chromium issue 12345
-
-        Performance and stability improvements on all platforms.\n"""
-    commit_msg = """Version 3.22.5 (based on hash1)
-
-Log text 1. Chromium issue 12345
-
-Performance and stability improvements on all platforms."""
-    self._TestSquashCommits(change_log, commit_msg)
-
-  def testSquashCommitsFormatted(self):
-    change_log = """1999-11-11: Version 3.22.5
-
-        Long commit message that fills more than 80 characters (Chromium issue
-        12345).
-
-        Performance and stability improvements on all platforms.\n"""
-    commit_msg = """Version 3.22.5 (based on hash1)
-
-Long commit message that fills more than 80 characters (Chromium issue 12345).
-
-Performance and stability improvements on all platforms."""
-    self._TestSquashCommits(change_log, commit_msg)
-
-  def testSquashCommitsQuotationMarks(self):
-    change_log = """Line with "quotation marks".\n"""
-    commit_msg = """Line with "quotation marks"."""
-    self._TestSquashCommits(change_log, commit_msg)
-
   def testBootstrapper(self):
     work_dir = self.MakeEmptyTempDirectory()
     class FakeScript(ScriptsBase):
@@ -699,188 +428,13 @@ Performance and stability improvements on all platforms."""
     ])
     FakeScript(fake_config, self).Run(["--work-dir", work_dir])
 
-  def _PushToCandidates(self, force=False, manual=False):
-    TextToFile("", os.path.join(TEST_CONFIG["DEFAULT_CWD"], ".git"))
-
-    # The version file on master has build level 5, while the version
-    # file from candidates has build level 4.
-    self.WriteFakeVersionFile(build=5)
-
-    TEST_CONFIG["CHANGELOG_ENTRY_FILE"] = self.MakeEmptyTempFile()
-    master_change_log = "2014-03-17: Sentinel\n"
-    TextToFile(master_change_log,
-               os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-    os.environ["EDITOR"] = "vi"
-
-    commit_msg_squashed = """Version 3.22.5 (squashed - based on push_hash)
-
-Log text 1 (issue 321).
-
-Performance and stability improvements on all platforms."""
-
-    commit_msg = """Version 3.22.5 (based on push_hash)
-
-Log text 1 (issue 321).
-
-Performance and stability improvements on all platforms."""
-
-    def ResetChangeLog():
-      """On 'git co -b new_branch origin/candidates',
-      and 'git checkout -- ChangeLog',
-      the ChangLog will be reset to its content on candidates."""
-      candidates_change_log = """1999-04-05: Version 3.22.4
-
-        Performance and stability improvements on all platforms.\n"""
-      TextToFile(candidates_change_log,
-                 os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-
-    def ResetToCandidates():
-      ResetChangeLog()
-      self.WriteFakeVersionFile()
-
-    def CheckVersionCommit():
-      commit = FileToText(TEST_CONFIG["COMMITMSG_FILE"])
-      self.assertEquals(commit_msg, commit)
-      version = FileToText(
-          os.path.join(TEST_CONFIG["DEFAULT_CWD"], VERSION_FILE))
-      self.assertTrue(re.search(r"#define V8_MINOR_VERSION\s+22", version))
-      self.assertTrue(re.search(r"#define V8_BUILD_NUMBER\s+5", version))
-      self.assertFalse(re.search(r"#define V8_BUILD_NUMBER\s+6", version))
-      self.assertTrue(re.search(r"#define V8_PATCH_LEVEL\s+0", version))
-      self.assertTrue(
-          re.search(r"#define V8_IS_CANDIDATE_VERSION\s+0", version))
-
-      # Check that the change log on the candidates branch got correctly
-      # modified.
-      change_log = FileToText(
-          os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-      self.assertEquals(
-"""1999-07-31: Version 3.22.5
-
-        Log text 1 (issue 321).
-
-        Performance and stability improvements on all platforms.
-
-
-1999-04-05: Version 3.22.4
-
-        Performance and stability improvements on all platforms.\n""",
-          change_log)
-
-    force_flag = " -f" if not manual else ""
-    expectations = []
-    if not force:
-      expectations.append(Cmd("which vi", "/usr/bin/vi"))
-    expectations += [
-      Cmd("git status -s -uno", ""),
-      Cmd("git checkout -f origin/master", ""),
-      Cmd("git fetch", ""),
-      Cmd("git branch", "  branch1\n* branch2\n"),
-      Cmd("git branch", "  branch1\n* branch2\n"),
-      Cmd(("git new-branch %s --upstream origin/master" %
-           TEST_CONFIG["BRANCHNAME"]), ""),
-      Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
-      Cmd("git tag", self.TAGS),
-      Cmd("git checkout -f origin/master -- include/v8-version.h",
-          "", cb=self.WriteFakeVersionFile),
-      Cmd("git log -1 --format=%H 3.22.4", "release_hash\n"),
-      Cmd("git log -1 --format=%s release_hash",
-          "Version 3.22.4 (based on abc3)\n"),
-      Cmd("git log --format=%H abc3..push_hash", "rev1\n"),
-      Cmd("git log -1 --format=%s rev1", "Log text 1.\n"),
-      Cmd("git log -1 --format=%B rev1", "Text\nLOG=YES\nBUG=v8:321\nText\n"),
-      Cmd("git log -1 --format=%an rev1", "author1@chromium.org\n"),
-    ]
-    if manual:
-      expectations.append(RL(""))  # Open editor.
-    if not force:
-      expectations.append(
-          Cmd("vi %s" % TEST_CONFIG["CHANGELOG_ENTRY_FILE"], ""))
-    expectations += [
-      Cmd("git fetch", ""),
-      Cmd("git checkout -f origin/master", ""),
-      Cmd("git diff origin/candidates push_hash", "patch content\n"),
-      Cmd(("git new-branch %s --upstream origin/candidates" %
-           TEST_CONFIG["CANDIDATESBRANCH"]), "", cb=ResetToCandidates),
-      Cmd("git apply --index --reject \"%s\"" % TEST_CONFIG["PATCH_FILE"], ""),
-      Cmd("git checkout -f origin/candidates -- ChangeLog", "",
-          cb=ResetChangeLog),
-      Cmd("git checkout -f origin/candidates -- include/v8-version.h", "",
-          cb=self.WriteFakeVersionFile),
-      Cmd("git commit -am \"%s\"" % commit_msg_squashed, ""),
-    ]
-    if manual:
-      expectations.append(RL("Y"))  # Sanity check.
-    expectations += [
-      Cmd("git cl land -f --bypass-hooks", ""),
-      Cmd("git checkout -f master", ""),
-      Cmd("git fetch", ""),
-      Cmd("git branch -D %s" % TEST_CONFIG["CANDIDATESBRANCH"], ""),
-      Cmd(("git new-branch %s --upstream origin/candidates" %
-           TEST_CONFIG["CANDIDATESBRANCH"]), "", cb=ResetToCandidates),
-      Cmd("git commit -aF \"%s\"" % TEST_CONFIG["COMMITMSG_FILE"], "",
-          cb=CheckVersionCommit),
-      Cmd("git cl land -f --bypass-hooks", ""),
-      Cmd("git fetch", ""),
-      Cmd("git log -1 --format=%H --grep="
-          "\"Version 3.22.5 (based on push_hash)\""
-          " origin/candidates", "hsh_to_tag"),
-      Cmd("git tag 3.22.5 hsh_to_tag", ""),
-      Cmd("git push origin refs/tags/3.22.5:refs/tags/3.22.5", ""),
-      Cmd("git checkout -f origin/master", ""),
-      Cmd("git branch -D %s" % TEST_CONFIG["BRANCHNAME"], ""),
-      Cmd("git branch -D %s" % TEST_CONFIG["CANDIDATESBRANCH"], ""),
-    ]
-    self.Expect(expectations)
-
-    args = ["-a", "author@chromium.org", "--revision", "push_hash"]
-    if force: args.append("-f")
-    if manual: args.append("-m")
-    else: args += ["-r", "reviewer@chromium.org"]
-    PushToCandidates(TEST_CONFIG, self).Run(args)
-
-    cl = FileToText(os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-    self.assertTrue(re.search(r"^\d\d\d\d\-\d+\-\d+: Version 3\.22\.5", cl))
-    self.assertTrue(re.search(r"        Log text 1 \(issue 321\).", cl))
-    self.assertTrue(re.search(r"1999\-04\-05: Version 3\.22\.4", cl))
-
-    # Note: The version file is on build number 5 again in the end of this test
-    # since the git command that merges to master is mocked out.
-
-  def testPushToCandidatesManual(self):
-    self._PushToCandidates(manual=True)
-
-  def testPushToCandidatesSemiAutomatic(self):
-    self._PushToCandidates()
-
-  def testPushToCandidatesForced(self):
-    self._PushToCandidates(force=True)
-
   def testCreateRelease(self):
     TextToFile("", os.path.join(TEST_CONFIG["DEFAULT_CWD"], ".git"))
 
-    # The version file on master has build level 5.
+    # The version file on main has build level 5.
     self.WriteFakeVersionFile(build=5)
 
-    master_change_log = "2014-03-17: Sentinel\n"
-    TextToFile(master_change_log,
-               os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-
-    commit_msg = """Version 3.22.5
-
-Log text 1 (issue 321).
-
-Performance and stability improvements on all platforms.
-
-TBR=reviewer@chromium.org"""
-
-    def ResetChangeLog():
-      last_change_log = """1999-04-05: Version 3.22.4
-
-        Performance and stability improvements on all platforms.\n"""
-      TextToFile(last_change_log,
-                 os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-
+    commit_msg = """Version 3.22.5"""
 
     def CheckVersionCommit():
       commit = FileToText(TEST_CONFIG["COMMITMSG_FILE"])
@@ -894,59 +448,36 @@ TBR=reviewer@chromium.org"""
       self.assertTrue(
           re.search(r"#define V8_IS_CANDIDATE_VERSION\s+0", version))
 
-      # Check that the change log on the candidates branch got correctly
-      # modified.
-      change_log = FileToText(
-          os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-      self.assertEquals(
-"""1999-07-31: Version 3.22.5
-
-        Log text 1 (issue 321).
-
-        Performance and stability improvements on all platforms.
-
-
-1999-04-05: Version 3.22.4
-
-        Performance and stability improvements on all platforms.\n""",
-          change_log)
-
     expectations = [
       Cmd("git fetch origin +refs/heads/*:refs/heads/*", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", "", cb=self.WriteFakeWatchlistsFile),
       Cmd("git branch", ""),
       Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
       Cmd("git tag", self.TAGS),
-      Cmd("git checkout -f origin/master -- include/v8-version.h",
+      Cmd("git checkout -f origin/main -- include/v8-version.h",
           "", cb=self.WriteFakeVersionFile),
       Cmd("git log -1 --format=%H 3.22.4", "release_hash\n"),
       Cmd("git log -1 --format=%s release_hash", "Version 3.22.4\n"),
       Cmd("git log -1 --format=%H release_hash^", "abc3\n"),
       Cmd("git log --format=%H abc3..push_hash", "rev1\n"),
-      Cmd("git log -1 --format=%s rev1", "Log text 1.\n"),
-      Cmd("git log -1 --format=%B rev1", "Text\nLOG=YES\nBUG=v8:321\nText\n"),
-      Cmd("git log -1 --format=%an rev1", "author1@chromium.org\n"),
       Cmd("git push origin push_hash:refs/heads/3.22.5", ""),
-      Cmd("git reset --hard origin/master", ""),
+      Cmd("git reset --hard origin/main", ""),
       Cmd("git new-branch work-branch --upstream origin/3.22.5", ""),
-      Cmd("git checkout -f 3.22.4 -- ChangeLog", "", cb=ResetChangeLog),
       Cmd("git checkout -f 3.22.4 -- include/v8-version.h", "",
           cb=self.WriteFakeVersionFile),
-      Cmd("git checkout -f 3.22.4 -- WATCHLISTS", "",
-          cb=self.WriteFakeWatchlistsFile),
       Cmd("git commit -aF \"%s\"" % TEST_CONFIG["COMMITMSG_FILE"], "",
           cb=CheckVersionCommit),
       Cmd("git cl upload --send-mail "
-          "-f --bypass-hooks --no-autocc --message-file "
-          "\"%s\" --gerrit" % TEST_CONFIG["COMMITMSG_FILE"], ""),
+          "-f --set-bot-commit --bypass-hooks --no-autocc --message-file "
+          "\"%s\"" % TEST_CONFIG["COMMITMSG_FILE"], ""),
       Cmd("git cl land --bypass-hooks -f", ""),
       Cmd("git fetch", ""),
       Cmd("git log -1 --format=%H --grep="
           "\"Version 3.22.5\" origin/3.22.5", "hsh_to_tag"),
       Cmd("git tag 3.22.5 hsh_to_tag", ""),
       Cmd("git push origin refs/tags/3.22.5:refs/tags/3.22.5", ""),
-      Cmd("git checkout -f origin/master", ""),
-      Cmd("git branch", "* master\n  work-branch\n"),
+      Cmd("git checkout -f origin/main", ""),
+      Cmd("git branch", "* main\n  work-branch\n"),
       Cmd("git branch -D work-branch", ""),
       Cmd("git gc", ""),
     ]
@@ -957,13 +488,8 @@ TBR=reviewer@chromium.org"""
             "--revision", "push_hash"]
     CreateRelease(TEST_CONFIG, self).Run(args)
 
-    cl = FileToText(os.path.join(TEST_CONFIG["DEFAULT_CWD"], CHANGELOG_FILE))
-    self.assertTrue(re.search(r"^\d\d\d\d\-\d+\-\d+: Version 3\.22\.5", cl))
-    self.assertTrue(re.search(r"        Log text 1 \(issue 321\).", cl))
-    self.assertTrue(re.search(r"1999\-04\-05: Version 3\.22\.4", cl))
-
     # Note: The version file is on build number 5 again in the end of this test
-    # since the git command that merges to master is mocked out.
+    # since the git command that merges to main is mocked out.
 
     # Check for correct content of the WATCHLISTS file
 
@@ -995,10 +521,12 @@ git-svn-id: https://v8.googlecode.com/svn/branches/bleeding_edge@123456 123
 
 """
 
-  ROLL_COMMIT_MSG = """Update V8 to version 3.22.4.
+  ROLL_HASH = "1234567890123456789012345678901234567890"
+  HASH_ALT_1 = "9999999999999999999999999999999999999999"
+  ROLL_COMMIT_MSG = f"""Update V8 to version 3.22.4.
 
 Summary of changes available at:
-https://chromium.googlesource.com/v8/v8/+log/last_rol..roll_hsh
+https://chromium.googlesource.com/v8/v8/+log/last_rol..{ROLL_HASH[:8]}
 
 Please follow these instructions for assigning/CC'ing issues:
 https://v8.dev/docs/triage-issues
@@ -1013,7 +541,7 @@ CQ_INCLUDE_TRYBOTS=luci.chromium.try:mac_optional_gpu_tests_rel
 CQ_INCLUDE_TRYBOTS=luci.chromium.try:win_optional_gpu_tests_rel
 CQ_INCLUDE_TRYBOTS=luci.chromium.try:android_optional_gpu_tests_rel
 
-TBR=reviewer@chromium.org"""
+R=reviewer@chromium.org"""
 
   # Snippet from the original DEPS file.
   FAKE_DEPS = """
@@ -1031,18 +559,19 @@ deps = {
     TEST_CONFIG["CHROMIUM"] = self.MakeEmptyTempDirectory()
     json_output_file = os.path.join(TEST_CONFIG["CHROMIUM"], "out.json")
     TextToFile(self.FAKE_DEPS, os.path.join(TEST_CONFIG["CHROMIUM"], "DEPS"))
+    chrome_dir = TEST_CONFIG["CHROMIUM"]
     self.Expect([
       Cmd("git fetch origin", ""),
       Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
-      Cmd("git describe --tags last_roll_hsh", "3.22.4"),
-      Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
-      Cmd("git rev-list --max-age=395200 --tags",
-          "bad_tag\nroll_hsh\nhash_123"),
-      Cmd("git describe --tags bad_tag", ""),
-      Cmd("git describe --tags roll_hsh", "3.22.4"),
-      Cmd("git describe --tags hash_123", "3.22.3"),
-      Cmd("git describe --tags roll_hsh", "3.22.4"),
-      Cmd("git describe --tags hash_123", "3.22.3"),
+      Cmd("gclient getdep -r src/v8", "last_roll_hsh", cwd=chrome_dir),
+      Cmd("git tag --points-at last_roll_hsh", "3.22.4\n3.22.4-pgo"),
+      Cmd((
+          "git for-each-ref --count=80 --sort=-committerdate --format "
+          "'%(refname) %(objectname)' 'refs/tags/*-pgo'"
+      ), "\n".join([
+          f"refs/tags/3.22.4-pgo {self.ROLL_HASH}",
+          f"refs/tags/3.22.3-pgo {self.HASH_ALT_1}",
+      ])),
     ])
 
     result = auto_roll.AutoRoll(TEST_CONFIG, self).Run(
@@ -1050,8 +579,6 @@ deps = {
           "-c", TEST_CONFIG["CHROMIUM"],
           "--json-output", json_output_file])
     self.assertEquals(0, result)
-    json_output = json.loads(FileToText(json_output_file))
-    self.assertEquals("up_to_date", json_output["monitoring_state"])
 
 
   def testChromiumRoll(self):
@@ -1070,33 +597,34 @@ deps = {
     expectations = [
       Cmd("git fetch origin", ""),
       Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
-      Cmd("git describe --tags last_roll_hsh", "3.22.3.1"),
-      Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
-      Cmd("git rev-list --max-age=395200 --tags",
-          "bad_tag\nroll_hsh\nhash_123"),
-      Cmd("git describe --tags bad_tag", ""),
-      Cmd("git describe --tags roll_hsh", "3.22.4"),
-      Cmd("git describe --tags hash_123", "3.22.3"),
-      Cmd("git describe --tags roll_hsh", "3.22.4"),
-      Cmd("git log -1 --format=%s roll_hsh", "Version 3.22.4\n"),
-      Cmd("git describe --tags roll_hsh", "3.22.4"),
-      Cmd("git describe --tags last_roll_hsh", "3.22.2.1"),
+      Cmd("gclient getdep -r src/v8", "last_roll_hsh", cwd=chrome_dir),
+      Cmd("git tag --points-at last_roll_hsh", "3.22.3.1\n22.3.1-pgo"),
+      Cmd((
+          "git for-each-ref --count=80 --sort=-committerdate --format "
+          "'%(refname) %(objectname)' 'refs/tags/*-pgo'"
+      ), "\n".join([
+          f"refs/tags/3.22.4-pgo {self.ROLL_HASH}",
+          f"refs/tags/3.22.3-pgo {self.HASH_ALT_1}",
+      ])),
+      Cmd(f"git log -1 --format=%s {self.ROLL_HASH}", "Version 3.22.4\n"),
+      Cmd(f"git tag --points-at {self.ROLL_HASH}", "3.22.4\n3.22.4-pgo"),
+      Cmd("git tag --points-at last_roll_hsh", "3.22.2.1\n22.2.1-pgo"),
       Cmd("git status -s -uno", "", cwd=chrome_dir),
-      Cmd("git checkout -f master", "", cwd=chrome_dir),
+      Cmd("git checkout -f main", "", cwd=chrome_dir),
       Cmd("git branch", "", cwd=chrome_dir),
       Cmd("git pull", "", cwd=chrome_dir),
       Cmd("git fetch origin", ""),
       Cmd("git new-branch work-branch", "", cwd=chrome_dir),
-      Cmd("gclient setdep -r src/v8@roll_hsh", "", cb=WriteDeps,
+      Cmd(f"gclient setdep -r src/v8@{self.ROLL_HASH}", "", cb=WriteDeps,
           cwd=chrome_dir),
       Cmd(("git commit -am \"%s\" "
            "--author \"author@chromium.org <author@chromium.org>\"" %
            self.ROLL_COMMIT_MSG),
           "", cwd=chrome_dir),
       Cmd("git cl upload --send-mail -f "
-          "--cq-dry-run --bypass-hooks --gerrit", "",
+          "--cq-dry-run --set-bot-commit --bypass-hooks", "",
           cwd=chrome_dir),
-      Cmd("git checkout -f master", "", cwd=chrome_dir),
+      Cmd("git checkout -f main", "", cwd=chrome_dir),
       Cmd("git branch -D work-branch", "", cwd=chrome_dir),
     ]
     self.Expect(expectations)
@@ -1108,9 +636,6 @@ deps = {
     deps = FileToText(os.path.join(chrome_dir, "DEPS"))
     self.assertTrue(re.search("\"v8_revision\": \"22624\"", deps))
 
-    json_output = json.loads(FileToText(json_output_file))
-    self.assertEquals("success", json_output["monitoring_state"])
-
   def testCheckLastPushRecently(self):
     self.Expect([
       Cmd("git fetch origin +refs/tags/*:refs/tags/*", ""),
@@ -1175,7 +700,6 @@ Title1
 Revert "Something"
 
 BUG=123,234,345,456,567,v8:123
-LOG=N
 """
 
     def VerifyLand():
@@ -1191,21 +715,21 @@ LOG=N
 
     self.Expect([
       Cmd("git status -s -uno", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git fetch", ""),
       Cmd("git branch", "  branch1\n* branch2\n"),
       Cmd("git new-branch %s --upstream refs/remotes/origin/candidates" %
           TEST_CONFIG["BRANCHNAME"], ""),
       Cmd(("git log --format=%H --grep=\"Port ab12345\" "
-           "--reverse origin/master"),
+           "--reverse origin/main"),
           "ab45678\nab23456"),
       Cmd("git log -1 --format=%s ab45678", "Title1"),
       Cmd("git log -1 --format=%s ab23456", "Title2"),
       Cmd(("git log --format=%H --grep=\"Port ab23456\" "
-           "--reverse origin/master"),
+           "--reverse origin/main"),
           ""),
       Cmd(("git log --format=%H --grep=\"Port ab34567\" "
-           "--reverse origin/master"),
+           "--reverse origin/main"),
           "ab56789"),
       Cmd("git log -1 --format=%s ab56789", "Title3"),
       RL("Y"),  # Automatically add corresponding ports (ab34567, ab56789)?
@@ -1219,7 +743,7 @@ LOG=N
       Cmd("git log -1 --format=%s ab56789", "Revert \"Something\""),
       Cmd("git log -1 ab12345", "Title4\nBUG=123\nBUG=234"),
       Cmd("git log -1 ab23456", "Title2\n BUG = v8:123,345"),
-      Cmd("git log -1 ab34567", "Title3\nLOG=n\nBUG=567, 456"),
+      Cmd("git log -1 ab34567", "Title3\nBUG=567, 456"),
       Cmd("git log -1 ab45678", "Title1\nBUG="),
       Cmd("git log -1 ab56789", "Revert \"Something\"\nBUG=none"),
       Cmd("git log -1 -p ab12345", "patch4"),
@@ -1247,7 +771,7 @@ LOG=N
       Cmd("git commit -aF \"%s\"" % TEST_CONFIG["COMMITMSG_FILE"], ""),
       RL("reviewer@chromium.org"),  # V8 reviewer.
       Cmd("git cl upload --send-mail -r \"reviewer@chromium.org\" "
-          "--bypass-hooks --cc \"ulan@chromium.org\" --gerrit", ""),
+          "--bypass-hooks", ""),
       Cmd("git checkout -f %s" % TEST_CONFIG["BRANCHNAME"], ""),
       RL("LGTM"),  # Enter LGTM for V8 CL.
       Cmd("git cl presubmit", "Presubmit successfull\n"),
@@ -1265,7 +789,7 @@ LOG=N
           "hsh_to_tag"),
       Cmd("git tag 3.22.5.1 hsh_to_tag", ""),
       Cmd("git push origin refs/tags/3.22.5.1:refs/tags/3.22.5.1", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git branch -D %s" % TEST_CONFIG["BRANCHNAME"], ""),
     ])
 
@@ -1317,7 +841,6 @@ Merged: Revert \"Something\"
 Revision: ab56789
 
 BUG=123,234,345,456,567,v8:123
-LOG=N
 NOTRY=true
 NOPRESUBMIT=true
 NOTREECHECKS=true
@@ -1329,21 +852,21 @@ NOTREECHECKS=true
 
     self.Expect([
       Cmd("git status -s -uno", ""),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git fetch", ""),
       Cmd("git branch", "  branch1\n* branch2\n"),
       Cmd("git new-branch %s --upstream refs/remotes/origin/candidates" %
           TEST_CONFIG["BRANCHNAME"], ""),
       Cmd(("git log --format=%H --grep=\"^[Pp]ort ab12345\" "
-           "--reverse origin/master"),
+           "--reverse origin/main"),
           "ab45678\nab23456"),
       Cmd("git log -1 --format=%s ab45678", "Title1"),
       Cmd("git log -1 --format=%s ab23456", "Title2"),
       Cmd(("git log --format=%H --grep=\"^[Pp]ort ab23456\" "
-           "--reverse origin/master"),
+           "--reverse origin/main"),
           ""),
       Cmd(("git log --format=%H --grep=\"^[Pp]ort ab34567\" "
-           "--reverse origin/master"),
+           "--reverse origin/main"),
           "ab56789"),
       Cmd("git log -1 --format=%s ab56789", "Title3"),
       RL("Y"),  # Automatically add corresponding ports (ab34567, ab56789)?
@@ -1357,7 +880,7 @@ NOTREECHECKS=true
       Cmd("git log -1 --format=%s ab56789", "Revert \"Something\""),
       Cmd("git log -1 ab12345", "Title4\nBUG=123\nBUG=234"),
       Cmd("git log -1 ab23456", "Title2\n BUG = v8:123,345"),
-      Cmd("git log -1 ab34567", "Title3\nLOG=n\nBug: 567, 456,345"),
+      Cmd("git log -1 ab34567", "Title3\nBug: 567, 456,345"),
       Cmd("git log -1 ab45678", "Title1\nBug:"),
       Cmd("git log -1 ab56789", "Revert \"Something\"\nBUG=none"),
       Cmd("git log -1 -p ab12345", "patch4"),
@@ -1384,13 +907,13 @@ NOTREECHECKS=true
       Cmd("git commit -aF \"%s\"" % TEST_CONFIG["COMMITMSG_FILE"], ""),
       RL("reviewer@chromium.org"),  # V8 reviewer.
       Cmd("git cl upload --send-mail -r \"reviewer@chromium.org\" "
-          "--bypass-hooks --cc \"ulan@chromium.org\" --gerrit", ""),
+          "--bypass-hooks", ""),
       Cmd("git checkout -f %s" % TEST_CONFIG["BRANCHNAME"], ""),
       RL("LGTM"),  # Enter LGTM for V8 CL.
       Cmd("git cl presubmit", "Presubmit successfull\n"),
       Cmd("git cl land -f --bypass-hooks", "Closing issue\n",
           cb=VerifyLand),
-      Cmd("git checkout -f origin/master", ""),
+      Cmd("git checkout -f origin/main", ""),
       Cmd("git branch -D %s" % TEST_CONFIG["BRANCHNAME"], ""),
     ])
 
