@@ -22,30 +22,34 @@
 'use strict';
 
 const {
-  ArrayPrototypeMap,
-  ObjectCreate,
   ObjectDefineProperties,
   ObjectDefineProperty,
-  ReflectApply,
   Symbol,
 } = primordials;
 
 const cares = internalBinding('cares_wrap');
-const { toASCII } = require('internal/idna');
 const { isIP } = require('internal/net');
 const { customPromisifyArgs } = require('internal/util');
-const errors = require('internal/errors');
+const {
+  codes: {
+    ERR_INVALID_ARG_TYPE,
+    ERR_INVALID_ARG_VALUE,
+    ERR_MISSING_ARGS,
+  },
+  DNSException,
+} = require('internal/errors');
 const {
   bindDefaultResolver,
-  getDefaultResolver,
   setDefaultResolver,
-  Resolver,
   validateHints,
   emitInvalidHostnameWarning,
-  getDefaultVerbatim,
+  getDefaultResultOrder,
   setDefaultResultOrder,
   errorCodes: dnsErrorCodes,
 } = require('internal/dns/utils');
+const {
+  Resolver,
+} = require('internal/dns/callback_resolver');
 const {
   NODATA,
   FORMERR,
@@ -72,11 +76,6 @@ const {
   ADDRGETNETWORKPARAMS,
   CANCELLED,
 } = dnsErrorCodes;
-const {
-  ERR_INVALID_ARG_TYPE,
-  ERR_INVALID_ARG_VALUE,
-  ERR_MISSING_ARGS,
-} = errors.codes;
 const {
   validateBoolean,
   validateFunction,
@@ -89,12 +88,13 @@ const {
 const {
   GetAddrInfoReqWrap,
   GetNameInfoReqWrap,
-  QueryReqWrap,
+  DNS_ORDER_VERBATIM,
+  DNS_ORDER_IPV4_FIRST,
+  DNS_ORDER_IPV6_FIRST,
 } = cares;
 
 const kPerfHooksDnsLookupContext = Symbol('kPerfHooksDnsLookupContext');
 const kPerfHooksDnsLookupServiceContext = Symbol('kPerfHooksDnsLookupServiceContext');
-const kPerfHooksDnsLookupResolveContext = Symbol('kPerfHooksDnsLookupResolveContext');
 
 const {
   hasObserver,
@@ -102,24 +102,22 @@ const {
   stopPerf,
 } = require('internal/perf/observe');
 
-const dnsException = errors.dnsException;
-
 let promises = null; // Lazy loaded
 
 function onlookup(err, addresses) {
   if (err) {
-    return this.callback(dnsException(err, 'getaddrinfo', this.hostname));
+    return this.callback(new DNSException(err, 'getaddrinfo', this.hostname));
   }
   this.callback(null, addresses[0], this.family || isIP(addresses[0]));
   if (this[kPerfHooksDnsLookupContext] && hasObserver('dns')) {
-    stopPerf(this, kPerfHooksDnsLookupContext);
+    stopPerf(this, kPerfHooksDnsLookupContext, { detail: { addresses } });
   }
 }
 
 
 function onlookupall(err, addresses) {
   if (err) {
-    return this.callback(dnsException(err, 'getaddrinfo', this.hostname));
+    return this.callback(new DNSException(err, 'getaddrinfo', this.hostname));
   }
 
   const family = this.family;
@@ -127,13 +125,13 @@ function onlookupall(err, addresses) {
     const addr = addresses[i];
     addresses[i] = {
       address: addr,
-      family: family || isIP(addr)
+      family: family || isIP(addr),
     };
   }
 
   this.callback(null, addresses);
   if (this[kPerfHooksDnsLookupContext] && hasObserver('dns')) {
-    stopPerf(this, kPerfHooksDnsLookupContext);
+    stopPerf(this, kPerfHooksDnsLookupContext, { detail: { addresses } });
   }
 }
 
@@ -145,7 +143,7 @@ function lookup(hostname, options, callback) {
   let hints = 0;
   let family = 0;
   let all = false;
-  let verbatim = getDefaultVerbatim();
+  let dnsOrder = getDefaultResultOrder();
 
   // Parse arguments
   if (hostname) {
@@ -158,7 +156,7 @@ function lookup(hostname, options, callback) {
   } else if (typeof options === 'number') {
     validateFunction(callback, 'callback');
 
-    validateOneOf(options, 'family', validFamilies, true);
+    validateOneOf(options, 'family', validFamilies);
     family = options;
   } else if (options !== undefined && typeof options !== 'object') {
     validateFunction(arguments.length === 2 ? options : callback, 'callback');
@@ -180,7 +178,7 @@ function lookup(hostname, options, callback) {
           family = 6;
           break;
         default:
-          validateOneOf(options.family, 'options.family', validFamilies, true);
+          validateOneOf(options.family, 'options.family', validFamilies);
           family = options.family;
           break;
       }
@@ -191,7 +189,11 @@ function lookup(hostname, options, callback) {
     }
     if (options?.verbatim != null) {
       validateBoolean(options.verbatim, 'options.verbatim');
-      verbatim = options.verbatim;
+      dnsOrder = options.verbatim ? 'verbatim' : 'ipv4first';
+    }
+    if (options?.order != null) {
+      validateOneOf(options.order, 'options.order', ['ipv4first', 'ipv6first', 'verbatim']);
+      dnsOrder = options.dnsOrder;
     }
   }
 
@@ -222,11 +224,19 @@ function lookup(hostname, options, callback) {
   req.hostname = hostname;
   req.oncomplete = all ? onlookupall : onlookup;
 
+  let order = DNS_ORDER_VERBATIM;
+
+  if (dnsOrder === 'ipv4first') {
+    order = DNS_ORDER_IPV4_FIRST;
+  } else if (dnsOrder === 'ipv6first') {
+    order = DNS_ORDER_IPV6_FIRST;
+  }
+
   const err = cares.getaddrinfo(
-    req, toASCII(hostname), family, hints, verbatim
+    req, hostname, family, hints, order,
   );
   if (err) {
-    process.nextTick(callback, dnsException(err, 'getaddrinfo', hostname));
+    process.nextTick(callback, new DNSException(err, 'getaddrinfo', hostname));
     return {};
   }
   if (hasObserver('dns')) {
@@ -234,8 +244,10 @@ function lookup(hostname, options, callback) {
       hostname,
       family,
       hints,
-      verbatim,
+      verbatim: order === DNS_ORDER_VERBATIM,
+      order: dnsOrder,
     };
+
     startPerf(req, kPerfHooksDnsLookupContext, { type: 'dns', name: 'lookup', detail });
   }
   return req;
@@ -247,11 +259,11 @@ ObjectDefineProperty(lookup, customPromisifyArgs,
 
 function onlookupservice(err, hostname, service) {
   if (err)
-    return this.callback(dnsException(err, 'getnameinfo', this.hostname));
+    return this.callback(new DNSException(err, 'getnameinfo', this.hostname));
 
   this.callback(null, hostname, service);
   if (this[kPerfHooksDnsLookupServiceContext] && hasObserver('dns')) {
-    stopPerf(this, kPerfHooksDnsLookupServiceContext);
+    stopPerf(this, kPerfHooksDnsLookupServiceContext, { detail: { hostname, service } });
   }
 }
 
@@ -276,7 +288,7 @@ function lookupService(address, port, callback) {
   req.oncomplete = onlookupservice;
 
   const err = cares.getnameinfo(req, address, port);
-  if (err) throw dnsException(err, 'getnameinfo', address);
+  if (err) throw new DNSException(err, 'getnameinfo', address);
   if (hasObserver('dns')) {
     startPerf(req, kPerfHooksDnsLookupServiceContext, {
       type: 'dns',
@@ -293,91 +305,6 @@ function lookupService(address, port, callback) {
 ObjectDefineProperty(lookupService, customPromisifyArgs,
                      { __proto__: null, value: ['hostname', 'service'], enumerable: false });
 
-
-function onresolve(err, result, ttls) {
-  if (ttls && this.ttl)
-    result = ArrayPrototypeMap(
-      result, (address, index) => ({ address, ttl: ttls[index] }));
-
-  if (err)
-    this.callback(dnsException(err, this.bindingName, this.hostname));
-  else {
-    this.callback(null, result);
-    if (this[kPerfHooksDnsLookupResolveContext] && hasObserver('dns')) {
-      stopPerf(this, kPerfHooksDnsLookupResolveContext);
-    }
-  }
-}
-
-function resolver(bindingName) {
-  function query(name, /* options, */ callback) {
-    let options;
-    if (arguments.length > 2) {
-      options = callback;
-      callback = arguments[2];
-    }
-
-    validateString(name, 'name');
-    validateFunction(callback, 'callback');
-
-    const req = new QueryReqWrap();
-    req.bindingName = bindingName;
-    req.callback = callback;
-    req.hostname = name;
-    req.oncomplete = onresolve;
-    req.ttl = !!(options && options.ttl);
-    const err = this._handle[bindingName](req, toASCII(name));
-    if (err) throw dnsException(err, bindingName, name);
-    if (hasObserver('dns')) {
-      startPerf(req, kPerfHooksDnsLookupResolveContext, {
-        type: 'dns',
-        name: bindingName,
-        detail: {
-          host: name,
-          ttl: req.ttl,
-        },
-      });
-    }
-    return req;
-  }
-  ObjectDefineProperty(query, 'name', { __proto__: null, value: bindingName });
-  return query;
-}
-
-const resolveMap = ObjectCreate(null);
-Resolver.prototype.resolveAny = resolveMap.ANY = resolver('queryAny');
-Resolver.prototype.resolve4 = resolveMap.A = resolver('queryA');
-Resolver.prototype.resolve6 = resolveMap.AAAA = resolver('queryAaaa');
-Resolver.prototype.resolveCaa = resolveMap.CAA = resolver('queryCaa');
-Resolver.prototype.resolveCname = resolveMap.CNAME = resolver('queryCname');
-Resolver.prototype.resolveMx = resolveMap.MX = resolver('queryMx');
-Resolver.prototype.resolveNs = resolveMap.NS = resolver('queryNs');
-Resolver.prototype.resolveTxt = resolveMap.TXT = resolver('queryTxt');
-Resolver.prototype.resolveSrv = resolveMap.SRV = resolver('querySrv');
-Resolver.prototype.resolvePtr = resolveMap.PTR = resolver('queryPtr');
-Resolver.prototype.resolveNaptr = resolveMap.NAPTR = resolver('queryNaptr');
-Resolver.prototype.resolveSoa = resolveMap.SOA = resolver('querySoa');
-Resolver.prototype.reverse = resolver('getHostByAddr');
-
-Resolver.prototype.resolve = resolve;
-
-function resolve(hostname, rrtype, callback) {
-  let resolver;
-  if (typeof rrtype === 'string') {
-    resolver = resolveMap[rrtype];
-  } else if (typeof rrtype === 'function') {
-    resolver = resolveMap.A;
-    callback = rrtype;
-  } else {
-    throw new ERR_INVALID_ARG_TYPE('rrtype', 'string', rrtype);
-  }
-
-  if (typeof resolver === 'function') {
-    return ReflectApply(resolver, this, [hostname, callback]);
-  }
-  throw new ERR_INVALID_ARG_VALUE('rrtype', rrtype);
-}
-
 function defaultResolverSetServers(servers) {
   const resolver = new Resolver();
 
@@ -394,6 +321,7 @@ module.exports = {
   lookupService,
 
   Resolver,
+  getDefaultResultOrder,
   setDefaultResultOrder,
   setServers: defaultResolverSetServers,
 
@@ -429,7 +357,7 @@ module.exports = {
   CANCELLED,
 };
 
-bindDefaultResolver(module.exports, getDefaultResolver());
+bindDefaultResolver(module.exports, Resolver.prototype);
 
 ObjectDefineProperties(module.exports, {
   promises: {
@@ -441,6 +369,6 @@ ObjectDefineProperties(module.exports, {
         promises = require('internal/dns/promises');
       }
       return promises;
-    }
-  }
+    },
+  },
 });
