@@ -23,7 +23,8 @@
 
 const {
   ArrayPrototypeJoin,
-  ArrayPrototypeShift,
+  ArrayPrototypePop,
+  ArrayPrototypePush,
   ArrayPrototypeSlice,
   ArrayPrototypeSplice,
   ArrayPrototypeUnshift,
@@ -32,8 +33,7 @@ const {
   ErrorCaptureStackTrace,
   FunctionPrototypeBind,
   FunctionPrototypeCall,
-  NumberIsNaN,
-  ObjectCreate,
+  NumberMAX_SAFE_INTEGER,
   ObjectDefineProperty,
   ObjectDefineProperties,
   ObjectGetPrototypeOf,
@@ -51,7 +51,7 @@ const {
 } = primordials;
 const kRejection = SymbolFor('nodejs.rejection');
 
-const { kEmptyObject } = require('internal/util');
+const { SymbolDispose, kEmptyObject } = require('internal/util');
 
 const {
   inspect,
@@ -59,6 +59,9 @@ const {
 } = require('internal/util/inspect');
 
 let spliceOne;
+let FixedQueue;
+let kFirstEventParam;
+let kResistStopPropagation;
 
 const {
   AbortError,
@@ -66,24 +69,28 @@ const {
   codes: {
     ERR_INVALID_ARG_TYPE,
     ERR_INVALID_THIS,
-    ERR_OUT_OF_RANGE,
-    ERR_UNHANDLED_ERROR
+    ERR_UNHANDLED_ERROR,
   },
   genericNodeError,
 } = require('internal/errors');
 
 const {
+  validateInteger,
   validateAbortSignal,
   validateBoolean,
   validateFunction,
+  validateNumber,
   validateString,
 } = require('internal/validators');
+const { addAbortListener } = require('internal/events/abort_listener');
 
 const kCapture = Symbol('kCapture');
 const kErrorMonitor = Symbol('events.errorMonitor');
+const kShapeMode = Symbol('shapeMode');
 const kMaxEventTargetListeners = Symbol('events.maxEventTargetListeners');
 const kMaxEventTargetListenersWarned =
   Symbol('events.maxEventTargetListenersWarned');
+const kWatermarkData = SymbolFor('nodejs.watermarkData');
 
 let EventEmitterAsyncResource;
 // The EventEmitterAsyncResource has to be initialized lazily because event.js
@@ -95,7 +102,7 @@ let EventEmitterAsyncResource;
 function lazyEventEmitterAsyncResource() {
   if (EventEmitterAsyncResource === undefined) {
     const {
-      AsyncResource
+      AsyncResource,
     } = require('async_hooks');
 
     const kEventEmitter = Symbol('kEventEmitter');
@@ -213,9 +220,11 @@ function EventEmitter(opts) {
   EventEmitter.init.call(this, opts);
 }
 module.exports = EventEmitter;
+module.exports.addAbortListener = addAbortListener;
 module.exports.once = once;
 module.exports.on = on;
 module.exports.getEventListeners = getEventListeners;
+module.exports.getMaxListeners = getMaxListeners;
 // Backwards-compat with node 0.10.x
 EventEmitter.EventEmitter = EventEmitter;
 
@@ -232,7 +241,7 @@ ObjectDefineProperty(EventEmitter, 'captureRejections', {
 
     EventEmitter.prototype[kCapture] = value;
   },
-  enumerable: true
+  enumerable: true,
 });
 
 ObjectDefineProperty(EventEmitter, 'EventEmitterAsyncResource', {
@@ -250,7 +259,7 @@ ObjectDefineProperty(EventEmitter.prototype, kCapture, {
   __proto__: null,
   value: false,
   writable: true,
-  enumerable: false
+  enumerable: false,
 });
 
 EventEmitter.prototype._events = undefined;
@@ -273,13 +282,9 @@ ObjectDefineProperty(EventEmitter, 'defaultMaxListeners', {
     return defaultMaxListeners;
   },
   set: function(arg) {
-    if (typeof arg !== 'number' || arg < 0 || NumberIsNaN(arg)) {
-      throw new ERR_OUT_OF_RANGE('defaultMaxListeners',
-                                 'a non-negative number',
-                                 arg);
-    }
+    validateNumber(arg, 'defaultMaxListeners', 0);
     defaultMaxListeners = arg;
-  }
+  },
 });
 
 ObjectDefineProperties(EventEmitter, {
@@ -296,7 +301,7 @@ ObjectDefineProperties(EventEmitter, {
     enumerable: false,
     configurable: false,
     writable: false,
-  }
+  },
 });
 
 /**
@@ -307,8 +312,7 @@ ObjectDefineProperties(EventEmitter, {
  */
 EventEmitter.setMaxListeners =
   function(n = defaultMaxListeners, ...eventTargets) {
-    if (typeof n !== 'number' || n < 0 || NumberIsNaN(n))
-      throw new ERR_OUT_OF_RANGE('n', 'a non-negative number', n);
+    validateNumber(n, 'setMaxListeners', 0);
     if (eventTargets.length === 0) {
       defaultMaxListeners = n;
     } else {
@@ -338,8 +342,11 @@ EventEmitter.init = function(opts) {
 
   if (this._events === undefined ||
       this._events === ObjectGetPrototypeOf(this)._events) {
-    this._events = ObjectCreate(null);
+    this._events = { __proto__: null };
     this._eventsCount = 0;
+    this[kShapeMode] = false;
+  } else {
+    this[kShapeMode] = true;
   }
 
   this._maxListeners = this._maxListeners || undefined;
@@ -404,9 +411,7 @@ function emitUnhandledRejectionOrErr(ee, err, type, args) {
  * @returns {EventEmitter}
  */
 EventEmitter.prototype.setMaxListeners = function setMaxListeners(n) {
-  if (typeof n !== 'number' || n < 0 || NumberIsNaN(n)) {
-    throw new ERR_OUT_OF_RANGE('n', 'a non-negative number', n);
-  }
+  validateNumber(n, 'setMaxListeners', 0);
   this._maxListeners = n;
   return this;
 };
@@ -480,7 +485,7 @@ EventEmitter.prototype.emit = function emit(type, ...args) {
         ObjectDefineProperty(er, kEnhanceStackBeforeInspector, {
           __proto__: null,
           value: FunctionPrototypeBind(enhanceStackTrace, this, er, capture),
-          configurable: true
+          configurable: true,
         });
       } catch {
         // Continue regardless of error.
@@ -547,7 +552,7 @@ function _addListener(target, type, listener, prepend) {
 
   events = target._events;
   if (events === undefined) {
-    events = target._events = ObjectCreate(null);
+    events = target._events = { __proto__: null };
     target._eventsCount = 0;
   } else {
     // To avoid recursion in the case that type === "newListener"! Before
@@ -586,7 +591,7 @@ function _addListener(target, type, listener, prepend) {
       // No error code for this since it is a Warning
       const w = genericNodeError(
         `Possible EventEmitter memory leak detected. ${existing.length} ${String(type)} listeners ` +
-        `added to ${inspect(target, { depth: -1 })}. Use emitter.setMaxListeners() to increase limit`,
+        `added to ${inspect(target, { depth: -1 })}. MaxListeners is ${m}. Use emitter.setMaxListeners() to increase limit`,
         { name: 'MaxListenersExceededWarning', emitter: target, type: type, count: existing.length });
       process.emitWarning(w);
     }
@@ -684,9 +689,13 @@ EventEmitter.prototype.removeListener =
         return this;
 
       if (list === listener || list.listener === listener) {
-        if (--this._eventsCount === 0)
-          this._events = ObjectCreate(null);
-        else {
+        this._eventsCount -= 1;
+
+        if (this[kShapeMode]) {
+          events[type] = undefined;
+        } else if (this._eventsCount === 0) {
+          this._events = { __proto__: null };
+        } else {
           delete events[type];
           if (events.removeListener)
             this.emit('removeListener', type, list.listener || listener);
@@ -740,14 +749,15 @@ EventEmitter.prototype.removeAllListeners =
       // Not listening for removeListener, no need to emit
       if (events.removeListener === undefined) {
         if (arguments.length === 0) {
-          this._events = ObjectCreate(null);
+          this._events = { __proto__: null };
           this._eventsCount = 0;
         } else if (events[type] !== undefined) {
           if (--this._eventsCount === 0)
-            this._events = ObjectCreate(null);
+            this._events = { __proto__: null };
           else
             delete events[type];
         }
+        this[kShapeMode] = false;
         return this;
       }
 
@@ -758,8 +768,9 @@ EventEmitter.prototype.removeAllListeners =
           this.removeAllListeners(key);
         }
         this.removeAllListeners('removeListener');
-        this._events = ObjectCreate(null);
+        this._events = { __proto__: null };
         this._eventsCount = 0;
+        this[kShapeMode] = false;
         return this;
       }
 
@@ -835,17 +846,34 @@ EventEmitter.prototype.listenerCount = listenerCount;
  * Returns the number of listeners listening to event name
  * specified as `type`.
  * @param {string | symbol} type
+ * @param {Function} listener
  * @returns {number}
  */
-function listenerCount(type) {
+function listenerCount(type, listener) {
   const events = this._events;
 
   if (events !== undefined) {
     const evlistener = events[type];
 
     if (typeof evlistener === 'function') {
+      if (listener != null) {
+        return listener === evlistener || listener === evlistener.listener ? 1 : 0;
+      }
+
       return 1;
     } else if (evlistener !== undefined) {
+      if (listener != null) {
+        let matching = 0;
+
+        for (let i = 0, l = evlistener.length; i < l; i++) {
+          if (evlistener[i] === listener || evlistener[i].listener === listener) {
+            matching++;
+          }
+        }
+
+        return matching;
+      }
+
       return evlistener.length;
     }
   }
@@ -916,11 +944,28 @@ function getEventListeners(emitterOrTarget, type) {
                                  emitterOrTarget);
 }
 
+/**
+ * Returns the max listeners set.
+ * @param {EventEmitter | EventTarget} emitterOrTarget
+ * @returns {number}
+ */
+function getMaxListeners(emitterOrTarget) {
+  if (typeof emitterOrTarget?.getMaxListeners === 'function') {
+    return _getMaxListeners(emitterOrTarget);
+  } else if (emitterOrTarget?.[kMaxEventTargetListeners]) {
+    return emitterOrTarget[kMaxEventTargetListeners];
+  }
+
+  throw new ERR_INVALID_ARG_TYPE('emitter',
+                                 ['EventEmitter', 'EventTarget'],
+                                 emitterOrTarget);
+}
+
 /**
  * Creates a `Promise` that is fulfilled when the emitter
  * emits the given event.
  * @param {EventEmitter} emitter
- * @param {string} name
+ * @param {string | symbol} name
  * @param {{ signal: AbortSignal; }} [options]
  * @returns {Promise}
  */
@@ -946,7 +991,10 @@ async function once(emitter, name, options = kEmptyObject) {
       }
       resolve(args);
     };
-    eventTargetAgnosticAddListener(emitter, name, resolver, { once: true });
+
+    kResistStopPropagation ??= require('internal/event_target').kResistStopPropagation;
+    const opts = { __proto__: null, once: true, [kResistStopPropagation]: true };
+    eventTargetAgnosticAddListener(emitter, name, resolver, opts);
     if (name !== 'error' && typeof emitter.once === 'function') {
       // EventTarget does not have `error` event semantics like Node
       // EventEmitters, we listen to `error` events only on EventEmitters.
@@ -959,7 +1007,7 @@ async function once(emitter, name, options = kEmptyObject) {
     }
     if (signal != null) {
       eventTargetAgnosticAddListener(
-        signal, 'abort', abortListener, { once: true });
+        signal, 'abort', abortListener, { __proto__: null, once: true, [kResistStopPropagation]: true });
     }
   });
 }
@@ -999,25 +1047,46 @@ function eventTargetAgnosticAddListener(emitter, name, listener, flags) {
  * Returns an `AsyncIterator` that iterates `event` events.
  * @param {EventEmitter} emitter
  * @param {string | symbol} event
- * @param {{ signal: AbortSignal; }} [options]
+ * @param {{
+ *    signal: AbortSignal;
+ *    close?: string[];
+ *    highWaterMark?: number,
+ *    lowWaterMark?: number
+ *   }} [options]
  * @returns {AsyncIterator}
  */
-function on(emitter, event, options) {
-  const signal = options?.signal;
+function on(emitter, event, options = kEmptyObject) {
+  // Parameters validation
+  const signal = options.signal;
   validateAbortSignal(signal, 'options.signal');
   if (signal?.aborted)
     throw new AbortError(undefined, { cause: signal?.reason });
+  // Support both highWaterMark and highWatermark for backward compatibility
+  const highWatermark = options.highWaterMark ?? options.highWatermark ?? NumberMAX_SAFE_INTEGER;
+  validateInteger(highWatermark, 'options.highWaterMark', 1);
+  // Support both lowWaterMark and lowWatermark for backward compatibility
+  const lowWatermark = options.lowWaterMark ?? options.lowWatermark ?? 1;
+  validateInteger(lowWatermark, 'options.lowWaterMark', 1);
 
-  const unconsumedEvents = [];
-  const unconsumedPromises = [];
+  // Preparing controlling queues and variables
+  FixedQueue ??= require('internal/fixed_queue');
+  const unconsumedEvents = new FixedQueue();
+  const unconsumedPromises = new FixedQueue();
+  let paused = false;
   let error = null;
   let finished = false;
+  let size = 0;
 
   const iterator = ObjectSetPrototypeOf({
     next() {
       // First, we consume all unread events
-      const value = unconsumedEvents.shift();
-      if (value) {
+      if (size) {
+        const value = unconsumedEvents.shift();
+        size--;
+        if (paused && size < lowWatermark) {
+          emitter.resume();
+          paused = false;
+        }
         return PromiseResolve(createIterResult(value, false));
       }
 
@@ -1032,9 +1101,7 @@ function on(emitter, event, options) {
       }
 
       // If the iterator is finished, resolve to done
-      if (finished) {
-        return PromiseResolve(createIterResult(undefined, true));
-      }
+      if (finished) return closeHandler();
 
       // Wait until an event happens
       return new Promise(function(resolve, reject) {
@@ -1043,24 +1110,7 @@ function on(emitter, event, options) {
     },
 
     return() {
-      eventTargetAgnosticRemoveListener(emitter, event, eventHandler);
-      eventTargetAgnosticRemoveListener(emitter, 'error', errorHandler);
-
-      if (signal) {
-        eventTargetAgnosticRemoveListener(
-          signal,
-          'abort',
-          abortListener,
-          { once: true });
-      }
-
-      finished = true;
-
-      for (const promise of unconsumedPromises) {
-        promise.resolve(createIterResult(undefined, true));
-      }
-
-      return PromiseResolve(createIterResult(undefined, true));
+      return closeHandler();
     },
 
     throw(err) {
@@ -1068,55 +1118,106 @@ function on(emitter, event, options) {
         throw new ERR_INVALID_ARG_TYPE('EventEmitter.AsyncIterator',
                                        'Error', err);
       }
-      error = err;
-      eventTargetAgnosticRemoveListener(emitter, event, eventHandler);
-      eventTargetAgnosticRemoveListener(emitter, 'error', errorHandler);
+      errorHandler(err);
     },
-
     [SymbolAsyncIterator]() {
       return this;
-    }
+    },
+    [kWatermarkData]: {
+      /**
+       * The current queue size
+       */
+      get size() {
+        return size;
+      },
+      /**
+       * The low watermark. The emitter is resumed every time size is lower than it
+       */
+      get low() {
+        return lowWatermark;
+      },
+      /**
+       * The high watermark. The emitter is paused every time size is higher than it
+       */
+      get high() {
+        return highWatermark;
+      },
+      /**
+       * It checks whether the emitter is paused by the watermark controller or not
+       */
+      get isPaused() {
+        return paused;
+      },
+    },
   }, AsyncIteratorPrototype);
 
-  eventTargetAgnosticAddListener(emitter, event, eventHandler);
+  // Adding event handlers
+  const { addEventListener, removeAll } = listenersController();
+  kFirstEventParam ??= require('internal/events/symbols').kFirstEventParam;
+  addEventListener(emitter, event, options[kFirstEventParam] ? eventHandler : function(...args) {
+    return eventHandler(args);
+  });
   if (event !== 'error' && typeof emitter.on === 'function') {
-    emitter.on('error', errorHandler);
+    addEventListener(emitter, 'error', errorHandler);
+  }
+  const closeEvents = options?.close;
+  if (closeEvents?.length) {
+    for (let i = 0; i < closeEvents.length; i++) {
+      addEventListener(emitter, closeEvents[i], closeHandler);
+    }
   }
 
-  if (signal) {
-    eventTargetAgnosticAddListener(
-      signal,
-      'abort',
-      abortListener,
-      { once: true });
-  }
+  const abortListenerDisposable = signal ? addAbortListener(signal, abortListener) : null;
+
+  return iterator;
 
   function abortListener() {
     errorHandler(new AbortError(undefined, { cause: signal?.reason }));
   }
 
-  function eventHandler(...args) {
-    const promise = ArrayPrototypeShift(unconsumedPromises);
-    if (promise) {
-      promise.resolve(createIterResult(args, false));
-    } else {
-      unconsumedEvents.push(args);
-    }
+  function eventHandler(value) {
+    if (unconsumedPromises.isEmpty()) {
+      size++;
+      if (!paused && size > highWatermark) {
+        paused = true;
+        emitter.pause();
+      }
+      unconsumedEvents.push(value);
+    } else unconsumedPromises.shift().resolve(createIterResult(value, false));
   }
 
   function errorHandler(err) {
+    if (unconsumedPromises.isEmpty()) error = err;
+    else unconsumedPromises.shift().reject(err);
+
+    closeHandler();
+  }
+
+  function closeHandler() {
+    abortListenerDisposable?.[SymbolDispose]();
+    removeAll();
     finished = true;
-
-    const toError = ArrayPrototypeShift(unconsumedPromises);
-
-    if (toError) {
-      toError.reject(err);
-    } else {
-      // The next time we call next()
-      error = err;
+    const doneResult = createIterResult(undefined, true);
+    while (!unconsumedPromises.isEmpty()) {
+      unconsumedPromises.shift().resolve(doneResult);
     }
 
-    iterator.return();
+    return PromiseResolve(doneResult);
   }
-  return iterator;
+}
+
+function listenersController() {
+  const listeners = [];
+
+  return {
+    addEventListener(emitter, event, handler, flags) {
+      eventTargetAgnosticAddListener(emitter, event, handler, flags);
+      ArrayPrototypePush(listeners, [emitter, event, handler, flags]);
+    },
+    removeAll() {
+      while (listeners.length > 0) {
+        ReflectApply(eventTargetAgnosticRemoveListener, undefined, ArrayPrototypePop(listeners));
+      }
+    },
+  };
 }
