@@ -254,7 +254,7 @@ function exists(path, callback) {
   validateFunction(callback, 'cb');
 
   function suppressedCallback(err) {
-    callback(err ? false : true);
+    callback(!err);
   }
 
   try {
@@ -759,8 +759,8 @@ function readSync(fd, buffer, offsetOrOptions, length, position) {
  * @param {(
  *   err?: Error,
  *   bytesRead?: number,
- *   buffers?: ArrayBufferView[];
- *   ) => any} callback
+ *   buffers?: ArrayBufferView[]
+ * ) => any} callback
  * @returns {void}
  */
 function readv(fd, buffers, position, callback) {
@@ -813,9 +813,9 @@ function readvSync(fd, buffers, position) {
  * @param {number | null} [position]
  * @param {(
  *   err?: Error,
- *   bytesWritten?: number;
+ *   bytesWritten?: number,
  *   buffer?: Buffer | TypedArray | DataView
- *   ) => any} callback
+ * ) => any} callback
  * @returns {void}
  */
 function write(fd, buffer, offsetOrOptions, length, position, callback) {
@@ -891,6 +891,8 @@ ObjectDefineProperty(write, kCustomPromisifyArgsSymbol,
  *   length?: number;
  *   position?: number | null;
  *   }} [offsetOrOptions]
+ * @param {number} [length]
+ * @param {number} [position]
  * @returns {number}
  */
 function writeSync(fd, buffer, offsetOrOptions, length, position) {
@@ -1084,14 +1086,11 @@ function truncateSync(path, len) {
   }
   // Allow error to be thrown, but still close fd.
   const fd = fs.openSync(path, 'r+');
-  let ret;
-
   try {
-    ret = fs.ftruncateSync(fd, len);
+    fs.ftruncateSync(fd, len);
   } finally {
     fs.closeSync(fd);
   }
-  return ret;
 }
 
 /**
@@ -1450,8 +1449,8 @@ function readdirSyncRecursive(basePath, options) {
  *   }} [options]
  * @param {(
  *   err?: Error,
- *   files?: string[] | Buffer[] | Dirent[];
- *   ) => any} callback
+ *   files?: string[] | Buffer[] | Dirent[]
+ * ) => any} callback
  * @returns {void}
  */
 function readdir(path, options, callback) {
@@ -1555,7 +1554,8 @@ function lstat(path, options = { bigint: false }, callback) {
   callback = makeStatsCallback(callback);
   path = getValidatedPath(path);
   if (permission.isEnabled() && !permission.has('fs.read', path)) {
-    callback(new ERR_ACCESS_DENIED('Access to this API has been restricted', 'FileSystemRead', path));
+    const resource = BufferIsBuffer(path) ? BufferToString(path) : path;
+    callback(new ERR_ACCESS_DENIED('Access to this API has been restricted', 'FileSystemRead', resource));
     return;
   }
 
@@ -1635,7 +1635,8 @@ function fstatSync(fd, options = { bigint: false }) {
 function lstatSync(path, options = { bigint: false, throwIfNoEntry: true }) {
   path = getValidatedPath(path);
   if (permission.isEnabled() && !permission.has('fs.read', path)) {
-    throw new ERR_ACCESS_DENIED('Access to this API has been restricted', 'FileSystemRead', path);
+    const resource = BufferIsBuffer(path) ? BufferToString(path) : path;
+    throw new ERR_ACCESS_DENIED('Access to this API has been restricted', 'FileSystemRead', resource);
   }
   const stats = binding.lstat(
     pathModule.toNamespacedPath(path),
@@ -1957,13 +1958,11 @@ function lchmodSync(path, mode) {
 
   // Prefer to return the chmod error, if one occurs,
   // but still try to close, and report closing errors if they occur.
-  let ret;
   try {
-    ret = fs.fchmodSync(fd, mode);
+    fs.fchmodSync(fd, mode);
   } finally {
     fs.closeSync(fd);
   }
-  return ret;
 }
 
 /**
@@ -2838,7 +2837,7 @@ function realpath(p, options, callback) {
 
   // On windows, check that the root exists. On unix there is no need.
   if (isWindows && !knownHard.has(base)) {
-    fs.lstat(base, (err, stats) => {
+    fs.lstat(base, (err) => {
       if (err) return callback(err);
       knownHard.add(base);
       LOOP();
@@ -2999,7 +2998,7 @@ function mkdtempSync(prefix, options) {
  * @param {string | Buffer | URL} src
  * @param {string | Buffer | URL} dest
  * @param {number} [mode]
- * @param {() => any} callback
+ * @param {(err?: Error) => any} callback
  * @returns {void}
  */
 function copyFile(src, dest, mode, callback) {
@@ -3045,7 +3044,7 @@ function copyFileSync(src, dest, mode) {
  * @param {string | URL} src
  * @param {string | URL} dest
  * @param {object} [options]
- * @param {() => any} callback
+ * @param {(err?: Error) => any} callback
  * @returns {void}
  */
 function cp(src, dest, options, callback) {
