@@ -8,7 +8,6 @@ const {
   ArrayIsArray,
   ArrayPrototypeForEach,
   ArrayPrototypePush,
-  ArrayPrototypeSome,
   ArrayPrototypeUnshift,
   Boolean,
   ErrorCaptureStackTrace,
@@ -68,7 +67,6 @@ const {
   CHAR_LOWERCASE_N: kTraceInstant,
   CHAR_UPPERCASE_C: kTraceCount,
 } = require('internal/constants');
-const { styleText } = require('util');
 const kCounts = Symbol('counts');
 
 const kTraceConsoleCategory = 'node,node.console';
@@ -274,7 +272,7 @@ ObjectDefineProperties(Console.prototype, {
   [kWriteToConsole]: {
     __proto__: null,
     ...consolePropAttributes,
-    value: function(streamSymbol, string, color = '') {
+    value: function(streamSymbol, string) {
       const ignoreErrors = this._ignoreErrors;
       const groupIndent = this[kGroupIndent];
 
@@ -289,11 +287,6 @@ ObjectDefineProperties(Console.prototype, {
         }
         string = groupIndent + string;
       }
-
-      if (color) {
-        string = styleText(color, string);
-      }
-
       string += '\n';
 
       if (ignoreErrors === false) return stream.write(string);
@@ -384,15 +377,12 @@ const consoleMethods = {
   log(...args) {
     this[kWriteToConsole](kUseStdout, this[kFormatForStdout](args));
   },
+
+
   warn(...args) {
-    const color = (shouldColorize(args) && 'yellow') || '';
-    this[kWriteToConsole](kUseStderr, this[kFormatForStderr](args), color);
+    this[kWriteToConsole](kUseStderr, this[kFormatForStderr](args));
   },
 
-  error(...args) {
-    const color = (shouldColorize(args) && 'red') || '';
-    this[kWriteToConsole](kUseStderr, this[kFormatForStderr](args), color);
-  },
 
   dir(object, options) {
     this[kWriteToConsole](kUseStdout, inspect(object, {
@@ -685,12 +675,6 @@ const iterKey = '(iteration index)';
 
 const isArray = (v) => ArrayIsArray(v) || isTypedArray(v) || isBuffer(v);
 
-// TODO: remove string type check once the styleText supports objects
-// Return true if all args are type string
-const shouldColorize = (args) => {
-  return lazyUtilColors().hasColors && !ArrayPrototypeSome(args, (arg) => typeof arg !== 'string');
-};
-
 function noop() {}
 
 for (const method of ReflectOwnKeys(consoleMethods))
@@ -699,6 +683,7 @@ for (const method of ReflectOwnKeys(consoleMethods))
 Console.prototype.debug = Console.prototype.log;
 Console.prototype.info = Console.prototype.log;
 Console.prototype.dirxml = Console.prototype.log;
+Console.prototype.error = Console.prototype.warn;
 Console.prototype.groupCollapsed = Console.prototype.group;
 
 function initializeGlobalConsole(globalConsole) {
