@@ -2,6 +2,7 @@
 
 const {
   ArrayFrom,
+  ArrayPrototypeReduce,
   Boolean,
   Error,
   FunctionPrototypeCall,
@@ -34,7 +35,13 @@ const {
     ERR_INVALID_THIS,
   },
 } = require('internal/errors');
-const { validateObject, validateString } = require('internal/validators');
+const {
+  validateAbortSignal,
+  validateObject,
+  validateString,
+  validateInternalField,
+  kValidateObjectAllowObjects,
+} = require('internal/validators');
 
 const {
   customInspectSymbol,
@@ -62,7 +69,7 @@ const kWeakHandler = Symbol('kWeak');
 const kResistStopPropagation = Symbol('kResistStopPropagation');
 
 const kHybridDispatch = SymbolFor('nodejs.internal.kHybridDispatch');
-const kRemoveWeakListenerHelper = Symbol('nodejs.internal.removeWeakListenerHelper');
+const kRemoveWeakListenerHelper = Symbol('kRemoveWeakListenerHelper');
 const kCreateEvent = Symbol('kCreateEvent');
 const kNewListener = Symbol('kNewListener');
 const kRemoveListener = Symbol('kRemoveListener');
@@ -108,14 +115,14 @@ class Event {
    *   composed?: boolean,
    * }} [options]
    */
-  constructor(type, options = kEmptyObject) {
+  constructor(type, options = undefined) {
     if (arguments.length === 0)
       throw new ERR_MISSING_ARGS('type');
-    validateObject(options, 'options');
-    const { bubbles, cancelable, composed } = options;
-    this.#cancelable = !!cancelable;
-    this.#bubbles = !!bubbles;
-    this.#composed = !!composed;
+    if (options != null)
+      validateObject(options, 'options');
+    this.#bubbles = !!options?.bubbles;
+    this.#cancelable = !!options?.cancelable;
+    this.#composed = !!options?.composed;
 
     this[kType] = `${type}`;
     if (options?.[kTrustEvent]) {
@@ -126,6 +133,23 @@ class Event {
     this[kIsBeingDispatched] = false;
   }
 
+  /**
+   * @param {string} type
+   * @param {boolean} [bubbles]
+   * @param {boolean} [cancelable]
+   */
+  initEvent(type, bubbles = false, cancelable = false) {
+    if (arguments.length === 0)
+      throw new ERR_MISSING_ARGS('type');
+
+    if (this[kIsBeingDispatched]) {
+      return;
+    }
+    this[kType] = `${type}`;
+    this.#bubbles = !!bubbles;
+    this.#cancelable = !!cancelable;
+  }
+
   [customInspectSymbol](depth, options) {
     if (!isEvent(this))
       throw new ERR_INVALID_THIS('Event');
@@ -148,6 +172,10 @@ class Event {
   stopImmediatePropagation() {
     if (!isEvent(this))
       throw new ERR_INVALID_THIS('Event');
+    // Spec mention "stopImmediatePropagation should set both "stop propagation"
+    // and "stop immediate propagation" flags"
+    // cf: from https://dom.spec.whatwg.org/#dom-event-stopimmediatepropagation
+    this.stopPropagation();
     this[kStop] = true;
   }
 
@@ -172,7 +200,7 @@ class Event {
   get currentTarget() {
     if (!isEvent(this))
       throw new ERR_INVALID_THIS('Event');
-    return this[kTarget];
+    return this[kIsBeingDispatched] ? this[kTarget] : null;
   }
 
   /**
@@ -285,7 +313,7 @@ class Event {
     if (!isEvent(this))
       throw new ERR_INVALID_THIS('Event');
     if (value) {
-      this.stopPropagation();
+      this.#propagationStopped = true;
     }
   }
 
@@ -294,11 +322,6 @@ class Event {
       throw new ERR_INVALID_THIS('Event');
     this.#propagationStopped = true;
   }
-
-  static NONE = 0;
-  static CAPTURING_PHASE = 1;
-  static AT_TARGET = 2;
-  static BUBBLING_PHASE = 3;
 }
 
 ObjectDefineProperties(
@@ -310,6 +333,7 @@ ObjectDefineProperties(
       configurable: true,
       value: 'Event',
     },
+    initEvent: kEnumerableProperty,
     stopImmediatePropagation: kEnumerableProperty,
     preventDefault: kEnumerableProperty,
     target: kEnumerableProperty,
@@ -333,6 +357,22 @@ ObjectDefineProperties(
     isTrusted: isTrustedDescriptor,
   });
 
+const staticProps = ['NONE', 'CAPTURING_PHASE', 'AT_TARGET', 'BUBBLING_PHASE'];
+
+ObjectDefineProperties(
+  Event,
+  ArrayPrototypeReduce(staticProps, (result, staticProp, index = 0) => {
+    result[staticProp] = {
+      __proto__: null,
+      writable: false,
+      configurable: false,
+      enumerable: true,
+      value: index,
+    };
+    return result;
+  }, {}),
+);
+
 function isCustomEvent(value) {
   return isEvent(value) && (value?.[kDetail] !== undefined);
 }
@@ -376,15 +416,6 @@ ObjectDefineProperties(CustomEvent.prototype, {
   detail: kEnumerableProperty,
 });
 
-class NodeCustomEvent extends Event {
-  constructor(type, options) {
-    super(type, options);
-    if (options?.detail) {
-      this.detail = options.detail;
-    }
-  }
-}
-
 // Weak listener cleanup
 // This has to be lazy for snapshots to work
 let weakListenersState = null;
@@ -512,6 +543,7 @@ function initEventTarget(self) {
   self[kEvents] = new SafeMap();
   self[kMaxEventTargetListeners] = EventEmitter.defaultMaxListeners;
   self[kMaxEventTargetListenersWarned] = false;
+  self[kHandlers] = new SafeMap();
 }
 
 class EventTarget {
@@ -533,7 +565,7 @@ class EventTarget {
       // eslint-disable-next-line no-restricted-syntax
       const w = new Error('Possible EventTarget memory leak detected. ' +
                           `${size} ${type} listeners ` +
-                          `added to ${inspect(this, { depth: -1 })}. Use ` +
+                          `added to ${inspect(this, { depth: -1 })}. MaxListeners is ${this[kMaxEventTargetListeners]}. Use ` +
                           'events.setMaxListeners() to increase limit');
       w.name = 'MaxListenersExceededWarning';
       w.target = this;
@@ -581,6 +613,8 @@ class EventTarget {
       resistStopPropagation,
     } = validateEventListenerOptions(options);
 
+    validateAbortSignal(signal, 'options.signal');
+
     if (!validateEventListener(listener)) {
       // The DOM silently allows passing undefined as a second argument
       // No error code for this since it is a Warning
@@ -802,7 +836,7 @@ class EventTarget {
   }
 
   [kCreateEvent](nodeValue, type) {
-    return new NodeCustomEvent(type, { detail: nodeValue });
+    return new CustomEvent(type, { detail: nodeValue });
   }
   [customInspectSymbol](depth, options) {
     if (!isEventTarget(this))
@@ -1020,9 +1054,7 @@ function validateEventListenerOptions(options) {
 
   if (options === null)
     return kEmptyObject;
-  validateObject(options, 'options', {
-    allowArray: true, allowFunction: true,
-  });
+  validateObject(options, 'options', kValidateObjectAllowObjects);
   return {
     once: Boolean(options.once),
     capture: Boolean(options.capture),
@@ -1075,36 +1107,48 @@ function makeEventHandler(handler) {
   return eventHandler;
 }
 
-function defineEventHandler(emitter, name) {
+function defineEventHandler(emitter, name, event = name) {
   // 8.1.5.1 Event handlers - basically `on[eventName]` attributes
-  ObjectDefineProperty(emitter, `on${name}`, {
+  const propName = `on${name}`;
+  function get() {
+    validateInternalField(this, kHandlers, 'EventTarget');
+    return this[kHandlers]?.get(event)?.handler ?? null;
+  }
+  ObjectDefineProperty(get, 'name', {
     __proto__: null,
-    get() {
-      return this[kHandlers]?.get(name)?.handler ?? null;
-    },
-    set(value) {
-      if (!this[kHandlers]) {
-        this[kHandlers] = new SafeMap();
+    value: `get ${propName}`,
+  });
+
+  function set(value) {
+    validateInternalField(this, kHandlers, 'EventTarget');
+    let wrappedHandler = this[kHandlers]?.get(event);
+    if (wrappedHandler) {
+      if (typeof wrappedHandler.handler === 'function') {
+        this[kEvents].get(event).size--;
+        const size = this[kEvents].get(event).size;
+        this[kRemoveListener](size, event, wrappedHandler.handler, false);
       }
-      let wrappedHandler = this[kHandlers]?.get(name);
-      if (wrappedHandler) {
-        if (typeof wrappedHandler.handler === 'function') {
-          this[kEvents].get(name).size--;
-          const size = this[kEvents].get(name).size;
-          this[kRemoveListener](size, name, wrappedHandler.handler, false);
-        }
-        wrappedHandler.handler = value;
-        if (typeof wrappedHandler.handler === 'function') {
-          this[kEvents].get(name).size++;
-          const size = this[kEvents].get(name).size;
-          this[kNewListener](size, name, value, false, false, false, false);
-        }
-      } else {
-        wrappedHandler = makeEventHandler(value);
-        this.addEventListener(name, wrappedHandler);
+      wrappedHandler.handler = value;
+      if (typeof wrappedHandler.handler === 'function') {
+        this[kEvents].get(event).size++;
+        const size = this[kEvents].get(event).size;
+        this[kNewListener](size, event, value, false, false, false, false);
       }
-      this[kHandlers].set(name, wrappedHandler);
-    },
+    } else {
+      wrappedHandler = makeEventHandler(value);
+      this.addEventListener(event, wrappedHandler);
+    }
+    this[kHandlers].set(event, wrappedHandler);
+  }
+  ObjectDefineProperty(set, 'name', {
+    __proto__: null,
+    value: `set ${propName}`,
+  });
+
+  ObjectDefineProperty(emitter, propName, {
+    __proto__: null,
+    get,
+    set,
     configurable: true,
     enumerable: true,
   });
