@@ -3,15 +3,16 @@
 // Stdin is not a TTY, we will read it and execute it.
 
 const {
-  prepareMainThreadExecution
-} = require('internal/bootstrap/pre_execution');
+  prepareMainThreadExecution,
+  markBootstrapComplete,
+} = require('internal/process/pre_execution');
 
 const { getOptionValue } = require('internal/options');
 
 const {
-  evalModule,
+  evalModuleEntryPoint,
   evalScript,
-  readStdin
+  readStdin,
 } = require('internal/process/execution');
 
 prepareMainThreadExecution();
@@ -23,11 +24,15 @@ readStdin((code) => {
   process._eval = code;
 
   const print = getOptionValue('--print');
-  if (getOptionValue('--input-type') === 'module')
-    evalModule(code, print);
-  else
+  const shouldLoadESM = getOptionValue('--import').length > 0;
+  if (getOptionValue('--input-type') === 'module' ||
+    (getOptionValue('--experimental-default-type') === 'module' && getOptionValue('--input-type') !== 'commonjs')) {
+    evalModuleEntryPoint(code, print);
+  } else {
     evalScript('[stdin]',
                code,
                getOptionValue('--inspect-brk'),
-               print);
+               print,
+               shouldLoadESM);
+  }
 });
