@@ -8,34 +8,24 @@ const {
   ArrayPrototypePushApply,
   ArrayPrototypeSplice,
   ObjectDefineProperty,
-  PromisePrototypeCatch,
-  globalThis: { Atomics },
+  PromisePrototypeThen,
+  RegExpPrototypeExec,
+  SafeWeakMap,
+  globalThis: {
+    Atomics,
+    SharedArrayBuffer,
+  },
 } = primordials;
 
 const {
-  patchProcessObject,
-  setupCoverageHooks,
-  setupInspectorHooks,
-  setupWarningHandler,
-  setupFetch,
-  setupWebCrypto,
-  setupCustomEvent,
-  setupDebugEnv,
-  setupPerfHooks,
-  initializeDeprecations,
-  initializeWASI,
-  initializeCJSLoader,
-  initializeESMLoader,
-  initializeFrozenIntrinsics,
-  initializeReport,
-  initializeSourceMapsHandlers,
-  loadPreloadModules,
-  setupTraceCategoryState
-} = require('internal/bootstrap/pre_execution');
+  prepareWorkerThreadExecution,
+  setupUserModules,
+  markBootstrapComplete,
+} = require('internal/process/pre_execution');
 
 const {
   threadId,
-  getEnvMessagePort
+  getEnvMessagePort,
 } = internalBinding('worker');
 
 const workerIo = require('internal/worker/io');
@@ -51,35 +41,22 @@ const {
     STDIO_PAYLOAD,
     STDIO_WANTS_MORE_DATA,
   },
-  kStdioWantsMoreDataCallback
+  kStdioWantsMoreDataCallback,
 } = workerIo;
 
 const {
-  onGlobalUncaughtException
+  onGlobalUncaughtException,
 } = require('internal/process/execution');
 
-const publicWorker = require('worker_threads');
 let debug = require('internal/util/debuglog').debuglog('worker', (fn) => {
   debug = fn;
 });
 
 const assert = require('internal/assert');
+const { getOptionValue } = require('internal/options');
+const { exitCodes: { kGenericUserError } } = internalBinding('errors');
 
-patchProcessObject();
-setupInspectorHooks();
-setupDebugEnv();
-
-setupWarningHandler();
-setupFetch();
-setupWebCrypto();
-setupCustomEvent();
-initializeSourceMapsHandlers();
-
-// Since worker threads cannot switch cwd, we do not need to
-// overwrite the process.env.NODE_V8_COVERAGE variable.
-if (process.env.NODE_V8_COVERAGE) {
-  setupCoverageHooks(process.env.NODE_V8_COVERAGE);
-}
+prepareWorkerThreadExecution();
 
 debug(`[${threadId}] is setting up worker child environment`);
 
@@ -94,13 +71,13 @@ if (process.env.NODE_CHANNEL_FD) {
   ObjectDefineProperty(process, 'channel', {
     __proto__: null,
     enumerable: false,
-    get: workerThreadSetup.unavailable('process.channel')
+    get: workerThreadSetup.unavailable('process.channel'),
   });
 
   ObjectDefineProperty(process, 'connected', {
     __proto__: null,
     enumerable: false,
-    get: workerThreadSetup.unavailable('process.connected')
+    get: workerThreadSetup.unavailable('process.connected'),
   });
 
   process.send = workerThreadSetup.unavailable('process.send()');
@@ -114,55 +91,54 @@ port.on('message', (message) => {
     const {
       argv,
       cwdCounter,
-      filename,
       doEval,
-      workerData,
       environmentData,
-      publicPort,
+      filename,
+      hasStdin,
       manifestSrc,
       manifestURL,
-      hasStdin
+      publicPort,
+      workerData,
     } = message;
 
-    setupTraceCategoryState();
-    setupPerfHooks();
-    initializeReport();
-    if (manifestSrc) {
-      require('internal/process/policy').setup(manifestSrc, manifestURL);
-    }
-    initializeDeprecations();
-    initializeWASI();
-    initializeCJSLoader();
-    initializeESMLoader();
+    if (doEval !== 'internal') {
+      if (argv !== undefined) {
+        ArrayPrototypePushApply(process.argv, argv);
+      }
 
-    if (argv !== undefined) {
-      ArrayPrototypePushApply(process.argv, argv);
+      const publicWorker = require('worker_threads');
+      publicWorker.parentPort = publicPort;
+      publicWorker.workerData = workerData;
     }
-    publicWorker.parentPort = publicPort;
-    publicWorker.workerData = workerData;
 
     require('internal/worker').assignEnvironmentData(environmentData);
 
-    // The counter is only passed to the workers created by the main thread, not
-    // to workers created by other workers.
-    let cachedCwd = '';
-    let lastCounter = -1;
-    const originalCwd = process.cwd;
+    if (SharedArrayBuffer !== undefined && Atomics !== undefined) {
+      // The counter is only passed to the workers created by the main thread,
+      // not to workers created by other workers.
+      let cachedCwd = '';
+      let lastCounter = -1;
+      const originalCwd = process.cwd;
 
-    process.cwd = function() {
-      const currentCounter = Atomics.load(cwdCounter, 0);
-      if (currentCounter === lastCounter)
+      process.cwd = function() {
+        const currentCounter = Atomics.load(cwdCounter, 0);
+        if (currentCounter === lastCounter)
+          return cachedCwd;
+        lastCounter = currentCounter;
+        cachedCwd = originalCwd();
         return cachedCwd;
-      lastCounter = currentCounter;
-      cachedCwd = originalCwd();
-      return cachedCwd;
-    };
-    workerIo.sharedCwdCounter = cwdCounter;
+      };
+      workerIo.sharedCwdCounter = cwdCounter;
+    }
 
-    const CJSLoader = require('internal/modules/cjs/loader');
-    assert(!CJSLoader.hasLoadedAnyUserCJSModule);
-    loadPreloadModules();
-    initializeFrozenIntrinsics();
+    if (manifestSrc) {
+      require('internal/process/policy').setup(manifestSrc, manifestURL);
+    }
+    const isLoaderWorker =
+      doEval === 'internal' &&
+      filename === require('internal/modules/esm/utils').loaderWorkerId;
+    // Disable custom loaders in loader worker.
+    setupUserModules(isLoaderWorker);
 
     if (!hasStdin)
       process.stdin.push(null);
@@ -170,30 +146,48 @@ port.on('message', (message) => {
     debug(`[${threadId}] starts worker script ${filename} ` +
           `(eval = ${doEval}) at cwd = ${process.cwd()}`);
     port.postMessage({ type: UP_AND_RUNNING });
-    if (doEval === 'classic') {
-      const { evalScript } = require('internal/process/execution');
-      const name = '[worker eval]';
-      // This is necessary for CJS module compilation.
-      // TODO: pass this with something really internal.
-      ObjectDefineProperty(process, '_eval', {
-        __proto__: null,
-        configurable: true,
-        enumerable: true,
-        value: filename,
-      });
-      ArrayPrototypeSplice(process.argv, 1, 0, name);
-      evalScript(name, filename);
-    } else if (doEval === 'module') {
-      const { evalModule } = require('internal/process/execution');
-      PromisePrototypeCatch(evalModule(filename), (e) => {
-        workerOnGlobalUncaughtException(e, true);
-      });
-    } else {
-      // script filename
-      // runMain here might be monkey-patched by users in --require.
-      // XXX: the monkey-patchability here should probably be deprecated.
-      ArrayPrototypeSplice(process.argv, 1, 0, filename);
-      CJSLoader.Module.runMain(filename);
+    switch (doEval) {
+      case 'internal': {
+        // Create this WeakMap in js-land because V8 has no C++ API for WeakMap.
+        internalBinding('module_wrap').callbackMap = new SafeWeakMap();
+        require(filename)(workerData, publicPort);
+        break;
+      }
+
+      case 'classic': if (getOptionValue('--input-type') !== 'module') {
+        const { evalScript } = require('internal/process/execution');
+        const name = '[worker eval]';
+        // This is necessary for CJS module compilation.
+        // TODO: pass this with something really internal.
+        ObjectDefineProperty(process, '_eval', {
+          __proto__: null,
+          configurable: true,
+          enumerable: true,
+          value: filename,
+        });
+        ArrayPrototypeSplice(process.argv, 1, 0, name);
+        evalScript(name, filename);
+        break;
+      }
+
+      // eslint-disable-next-line no-fallthrough
+      case 'module': {
+        const { evalModuleEntryPoint } = require('internal/process/execution');
+        PromisePrototypeThen(evalModuleEntryPoint(filename), undefined, (e) => {
+          workerOnGlobalUncaughtException(e, true);
+        });
+        break;
+      }
+
+      default: {
+        // script filename
+        // runMain here might be monkey-patched by users in --require.
+        // XXX: the monkey-patchability here should probably be deprecated.
+        ArrayPrototypeSplice(process.argv, 1, 0, filename);
+        const CJSLoader = require('internal/modules/cjs/loader');
+        CJSLoader.Module.runMain(filename);
+        break;
+      }
     }
   } else if (message.type === STDIO_PAYLOAD) {
     const { stream, chunks } = message;
@@ -203,7 +197,7 @@ port.on('message', (message) => {
   } else {
     assert(
       message.type === STDIO_WANTS_MORE_DATA,
-      `Unknown worker message type ${message.type}`
+      `Unknown worker message type ${message.type}`,
     );
     const { stream } = message;
     process[stream][kStdioWantsMoreDataCallback]();
@@ -229,7 +223,7 @@ function workerOnGlobalUncaughtException(error, fromPromise) {
   if (!process._exiting) {
     try {
       process._exiting = true;
-      process.exitCode = 1;
+      process.exitCode = kGenericUserError;
       if (!handlerThrew) {
         process.emit('exit', process.exitCode);
       }
@@ -249,7 +243,7 @@ function workerOnGlobalUncaughtException(error, fromPromise) {
   if (serialized)
     port.postMessage({
       type: ERROR_MESSAGE,
-      error: serialized
+      error: serialized,
     });
   else
     port.postMessage({ type: COULD_NOT_SERIALIZE_ERROR });
@@ -266,4 +260,7 @@ process._fatalException = workerOnGlobalUncaughtException;
 
 markBootstrapComplete();
 
+// Necessary to reset RegExp statics before user code runs.
+RegExpPrototypeExec(/^/, '');
+
 port.start();
