@@ -1,13 +1,13 @@
 'use strict';
 
 const {
-  ArrayPrototypeForEach,
   ArrayPrototypeMap,
   Boolean,
   JSONParse,
   ObjectGetPrototypeOf,
   ObjectPrototypeHasOwnProperty,
   ObjectKeys,
+  ReflectApply,
   SafeArrayIterator,
   SafeMap,
   SafeSet,
@@ -29,22 +29,32 @@ function lazyTypes() {
   return _TYPES = require('internal/util/types');
 }
 
-const { readFileSync } = require('fs');
-const { extname, isAbsolute } = require('path');
 const {
-  hasEsmSyntax,
+  containsModuleSyntax,
+  compileFunctionForCJSLoader,
+} = internalBinding('contextify');
+
+const { BuiltinModule } = require('internal/bootstrap/realm');
+const assert = require('internal/assert');
+const { readFileSync } = require('fs');
+const { dirname, extname, isAbsolute } = require('path');
+const {
   loadBuiltinModule,
   stripBOM,
+  urlToFilename,
 } = require('internal/modules/helpers');
 const {
+  kIsCachedByESMLoader,
   Module: CJSModule,
-  cjsParseCache,
+  kModuleSource,
+  kModuleExport,
+  kModuleExportNames,
 } = require('internal/modules/cjs/loader');
-const { fileURLToPath, URL } = require('internal/url');
+const { fileURLToPath, pathToFileURL, URL } = require('internal/url');
 let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
   debug = fn;
 });
-const { emitExperimentalWarning } = require('internal/util');
+const { emitExperimentalWarning, kEmptyObject, setOwnProperty } = require('internal/util');
 const {
   ERR_UNKNOWN_BUILTIN_MODULE,
   ERR_INVALID_RETURN_PROPERTY_VALUE,
@@ -52,9 +62,19 @@ const {
 const { maybeCacheSourceMap } = require('internal/source_map/source_map_cache');
 const moduleWrap = internalBinding('module_wrap');
 const { ModuleWrap } = moduleWrap;
-const asyncESM = require('internal/process/esm_loader');
 const { emitWarningSync } = require('internal/process/warning');
 
+// Lazy-loading to avoid circular dependencies.
+let getSourceSync;
+/**
+ * @param {Parameters<typeof import('./load').getSourceSync>[0]} url
+ * @returns {ReturnType<typeof import('./load').getSourceSync>}
+ */
+function getSource(url) {
+  getSourceSync ??= require('internal/modules/esm/load').getSourceSync;
+  return getSourceSync(url);
+}
+
 /** @type {import('deps/cjs-module-lexer/lexer.js').parse} */
 let cjsParse;
 /**
@@ -64,7 +84,7 @@ let cjsParse;
  */
 async function initCJSParse() {
   if (typeof WebAssembly === 'undefined') {
-    cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
+    initCJSParseSync();
   } else {
     const { parse, init } =
         require('internal/deps/cjs-module-lexer/dist/lexer');
@@ -72,11 +92,19 @@ async function initCJSParse() {
       await init();
       cjsParse = parse;
     } catch {
-      cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
+      initCJSParseSync();
     }
   }
 }
 
+function initCJSParseSync() {
+  // TODO(joyeecheung): implement a binding that directly compiles using
+  // v8::WasmModuleObject::Compile() synchronously.
+  if (cjsParse === undefined) {
+    cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
+  }
+}
+
 const translators = new SafeMap();
 exports.translators = translators;
 exports.enrichCJSError = enrichCJSError;
@@ -113,7 +141,7 @@ function assertBufferSource(body, allowString, hookName) {
  */
 function stringify(body) {
   if (typeof body === 'string') { return body; }
-  assertBufferSource(body, false, 'transformSource');
+  assertBufferSource(body, false, 'load');
   const { TextDecoder } = require('internal/encoding');
   DECODER = DECODER === null ? new TextDecoder() : DECODER;
   return DECODER.decode(body);
@@ -131,31 +159,13 @@ function errPath(url) {
   return url;
 }
 
-/**
- * Dynamically imports a module using the ESM loader.
- * @param {string} specifier - The module specifier to import.
- * @param {object} options - An object containing options for the import.
- * @param {string} options.url - The URL of the module requesting the import.
- * @param {Record<string, string>} [attributes] - An object containing attributes for the import.
- * @returns {Promise<import('internal/modules/esm/loader.js').ModuleExports>} The imported module.
- */
-async function importModuleDynamically(specifier, { url }, attributes) {
-  return asyncESM.esmLoader.import(specifier, url, attributes);
-}
-
 // Strategy for loading a standard JavaScript module.
-translators.set('module', async function moduleStrategy(url, source, isMain) {
+translators.set('module', function moduleStrategy(url, source, isMain) {
   assertBufferSource(source, true, 'load');
   source = stringify(source);
-  maybeCacheSourceMap(url, source);
   debug(`Translating StandardModule ${url}`);
-  const module = new ModuleWrap(url, undefined, source, 0, 0);
-  const { registerModule } = require('internal/modules/esm/utils');
-  registerModule(module, {
-    __proto__: null,
-    initializeImportMeta: (meta, wrap) => this.importMetaInitialize(meta, { url }),
-    importModuleDynamically,
-  });
+  const { compileSourceTextModule } = require('internal/modules/esm/utils');
+  const module = compileSourceTextModule(url, source, this);
   return module;
 });
 
@@ -163,11 +173,11 @@ translators.set('module', async function moduleStrategy(url, source, isMain) {
  * Provide a more informative error for CommonJS imports.
  * @param {Error | any} err
  * @param {string} [content] Content of the file, if known.
- * @param {string} [filename] Useful only if `content` is unknown.
+ * @param {string} [filename] The filename of the erroring module.
  */
 function enrichCJSError(err, content, filename) {
   if (err != null && ObjectGetPrototypeOf(err) === SyntaxErrorPrototype &&
-      hasEsmSyntax(content || readFileSync(filename, 'utf-8'))) {
+      containsModuleSyntax(content, filename)) {
     // Emit the warning synchronously because we are in the middle of handling
     // a SyntaxError that will throw and likely terminate the process before an
     // asynchronous warning would be emitted.
@@ -178,35 +188,111 @@ function enrichCJSError(err, content, filename) {
   }
 }
 
-// Strategy for loading a node-style CommonJS module
-const isWindows = process.platform === 'win32';
-translators.set('commonjs', async function commonjsStrategy(url, source,
-                                                            isMain) {
+/**
+ * Loads a CommonJS module via the ESM Loader sync CommonJS translator.
+ * This translator creates its own version of the `require` function passed into CommonJS modules.
+ * Any monkey patches applied to the CommonJS Loader will not affect this module.
+ * Any `require` calls in this module will load all children in the same way.
+ * @param {import('internal/modules/cjs/loader').Module} module - The module to load.
+ * @param {string} source - The source code of the module.
+ * @param {string} url - The URL of the module.
+ * @param {string} filename - The filename of the module.
+ */
+function loadCJSModule(module, source, url, filename) {
+  let compileResult;
+  try {
+    compileResult = compileFunctionForCJSLoader(source, filename);
+  } catch (err) {
+    enrichCJSError(err, source, filename);
+    throw err;
+  }
+  // Cache the source map for the cjs module if present.
+  if (compileResult.sourceMapURL) {
+    maybeCacheSourceMap(url, source, null, false, undefined, compileResult.sourceMapURL);
+  }
+
+  const compiledWrapper = compileResult.function;
+
+  const cascadedLoader = require('internal/modules/esm/loader').getOrInitializeCascadedLoader();
+  const __dirname = dirname(filename);
+  // eslint-disable-next-line func-name-matching,func-style
+  const requireFn = function require(specifier) {
+    let importAttributes = kEmptyObject;
+    if (!StringPrototypeStartsWith(specifier, 'node:') && !BuiltinModule.normalizeRequirableId(specifier)) {
+      // TODO: do not depend on the monkey-patchable CJS loader here.
+      const path = CJSModule._resolveFilename(specifier, module);
+      switch (extname(path)) {
+        case '.json':
+          importAttributes = { __proto__: null, type: 'json' };
+          break;
+        case '.node':
+          return CJSModule._load(specifier, module);
+        default:
+            // fall through
+      }
+      specifier = `${pathToFileURL(path)}`;
+    }
+    const job = cascadedLoader.getModuleJobSync(specifier, url, importAttributes);
+    job.runSync();
+    return cjsCache.get(job.url).exports;
+  };
+  setOwnProperty(requireFn, 'resolve', function resolve(specifier) {
+    if (!StringPrototypeStartsWith(specifier, 'node:')) {
+      const path = CJSModule._resolveFilename(specifier, module);
+      if (specifier !== path) {
+        specifier = `${pathToFileURL(path)}`;
+      }
+    }
+    const { url: resolvedURL } = cascadedLoader.resolveSync(specifier, url, kEmptyObject);
+    return urlToFilename(resolvedURL);
+  });
+  setOwnProperty(requireFn, 'main', process.mainModule);
+
+  ReflectApply(compiledWrapper, module.exports,
+               [module.exports, requireFn, module, filename, __dirname]);
+  setOwnProperty(module, 'loaded', true);
+}
+
+// TODO: can we use a weak map instead?
+const cjsCache = new SafeMap();
+/**
+ * Creates a ModuleWrap object for a CommonJS module.
+ * @param {string} url - The URL of the module.
+ * @param {string} source - The source code of the module.
+ * @param {boolean} isMain - Whether the module is the main module.
+ * @param {typeof loadCJSModule} [loadCJS=loadCJSModule] - The function to load the CommonJS module.
+ * @returns {ModuleWrap} The ModuleWrap object for the CommonJS module.
+ */
+function createCJSModuleWrap(url, source, isMain, loadCJS = loadCJSModule) {
   debug(`Translating CJSModule ${url}`);
 
-  const filename = fileURLToPath(new URL(url));
+  const filename = urlToFilename(url);
+  // In case the source was not provided by the `load` step, we need fetch it now.
+  source = stringify(source ?? getSource(new URL(url)).source);
 
-  if (!cjsParse) { await initCJSParse(); }
-  const { module, exportNames } = cjsPreparseModuleExports(filename);
+  const { exportNames, module } = cjsPreparseModuleExports(filename, source);
+  cjsCache.set(url, module);
   const namesWithDefault = exportNames.has('default') ?
     [...exportNames] : ['default', ...exportNames];
 
+  if (isMain) {
+    setOwnProperty(process, 'mainModule', module);
+  }
+
   return new ModuleWrap(url, undefined, namesWithDefault, function() {
     debug(`Loading CJSModule ${url}`);
 
-    let exports;
-    if (asyncESM.esmLoader.cjsCache.has(module)) {
-      exports = asyncESM.esmLoader.cjsCache.get(module);
-      asyncESM.esmLoader.cjsCache.delete(module);
-    } else {
-      try {
-        exports = CJSModule._load(filename, undefined, isMain);
-      } catch (err) {
-        enrichCJSError(err, undefined, filename);
-        throw err;
-      }
+    if (!module.loaded) {
+      loadCJS(module, source, url, filename);
     }
 
+    let exports;
+    if (module[kModuleExport] !== undefined) {
+      exports = module[kModuleExport];
+      module[kModuleExport] = undefined;
+    } else {
+      ({ exports } = module);
+    }
     for (const exportName of exportNames) {
       if (!ObjectPrototypeHasOwnProperty(exports, exportName) ||
           exportName === 'default') {
@@ -222,36 +308,79 @@ translators.set('commonjs', async function commonjsStrategy(url, source,
       this.setExport(exportName, value);
     }
     this.setExport('default', exports);
+  }, module);
+}
+
+translators.set('commonjs-sync', function requireCommonJS(url, source, isMain) {
+  initCJSParseSync();
+  assert(!isMain);  // This is only used by imported CJS modules.
+
+  return createCJSModuleWrap(url, source, isMain, (module, source, url, filename) => {
+    assert(module === CJSModule._cache[filename]);
+    CJSModule._load(filename);
   });
 });
 
+// Handle CommonJS modules referenced by `require` calls.
+// This translator function must be sync, as `require` is sync.
+translators.set('require-commonjs', (url, source, isMain) => {
+  assert(cjsParse);
+
+  return createCJSModuleWrap(url, source);
+});
+
+// Handle CommonJS modules referenced by `import` statements or expressions,
+// or as the initial entry point when the ESM loader handles a CommonJS entry.
+translators.set('commonjs', async function commonjsStrategy(url, source,
+                                                            isMain) {
+  if (!cjsParse) {
+    await initCJSParse();
+  }
+
+  // For backward-compatibility, it's possible to return a nullish value for
+  // CJS source associated with a file: URL. In this case, the source is
+  // obtained by calling the monkey-patchable CJS loader.
+  const cjsLoader = source == null ? (module, source, url, filename) => {
+    try {
+      assert(module === CJSModule._cache[filename]);
+      CJSModule._load(filename);
+    } catch (err) {
+      enrichCJSError(err, source, filename);
+      throw err;
+    }
+  } : loadCJSModule;
+
+  try {
+    // We still need to read the FS to detect the exports.
+    source ??= readFileSync(new URL(url), 'utf8');
+  } catch {
+    // Continue regardless of error.
+  }
+  return createCJSModuleWrap(url, source, isMain, cjsLoader);
+
+});
+
 /**
  * Pre-parses a CommonJS module's exports and re-exports.
  * @param {string} filename - The filename of the module.
+ * @param {string} [source] - The source code of the module.
  */
-function cjsPreparseModuleExports(filename) {
+function cjsPreparseModuleExports(filename, source) {
+  // TODO: Do we want to keep hitting the user mutable CJS loader here?
   let module = CJSModule._cache[filename];
-  if (module) {
-    const cached = cjsParseCache.get(module);
-    if (cached) {
-      return { module, exportNames: cached.exportNames };
-    }
+  if (module && module[kModuleExportNames] !== undefined) {
+    return { module, exportNames: module[kModuleExportNames] };
   }
   const loaded = Boolean(module);
   if (!loaded) {
     module = new CJSModule(filename);
     module.filename = filename;
     module.paths = CJSModule._nodeModulePaths(module.path);
+    module[kIsCachedByESMLoader] = true;
+    module[kModuleSource] = source;
     CJSModule._cache[filename] = module;
   }
 
-  let source;
-  try {
-    source = readFileSync(filename, 'utf8');
-  } catch {
-    // Continue regardless of error.
-  }
-
   let exports, reexports;
   try {
     ({ exports, reexports } = cjsParse(source || ''));
@@ -263,39 +392,54 @@ function cjsPreparseModuleExports(filename) {
   const exportNames = new SafeSet(new SafeArrayIterator(exports));
 
   // Set first for cycles.
-  cjsParseCache.set(module, { source, exportNames, loaded });
+  module[kModuleExportNames] = exportNames;
 
   if (reexports.length) {
     module.filename = filename;
     module.paths = CJSModule._nodeModulePaths(module.path);
-  }
-  ArrayPrototypeForEach(reexports, (reexport) => {
-    let resolved;
-    try {
-      resolved = CJSModule._resolveFilename(reexport, module);
-    } catch {
-      return;
-    }
-    const ext = extname(resolved);
-    if ((ext === '.js' || ext === '.cjs' || !CJSModule._extensions[ext]) &&
-        isAbsolute(resolved)) {
-      const { exportNames: reexportNames } = cjsPreparseModuleExports(resolved);
-      for (const name of reexportNames) {
-        exportNames.add(name);
+    for (let i = 0; i < reexports.length; i++) {
+      const reexport = reexports[i];
+      let resolved;
+      try {
+        // TODO: this should be calling the `resolve` hook chain instead.
+        // Doing so would mean dropping support for CJS in the loader thread, as
+        // this call needs to be sync from the perspective of the main thread,
+        // which we can do via HooksProxy and Atomics, but we can't do within
+        // the loaders thread. Until this is done, the lexer will use the
+        // monkey-patchable CJS loader to get the path to the module file to
+        // load (which may or may not be aligned with the URL that the `resolve`
+        // hook have returned).
+        resolved = CJSModule._resolveFilename(reexport, module);
+      } catch {
+        continue;
+      }
+      // TODO: this should be calling the `load` hook chain and check if it returns
+      // `format: 'commonjs'` instead of relying on file extensions.
+      const ext = extname(resolved);
+      if ((ext === '.js' || ext === '.cjs' || !CJSModule._extensions[ext]) &&
+      isAbsolute(resolved)) {
+        // TODO: this should be calling the `load` hook chain to get the source
+        // (and fallback to reading the FS only if the source is nullish).
+        const source = readFileSync(resolved, 'utf-8');
+        const { exportNames: reexportNames } = cjsPreparseModuleExports(resolved, source);
+        for (const name of reexportNames) {
+          exportNames.add(name);
+        }
       }
     }
-  });
+  }
 
   return { module, exportNames };
 }
 
 // Strategy for loading a node builtin CommonJS module that isn't
 // through normal resolution
-translators.set('builtin', async function builtinStrategy(url) {
+translators.set('builtin', function builtinStrategy(url) {
   debug(`Translating BuiltinModule ${url}`);
   // Slice 'node:' scheme
   const id = StringPrototypeSlice(url, 5);
   const module = loadBuiltinModule(id, url);
+  cjsCache.set(url, module);
   if (!StringPrototypeStartsWith(url, 'node:') || !module) {
     throw new ERR_UNKNOWN_BUILTIN_MODULE(url);
   }
@@ -304,7 +448,8 @@ translators.set('builtin', async function builtinStrategy(url) {
 });
 
 // Strategy for loading a JSON file
-translators.set('json', async function jsonStrategy(url, source) {
+const isWindows = process.platform === 'win32';
+translators.set('json', function jsonStrategy(url, source) {
   emitExperimentalWarning('Importing JSON modules');
   assertBufferSource(source, true, 'load');
   debug(`Loading JSONModule ${url}`);
@@ -331,6 +476,7 @@ translators.set('json', async function jsonStrategy(url, source) {
     // A require call could have been called on the same file during loading and
     // that resolves synchronously. To make sure we always return the identical
     // export, we have to check again if the module already exists or not.
+    // TODO: remove CJS loader from here as well.
     module = CJSModule._cache[modulePath];
     if (module && module.loaded) {
       const exports = module.exports;
@@ -356,6 +502,7 @@ translators.set('json', async function jsonStrategy(url, source) {
   if (shouldCheckAndPopulateCJSModuleCache) {
     CJSModule._cache[modulePath] = module;
   }
+  cjsCache.set(url, module);
   return new ModuleWrap(url, undefined, ['default'], function() {
     debug(`Parsing JSONModule ${url}`);
     this.setExport('default', module.exports);
@@ -372,6 +519,8 @@ translators.set('wasm', async function(url, source) {
 
   let compiled;
   try {
+    // TODO(joyeecheung): implement a binding that directly compiles using
+    // v8::WasmModuleObject::Compile() synchronously.
     compiled = await WebAssembly.compile(source);
   } catch (err) {
     err.message = errPath(url) + ': ' + err.message;
