@@ -28,11 +28,10 @@ const {
   StringPrototypeStartsWith,
   Symbol,
   SymbolIterator,
-  Uint32Array,
 } = primordials;
 
 const {
-  errnoException,
+  ErrnoException,
   codes: {
     ERR_ASSERTION,
     ERR_INVALID_ARG_TYPE,
@@ -47,6 +46,8 @@ const {
   validateNumber,
   validateObject,
 } = require('internal/validators');
+const { getValidatedPath } = require('internal/fs/utils');
+const { toNamespacedPath } = require('path');
 const constants = internalBinding('constants').os.signals;
 
 const kInternal = Symbol('internal properties');
@@ -54,24 +55,17 @@ const kInternal = Symbol('internal properties');
 function assert(x, msg) {
   if (!x) throw new ERR_ASSERTION(msg || 'assertion error');
 }
+const { exitCodes: { kNoFailure } } = internalBinding('errors');
 
 const binding = internalBinding('process_methods');
 
-let hrValues;
-let hrBigintValues;
-
-function refreshHrtimeBuffer() {
-  // The 3 entries filled in by the original process.hrtime contains
-  // the upper/lower 32 bits of the second part of the value,
-  // and the remaining nanoseconds of the value.
-  hrValues = new Uint32Array(binding.hrtimeBuffer);
-  // Use a BigUint64Array in the closure because this is actually a bit
-  // faster than simply returning a BigInt from C++ in V8 7.1.
-  hrBigintValues = new BigUint64Array(binding.hrtimeBuffer, 0, 1);
-}
-
-// Create the buffers.
-refreshHrtimeBuffer();
+// The 3 entries filled in by the original process.hrtime contains
+// the upper/lower 32 bits of the second part of the value,
+// and the remaining nanoseconds of the value.
+const hrValues = binding.hrtimeBuffer;
+// Use a BigUint64Array because this is actually a bit
+// faster than simply returning a BigInt from C++ in V8 7.1.
+const hrBigintValues = new BigUint64Array(binding.hrtimeBuffer.buffer, 0, 1);
 
 function hrtime(time) {
   binding.hrtime();
@@ -108,6 +102,7 @@ function wrapProcessMethods(binding) {
     memoryUsage: _memoryUsage,
     rss,
     resourceUsage: _resourceUsage,
+    loadEnvFile: _loadEnvFile,
   } = binding;
 
   function _rawDebug(...args) {
@@ -178,22 +173,21 @@ function wrapProcessMethods(binding) {
   memoryUsage.rss = rss;
 
   function exit(code) {
-    const {
-      handleProcessExit,
-    } = require('internal/modules/esm/handle_process_exit');
+    const { handleProcessExit } = require('internal/modules/run_main');
     process.off('exit', handleProcessExit);
 
-    if (code || code === 0)
+    if (arguments.length !== 0) {
       process.exitCode = code;
+    }
 
     if (!process._exiting) {
       process._exiting = true;
-      process.emit('exit', process.exitCode || 0);
+      process.emit('exit', process.exitCode || kNoFailure);
     }
     // FIXME(joyeecheung): This is an undocumented API that gets monkey-patched
     // in the user land. Either document it, or deprecate it in favor of a
     // better public alternative.
-    process.reallyExit(process.exitCode || 0);
+    process.reallyExit(process.exitCode || kNoFailure);
 
     // If this is a worker, v8::Isolate::TerminateExecution() is called above.
     // That function spoofs the stack pointer to cause the stack guard
@@ -229,7 +223,7 @@ function wrapProcessMethods(binding) {
     }
 
     if (err)
-      throw errnoException(err, 'kill');
+      throw new ErrnoException(err, 'kill');
 
     return true;
   }
@@ -257,6 +251,19 @@ function wrapProcessMethods(binding) {
     };
   }
 
+  /**
+   * Loads the `.env` file to process.env.
+   * @param {string | URL | Buffer | undefined} path
+   */
+  function loadEnvFile(path = undefined) { // Provide optional value so that `loadEnvFile.length` returns 0
+    if (path != null) {
+      path = getValidatedPath(path);
+      _loadEnvFile(toNamespacedPath(path));
+    } else {
+      _loadEnvFile();
+    }
+  }
+
 
   return {
     _rawDebug,
@@ -265,6 +272,7 @@ function wrapProcessMethods(binding) {
     memoryUsage,
     kill,
     exit,
+    loadEnvFile,
   };
 }
 
@@ -279,7 +287,8 @@ function buildAllowedFlags() {
     envSettings: { kAllowedInEnvvar },
     types: { kBoolean },
   } = internalBinding('options');
-  const { options, aliases } = require('internal/options');
+  const { getCLIOptionsInfo } = require('internal/options');
+  const { options, aliases } = getCLIOptionsInfo();
 
   const allowedNodeEnvironmentFlags = [];
   for (const { 0: name, 1: info } of options) {
@@ -424,5 +433,4 @@ module.exports = {
   wrapProcessMethods,
   hrtime,
   hrtimeBigInt,
-  refreshHrtimeBuffer,
 };
