@@ -7,11 +7,10 @@ const {
   ArrayIsArray,
   Promise,
   SymbolAsyncIterator,
-  SymbolDispose,
 } = primordials;
 
 const eos = require('internal/streams/end-of-stream');
-const { once } = require('internal/util');
+const { SymbolDispose, once } = require('internal/util');
 const destroyImpl = require('internal/streams/destroy');
 const Duplex = require('internal/streams/duplex');
 const {
@@ -208,7 +207,7 @@ function pipelineImpl(streams, callback, opts) {
     finishImpl(new AbortError());
   }
 
-  addAbortListener ??= require('events').addAbortListener;
+  addAbortListener ??= require('internal/events/abort_listener').addAbortListener;
   let disposable;
   if (outerSignal) {
     disposable = addAbortListener(outerSignal, abort);
