@@ -1,9 +1,5 @@
 'use strict';
 
-const {
-  emitExperimentalWarning,
-} = require('internal/util');
-
 const {
   ArrayPrototypeAt,
   ArrayPrototypeForEach,
@@ -22,14 +18,17 @@ const {
   Promise,
   Symbol,
   SymbolAsyncIterator,
-  SymbolDispose,
 } = primordials;
 const {
   validateAbortSignal,
-  validateArray,
   validateNumber,
+  validateStringArray,
 } = require('internal/validators');
 
+const {
+  emitExperimentalWarning,
+  SymbolDispose,
+} = require('internal/util');
 const {
   AbortError,
   codes: { ERR_INVALID_STATE, ERR_INVALID_ARG_VALUE },
@@ -301,7 +300,7 @@ class MockTimers {
   }
 
   #clearTimer(timer) {
-    if (timer.priorityQueuePosition !== undefined) {
+    if (timer?.priorityQueuePosition !== undefined) {
       this.#executionQueue.removeAt(timer.priorityQueuePosition);
       timer.priorityQueuePosition = undefined;
     }
@@ -673,7 +672,7 @@ class MockTimers {
    */
   /**
    * Enables the MockTimers replacing the native timers with the fake ones.
-   * @param {EnableOptions} options
+   * @param {EnableOptions} [options]
    */
   enable(options = { __proto__: null, apis: SUPPORTED_APIS, now: 0 }) {
     const internalOptions = { __proto__: null, ...options };
@@ -693,7 +692,7 @@ class MockTimers {
       internalOptions.apis = SUPPORTED_APIS;
     }
 
-    validateArray(internalOptions.apis, 'options.apis');
+    validateStringArray(internalOptions.apis, 'options.apis');
     // Check that the timers passed are supported
     ArrayPrototypeForEach(internalOptions.apis, (timer) => {
       if (!ArrayPrototypeIncludes(SUPPORTED_APIS, timer)) {
