@@ -6,9 +6,9 @@ const {
   ArrayPrototypeForEach,
   ArrayPrototypeIndexOf,
   ArrayPrototypeSome,
-  ObjectCreate,
   ObjectDefineProperty,
   ObjectGetPrototypeOf,
+  ObjectPrototypeHasOwnProperty,
   ObjectSetPrototypeOf,
   ReflectApply,
   SafePromiseAllReturnVoid,
@@ -17,7 +17,6 @@ const {
   TypeError,
 } = primordials;
 
-const { isContext } = internalBinding('contextify');
 const {
   isModuleNamespaceObject,
 } = require('internal/util/types');
@@ -45,6 +44,7 @@ const {
   validateObject,
   validateUint32,
   validateString,
+  validateInternalField,
 } = require('internal/validators');
 
 const binding = internalBinding('module_wrap');
@@ -75,6 +75,15 @@ const kContext = Symbol('kContext');
 const kPerContextModuleId = Symbol('kPerContextModuleId');
 const kLink = Symbol('kLink');
 
+const { isContext } = require('internal/vm');
+
+function isModule(object) {
+  if (typeof object !== 'object' || object === null || !ObjectPrototypeHasOwnProperty(object, kWrap)) {
+    return false;
+  }
+  return true;
+}
+
 class Module {
   constructor(options) {
     emitExperimentalWarning('VM Modules');
@@ -131,6 +140,11 @@ class Module {
           importModuleDynamicallyWrap(options.importModuleDynamically) :
           undefined,
       };
+      // This will take precedence over the referrer as the object being
+      // passed into the callbacks.
+      registry.callbackReferrer = this;
+      const { registerModule } = require('internal/modules/esm/utils');
+      registerModule(this[kWrap], registry);
     } else {
       assert(syntheticEvaluationSteps);
       this[kWrap] = new ModuleWrap(identifier, context,
@@ -138,33 +152,21 @@ class Module {
                                    syntheticEvaluationSteps);
     }
 
-    // This will take precedence over the referrer as the object being
-    // passed into the callbacks.
-    registry.callbackReferrer = this;
-    const { registerModule } = require('internal/modules/esm/utils');
-    registerModule(this[kWrap], registry);
-
     this[kContext] = context;
   }
 
   get identifier() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     return this[kWrap].url;
   }
 
   get context() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     return this[kContext];
   }
 
   get namespace() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     if (this[kWrap].getStatus() < kInstantiated) {
       throw new ERR_VM_MODULE_STATUS('must not be unlinked or linking');
     }
@@ -172,16 +174,12 @@ class Module {
   }
 
   get status() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     return STATUS_MAP[this[kWrap].getStatus()];
   }
 
   get error() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     if (this[kWrap].getStatus() !== kErrored) {
       throw new ERR_VM_MODULE_STATUS('must be errored');
     }
@@ -189,9 +187,7 @@ class Module {
   }
 
   async link(linker) {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     validateFunction(linker, 'linker');
     if (this.status === 'linked') {
       throw new ERR_VM_MODULE_ALREADY_LINKED();
@@ -204,10 +200,7 @@ class Module {
   }
 
   async evaluate(options = kEmptyObject) {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
-
+    validateInternalField(this, kWrap, 'Module');
     validateObject(options, 'options');
 
     let timeout = options.timeout;
@@ -230,14 +223,12 @@ class Module {
   }
 
   [customInspectSymbol](depth, options) {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'Module');
     if (typeof depth === 'number' && depth < 0)
       return this;
 
     const constructor = getConstructorOf(this) || Module;
-    const o = ObjectCreate({ constructor });
+    const o = { __proto__: { constructor } };
     o.status = this.status;
     o.identifier = this.identifier;
     o.context = this.context;
@@ -307,7 +298,7 @@ class SourceTextModule extends Module {
 
       const promises = this[kWrap].link(async (identifier, attributes) => {
         const module = await linker(identifier, this, { attributes, assert: attributes });
-        if (module[kWrap] === undefined) {
+        if (!isModule(module)) {
           throw new ERR_VM_MODULE_NOT_MODULE();
         }
         if (module.context !== this.context) {
@@ -338,19 +329,13 @@ class SourceTextModule extends Module {
   }
 
   get dependencySpecifiers() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
-    if (this[kDependencySpecifiers] === undefined) {
-      this[kDependencySpecifiers] = this[kWrap].getStaticDependencySpecifiers();
-    }
+    validateInternalField(this, kDependencySpecifiers, 'SourceTextModule');
+    this[kDependencySpecifiers] ??= this[kWrap].getStaticDependencySpecifiers();
     return this[kDependencySpecifiers];
   }
 
   get status() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kDependencySpecifiers, 'SourceTextModule');
     if (this.#error !== kNoError) {
       return 'errored';
     }
@@ -361,9 +346,7 @@ class SourceTextModule extends Module {
   }
 
   get error() {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kDependencySpecifiers, 'SourceTextModule');
     if (this.#error !== kNoError) {
       return this.#error;
     }
@@ -416,9 +399,7 @@ class SyntheticModule extends Module {
   }
 
   setExport(name, value) {
-    if (this[kWrap] === undefined) {
-      throw new ERR_VM_MODULE_NOT_MODULE();
-    }
+    validateInternalField(this, kWrap, 'SyntheticModule');
     validateString(name, 'name');
     if (this[kWrap].getStatus() < kInstantiated) {
       throw new ERR_VM_MODULE_STATUS('must be linked');
@@ -433,7 +414,7 @@ function importModuleDynamicallyWrap(importModuleDynamically) {
     if (isModuleNamespaceObject(m)) {
       return m;
     }
-    if (!m || m[kWrap] === undefined) {
+    if (!isModule(m)) {
       throw new ERR_VM_MODULE_NOT_MODULE();
     }
     if (m.status === 'errored') {
