@@ -1164,8 +1164,8 @@ const posix = {
     let resolvedPath = '';
     let resolvedAbsolute = false;
 
-    for (let i = args.length - 1; i >= -1 && !resolvedAbsolute; i--) {
-      const path = i >= 0 ? args[i] : posixCwd();
+    for (let i = args.length - 1; i >= 0 && !resolvedAbsolute; i--) {
+      const path = args[i];
       validateString(path, `paths[${i}]`);
 
       // Skip empty entries
@@ -1178,6 +1178,13 @@ const posix = {
         StringPrototypeCharCodeAt(path, 0) === CHAR_FORWARD_SLASH;
     }
 
+    if (!resolvedAbsolute) {
+      const cwd = posixCwd();
+      resolvedPath = `${cwd}/${resolvedPath}`;
+      resolvedAbsolute =
+        StringPrototypeCharCodeAt(cwd, 0) === CHAR_FORWARD_SLASH;
+    }
+
     // At this point the path should be resolved to a full absolute path, but
     // handle relative paths to be safe (might happen when process.cwd() fails)
 
