@@ -931,7 +931,9 @@ function REPLServer(prompt,
       ReflectApply(_memory, self, [cmd]);
 
       if (e && !self[kBufferedCommandSymbol] &&
-          StringPrototypeStartsWith(StringPrototypeTrim(cmd), 'npm ')) {
+          StringPrototypeStartsWith(StringPrototypeTrim(cmd), 'npm ') &&
+          !(e instanceof Recoverable)
+      ) {
         self.output.write('npm should be run outside of the ' +
                                 'Node.js REPL, in your normal shell.\n' +
                                 '(Press Ctrl+D to exit.)\n');
@@ -1526,7 +1528,7 @@ function complete(line, callback) {
         let p;
         if ((typeof obj === 'object' && obj !== null) ||
             typeof obj === 'function') {
-          memberGroups.push(filteredOwnPropertyNames(obj));
+          ArrayPrototypePush(memberGroups, filteredOwnPropertyNames(obj));
           p = ObjectGetPrototypeOf(obj);
         } else {
           p = obj.constructor ? obj.constructor.prototype : null;
@@ -1534,7 +1536,7 @@ function complete(line, callback) {
         // Circular refs possible? Let's guard against that.
         let sentinel = 5;
         while (p !== null && sentinel-- !== 0) {
-          memberGroups.push(filteredOwnPropertyNames(p));
+          ArrayPrototypePush(memberGroups, filteredOwnPropertyNames(p));
           p = ObjectGetPrototypeOf(p);
         }
       } catch {
