@@ -62,10 +62,10 @@ using SSLPointer = DeleteFnPtr<SSL, SSL_free>;
 using PKCS8Pointer = DeleteFnPtr<PKCS8_PRIV_KEY_INFO, PKCS8_PRIV_KEY_INFO_free>;
 using EVPKeyPointer = DeleteFnPtr<EVP_PKEY, EVP_PKEY_free>;
 using EVPKeyCtxPointer = DeleteFnPtr<EVP_PKEY_CTX, EVP_PKEY_CTX_free>;
-using EVPMDPointer = DeleteFnPtr<EVP_MD_CTX, EVP_MD_CTX_free>;
+using EVPMDCtxPointer = DeleteFnPtr<EVP_MD_CTX, EVP_MD_CTX_free>;
 using RSAPointer = DeleteFnPtr<RSA, RSA_free>;
 using ECPointer = DeleteFnPtr<EC_KEY, EC_KEY_free>;
-using BignumPointer = DeleteFnPtr<BIGNUM, BN_free>;
+using BignumPointer = DeleteFnPtr<BIGNUM, BN_clear_free>;
 using BignumCtxPointer = DeleteFnPtr<BN_CTX, BN_CTX_free>;
 using NetscapeSPKIPointer = DeleteFnPtr<NETSCAPE_SPKI, NETSCAPE_SPKI_free>;
 using ECGroupPointer = DeleteFnPtr<EC_GROUP, EC_GROUP_free>;
@@ -136,7 +136,7 @@ void Decode(const v8::FunctionCallbackInfo<v8::Value>& args,
             void (*callback)(T*, const v8::FunctionCallbackInfo<v8::Value>&,
                              const char*, size_t)) {
   T* ctx;
-  ASSIGN_OR_RETURN_UNWRAP(&ctx, args.Holder());
+  ASSIGN_OR_RETURN_UNWRAP(&ctx, args.This());
 
   if (args[0]->IsString()) {
     StringBytes::InlineDecoder decoder;
@@ -233,6 +233,9 @@ class ByteSource {
     // Returns the (allocated) size in bytes.
     size_t size() const { return size_; }
 
+    // Returns if (allocated) size is zero.
+    bool empty() const { return size_ == 0; }
+
     // Finalizes the Builder and returns a read-only view that is optionally
     // truncated.
     ByteSource release(std::optional<size_t> resize = std::nullopt) && {
@@ -271,6 +274,8 @@ class ByteSource {
 
   size_t size() const { return size_; }
 
+  bool empty() const { return size_ == 0; }
+
   operator bool() const { return data_ != nullptr; }
 
   BignumPointer ToBN() const {
@@ -412,7 +417,7 @@ class CryptoJob : public AsyncWrap, public ThreadPoolWork {
     Environment* env = Environment::GetCurrent(args);
 
     CryptoJob<CryptoJobTraits>* job;
-    ASSIGN_OR_RETURN_UNWRAP(&job, args.Holder());
+    ASSIGN_OR_RETURN_UNWRAP(&job, args.This());
     if (job->mode() == kCryptoJobAsync)
       return job->ScheduleWork();
 
@@ -676,7 +681,8 @@ void array_push_back(const TypeName* evp_ref,
 }
 #endif
 
-inline bool IsAnyByteSource(v8::Local<v8::Value> arg) {
+// WebIDL AllowSharedBufferSource.
+inline bool IsAnyBufferSource(v8::Local<v8::Value> arg) {
   return arg->IsArrayBufferView() ||
          arg->IsArrayBuffer() ||
          arg->IsSharedArrayBuffer();
@@ -694,7 +700,7 @@ class ArrayBufferOrViewContents {
       return;
     }
 
-    CHECK(IsAnyByteSource(buf));
+    CHECK(IsAnyBufferSource(buf));
     if (buf->IsArrayBufferView()) {
       auto view = buf.As<v8::ArrayBufferView>();
       offset_ = view->ByteOffset();
@@ -717,8 +723,7 @@ class ArrayBufferOrViewContents {
     // Ideally, these would return nullptr if IsEmpty() or length_ is zero,
     // but some of the openssl API react badly if given a nullptr even when
     // length is zero, so we have to return something.
-    if (size() == 0)
-      return &buf;
+    if (empty()) return &buf;
     return reinterpret_cast<T*>(data_) + offset_;
   }
 
@@ -726,13 +731,14 @@ class ArrayBufferOrViewContents {
     // Ideally, these would return nullptr if IsEmpty() or length_ is zero,
     // but some of the openssl API react badly if given a nullptr even when
     // length is zero, so we have to return something.
-    if (size() == 0)
-      return &buf;
+    if (empty()) return &buf;
     return reinterpret_cast<T*>(data_) + offset_;
   }
 
   inline size_t size() const { return length_; }
 
+  inline bool empty() const { return length_ == 0; }
+
   // In most cases, input buffer sizes passed in to openssl need to
   // be limited to <= INT_MAX. This utility method helps us check.
   inline bool CheckSizeInt32() { return size() <= INT_MAX; }
@@ -742,14 +748,14 @@ class ArrayBufferOrViewContents {
   }
 
   inline ByteSource ToCopy() const {
-    if (size() == 0) return ByteSource();
+    if (empty()) return ByteSource();
     ByteSource::Builder buf(size());
     memcpy(buf.data<void>(), data(), size());
     return std::move(buf).release();
   }
 
   inline ByteSource ToNullTerminatedCopy() const {
-    if (size() == 0) return ByteSource();
+    if (empty()) return ByteSource();
     ByteSource::Builder buf(size() + 1);
     memcpy(buf.data<void>(), data(), size());
     buf.data<char>()[size()] = 0;
