@@ -1,8 +1,9 @@
-#include "node_util.h"
 #include "base_object-inl.h"
+#include "node_dotenv.h"
 #include "node_errors.h"
 #include "node_external_reference.h"
 #include "util-inl.h"
+#include "v8-fast-api-calls.h"
 
 namespace node {
 namespace util {
@@ -12,17 +13,17 @@ using v8::Array;
 using v8::ArrayBufferView;
 using v8::BigInt;
 using v8::Boolean;
+using v8::CFunction;
 using v8::Context;
 using v8::External;
 using v8::FunctionCallbackInfo;
-using v8::FunctionTemplate;
-using v8::HandleScope;
 using v8::IndexFilter;
 using v8::Integer;
 using v8::Isolate;
 using v8::KeyCollectionMode;
 using v8::Local;
 using v8::Object;
+using v8::ObjectTemplate;
 using v8::ONLY_CONFIGURABLE;
 using v8::ONLY_ENUMERABLE;
 using v8::ONLY_WRITABLE;
@@ -31,13 +32,12 @@ using v8::PropertyFilter;
 using v8::Proxy;
 using v8::SKIP_STRINGS;
 using v8::SKIP_SYMBOLS;
+using v8::StackFrame;
+using v8::StackTrace;
 using v8::String;
 using v8::Uint32;
 using v8::Value;
 
-// Used in ToUSVString().
-constexpr char16_t kUnicodeReplacementCharacter = 0xFFFD;
-
 // If a UTF-16 character is a low/trailing surrogate.
 CHAR_TEST(16, IsUnicodeTrail, (ch & 0xFC00) == 0xDC00)
 
@@ -137,6 +137,30 @@ static void GetProxyDetails(const FunctionCallbackInfo<Value>& args) {
   }
 }
 
+static void GetCallerLocation(const FunctionCallbackInfo<Value>& args) {
+  Isolate* isolate = args.GetIsolate();
+  Local<StackTrace> trace = StackTrace::CurrentStackTrace(isolate, 2);
+
+  // This function is frame zero. The caller is frame one. If there aren't two
+  // stack frames, return undefined.
+  if (trace->GetFrameCount() != 2) {
+    return;
+  }
+
+  Local<StackFrame> frame = trace->GetFrame(isolate, 1);
+  Local<Value> file = frame->GetScriptNameOrSourceURL();
+
+  if (file.IsEmpty()) {
+    return;
+  }
+
+  Local<Value> ret[] = {Integer::New(isolate, frame->GetLineNumber()),
+                        Integer::New(isolate, frame->GetColumn()),
+                        file};
+
+  args.GetReturnValue().Set(Array::New(args.GetIsolate(), ret, arraysize(ret)));
+}
+
 static void IsArrayBufferDetached(const FunctionCallbackInfo<Value>& args) {
   if (args[0]->IsArrayBuffer()) {
     auto buffer = args[0].As<v8::ArrayBuffer>();
@@ -178,114 +202,31 @@ void ArrayBufferViewHasBuffer(const FunctionCallbackInfo<Value>& args) {
   args.GetReturnValue().Set(args[0].As<ArrayBufferView>()->HasBuffer());
 }
 
-WeakReference::WeakReference(Environment* env,
-                             Local<Object> object,
-                             Local<Object> target)
-    : WeakReference(env, object, target, 0) {}
+static uint32_t GetUVHandleTypeCode(const uv_handle_type type) {
+  // TODO(anonrig): We can use an enum here and then create the array in the
+  // binding, which will remove the hard-coding in C++ and JS land.
 
-WeakReference::WeakReference(Environment* env,
-                             Local<Object> object,
-                             Local<Object> target,
-                             uint64_t reference_count)
-    : SnapshotableObject(env, object, type_int),
-      reference_count_(reference_count) {
-  MakeWeak();
-  if (!target.IsEmpty()) {
-    target_.Reset(env->isolate(), target);
-    if (reference_count_ == 0) {
-      target_.SetWeak();
-    }
+  // Currently, the return type of this function corresponds to the index of the
+  // array defined in the JS land. This is done as an optimization to reduce the
+  // string serialization overhead.
+  switch (type) {
+    case UV_TCP:
+      return 0;
+    case UV_TTY:
+      return 1;
+    case UV_UDP:
+      return 2;
+    case UV_FILE:
+      return 3;
+    case UV_NAMED_PIPE:
+      return 4;
+    case UV_UNKNOWN_HANDLE:
+      return 5;
+    default:
+      ABORT();
   }
 }
 
-bool WeakReference::PrepareForSerialization(Local<Context> context,
-                                            v8::SnapshotCreator* creator) {
-  if (target_.IsEmpty()) {
-    target_index_ = 0;
-    return true;
-  }
-
-  // Users can still hold strong references to target in addition to the
-  // reference that we manage here, and they could expect that the referenced
-  // object remains the same as long as that external strong reference
-  // is alive. Since we have no way to know if there is any other reference
-  // keeping the target alive, the best we can do to maintain consistency is to
-  // simply save a reference to the target in the snapshot (effectively making
-  // it strong) during serialization, and restore it during deserialization.
-  // If there's no known counted reference from our side, we'll make the
-  // reference here weak upon deserialization so that it can be GC'ed if users
-  // do not hold additional references to it.
-  Local<Object> target = target_.Get(context->GetIsolate());
-  target_index_ = creator->AddData(context, target);
-  DCHECK_NE(target_index_, 0);
-  target_.Reset();
-  return true;
-}
-
-InternalFieldInfoBase* WeakReference::Serialize(int index) {
-  DCHECK_EQ(index, BaseObject::kEmbedderType);
-  InternalFieldInfo* info =
-      InternalFieldInfoBase::New<InternalFieldInfo>(type());
-  info->target = target_index_;
-  info->reference_count = reference_count_;
-  return info;
-}
-
-void WeakReference::Deserialize(Local<Context> context,
-                                Local<Object> holder,
-                                int index,
-                                InternalFieldInfoBase* info) {
-  DCHECK_EQ(index, BaseObject::kEmbedderType);
-  HandleScope scope(context->GetIsolate());
-
-  InternalFieldInfo* weak_info = reinterpret_cast<InternalFieldInfo*>(info);
-  Local<Object> target;
-  if (weak_info->target != 0) {
-    target = context->GetDataFromSnapshotOnce<Object>(weak_info->target)
-                 .ToLocalChecked();
-  }
-  new WeakReference(Environment::GetCurrent(context),
-                    holder,
-                    target,
-                    weak_info->reference_count);
-}
-
-void WeakReference::New(const FunctionCallbackInfo<Value>& args) {
-  Environment* env = Environment::GetCurrent(args);
-  CHECK(args.IsConstructCall());
-  CHECK(args[0]->IsObject());
-  new WeakReference(env, args.This(), args[0].As<Object>());
-}
-
-void WeakReference::Get(const FunctionCallbackInfo<Value>& args) {
-  WeakReference* weak_ref = Unwrap<WeakReference>(args.Holder());
-  Isolate* isolate = args.GetIsolate();
-  if (!weak_ref->target_.IsEmpty())
-    args.GetReturnValue().Set(weak_ref->target_.Get(isolate));
-}
-
-void WeakReference::GetRef(const FunctionCallbackInfo<Value>& args) {
-  WeakReference* weak_ref = Unwrap<WeakReference>(args.Holder());
-  Isolate* isolate = args.GetIsolate();
-  args.GetReturnValue().Set(
-      v8::Number::New(isolate, weak_ref->reference_count_));
-}
-
-void WeakReference::IncRef(const FunctionCallbackInfo<Value>& args) {
-  WeakReference* weak_ref = Unwrap<WeakReference>(args.Holder());
-  weak_ref->reference_count_++;
-  if (weak_ref->target_.IsEmpty()) return;
-  if (weak_ref->reference_count_ == 1) weak_ref->target_.ClearWeak();
-}
-
-void WeakReference::DecRef(const FunctionCallbackInfo<Value>& args) {
-  WeakReference* weak_ref = Unwrap<WeakReference>(args.Holder());
-  CHECK_GE(weak_ref->reference_count_, 1);
-  weak_ref->reference_count_--;
-  if (weak_ref->target_.IsEmpty()) return;
-  if (weak_ref->reference_count_ == 0) weak_ref->target_.SetWeak();
-}
-
 static void GuessHandleType(const FunctionCallbackInfo<Value>& args) {
   Environment* env = Environment::GetCurrent(args);
   int fd;
@@ -293,71 +234,30 @@ static void GuessHandleType(const FunctionCallbackInfo<Value>& args) {
   CHECK_GE(fd, 0);
 
   uv_handle_type t = uv_guess_handle(fd);
-  const char* type = nullptr;
-
-  switch (t) {
-    case UV_TCP:
-      type = "TCP";
-      break;
-    case UV_TTY:
-      type = "TTY";
-      break;
-    case UV_UDP:
-      type = "UDP";
-      break;
-    case UV_FILE:
-      type = "FILE";
-      break;
-    case UV_NAMED_PIPE:
-      type = "PIPE";
-      break;
-    case UV_UNKNOWN_HANDLE:
-      type = "UNKNOWN";
-      break;
-    default:
-      ABORT();
-  }
-
-  args.GetReturnValue().Set(OneByteString(env->isolate(), type));
+  args.GetReturnValue().Set(GetUVHandleTypeCode(t));
 }
 
-static void ToUSVString(const FunctionCallbackInfo<Value>& args) {
+static uint32_t FastGuessHandleType(Local<Value> receiver, const uint32_t fd) {
+  uv_handle_type t = uv_guess_handle(fd);
+  return GetUVHandleTypeCode(t);
+}
+
+CFunction fast_guess_handle_type_(CFunction::Make(FastGuessHandleType));
+
+static void ParseEnv(const FunctionCallbackInfo<Value>& args) {
   Environment* env = Environment::GetCurrent(args);
-  CHECK_GE(args.Length(), 2);
+  CHECK_EQ(args.Length(), 1);  // content
   CHECK(args[0]->IsString());
-  CHECK(args[1]->IsNumber());
-
-  TwoByteValue value(env->isolate(), args[0]);
-
-  int64_t start = args[1]->IntegerValue(env->context()).FromJust();
-  CHECK_GE(start, 0);
-
-  for (size_t i = start; i < value.length(); i++) {
-    char16_t c = value[i];
-    if (!IsUnicodeSurrogate(c)) {
-      continue;
-    } else if (IsUnicodeSurrogateTrail(c) || i == value.length() - 1) {
-      value[i] = kUnicodeReplacementCharacter;
-    } else {
-      char16_t d = value[i + 1];
-      if (IsUnicodeTrail(d)) {
-        i++;
-      } else {
-        value[i] = kUnicodeReplacementCharacter;
-      }
-    }
-  }
-
-  args.GetReturnValue().Set(
-      String::NewFromTwoByte(env->isolate(),
-                             *value,
-                             v8::NewStringType::kNormal,
-                             value.length()).ToLocalChecked());
+  Utf8Value content(env->isolate(), args[0]);
+  Dotenv dotenv{};
+  dotenv.ParseContent(content.ToStringView());
+  args.GetReturnValue().Set(dotenv.ToObject(env));
 }
 
 void RegisterExternalReferences(ExternalReferenceRegistry* registry) {
   registry->Register(GetPromiseDetails);
   registry->Register(GetProxyDetails);
+  registry->Register(GetCallerLocation);
   registry->Register(IsArrayBufferDetached);
   registry->Register(PreviewEntries);
   registry->Register(GetOwnNonIndexProperties);
@@ -365,13 +265,10 @@ void RegisterExternalReferences(ExternalReferenceRegistry* registry) {
   registry->Register(GetExternalValue);
   registry->Register(Sleep);
   registry->Register(ArrayBufferViewHasBuffer);
-  registry->Register(WeakReference::New);
-  registry->Register(WeakReference::Get);
-  registry->Register(WeakReference::GetRef);
-  registry->Register(WeakReference::IncRef);
-  registry->Register(WeakReference::DecRef);
   registry->Register(GuessHandleType);
-  registry->Register(ToUSVString);
+  registry->Register(FastGuessHandleType);
+  registry->Register(fast_guess_handle_type_.GetTypeInfo());
+  registry->Register(ParseEnv);
 }
 
 void Initialize(Local<Object> target,
@@ -382,7 +279,7 @@ void Initialize(Local<Object> target,
   Isolate* isolate = env->isolate();
 
   {
-    Local<v8::ObjectTemplate> tmpl = v8::ObjectTemplate::New(isolate);
+    Local<ObjectTemplate> tmpl = ObjectTemplate::New(isolate);
 #define V(PropertyName, _)                                                     \
   tmpl->Set(FIXED_ONE_BYTE_STRING(env->isolate(), #PropertyName),              \
             env->PropertyName());
@@ -411,6 +308,18 @@ void Initialize(Local<Object> target,
     V(kRejected);
 #undef V
 
+#define V(name)                                                                \
+  constants                                                                    \
+      ->Set(context,                                                           \
+            FIXED_ONE_BYTE_STRING(isolate, #name),                             \
+            Integer::New(isolate, Environment::ExitInfoField::name))           \
+      .Check();
+
+    V(kExiting);
+    V(kExitCode);
+    V(kHasExitCode);
+#undef V
+
 #define V(name)                                                                \
   constants                                                                    \
       ->Set(context,                                                           \
@@ -432,6 +341,8 @@ void Initialize(Local<Object> target,
   SetMethodNoSideEffect(
       context, target, "getPromiseDetails", GetPromiseDetails);
   SetMethodNoSideEffect(context, target, "getProxyDetails", GetProxyDetails);
+  SetMethodNoSideEffect(
+      context, target, "getCallerLocation", GetCallerLocation);
   SetMethodNoSideEffect(
       context, target, "isArrayBufferDetached", IsArrayBufferDetached);
   SetMethodNoSideEffect(context, target, "previewEntries", PreviewEntries);
@@ -441,6 +352,7 @@ void Initialize(Local<Object> target,
       context, target, "getConstructorName", GetConstructorName);
   SetMethodNoSideEffect(context, target, "getExternalValue", GetExternalValue);
   SetMethod(context, target, "sleep", Sleep);
+  SetMethod(context, target, "parseEnv", ParseEnv);
 
   SetMethod(
       context, target, "arrayBufferViewHasBuffer", ArrayBufferViewHasBuffer);
@@ -453,24 +365,15 @@ void Initialize(Local<Object> target,
                   env->should_abort_on_uncaught_toggle().GetJSArray())
             .FromJust());
 
-  Local<FunctionTemplate> weak_ref =
-      NewFunctionTemplate(isolate, WeakReference::New);
-  weak_ref->InstanceTemplate()->SetInternalFieldCount(
-      WeakReference::kInternalFieldCount);
-  weak_ref->Inherit(BaseObject::GetConstructorTemplate(env));
-  SetProtoMethod(isolate, weak_ref, "get", WeakReference::Get);
-  SetProtoMethod(isolate, weak_ref, "getRef", WeakReference::GetRef);
-  SetProtoMethod(isolate, weak_ref, "incRef", WeakReference::IncRef);
-  SetProtoMethod(isolate, weak_ref, "decRef", WeakReference::DecRef);
-  SetConstructorFunction(context, target, "WeakReference", weak_ref);
-
-  SetMethod(context, target, "guessHandleType", GuessHandleType);
-
-  SetMethodNoSideEffect(context, target, "toUSVString", ToUSVString);
+  SetFastMethodNoSideEffect(context,
+                            target,
+                            "guessHandleType",
+                            GuessHandleType,
+                            &fast_guess_handle_type_);
 }
 
 }  // namespace util
 }  // namespace node
 
-NODE_MODULE_CONTEXT_AWARE_INTERNAL(util, node::util::Initialize)
-NODE_MODULE_EXTERNAL_REFERENCE(util, node::util::RegisterExternalReferences)
+NODE_BINDING_CONTEXT_AWARE_INTERNAL(util, node::util::Initialize)
+NODE_BINDING_EXTERNAL_REFERENCE(util, node::util::RegisterExternalReferences)
