@@ -1,6 +1,9 @@
 'use strict';
 
 const common = require('../common');
+if ((!common.hasCrypto) || (!common.hasIntl)) {
+  common.skip('ESLint tests require crypto and Intl');
+}
 
 common.skipIfEslintMissing();
 
@@ -15,7 +18,7 @@ new RuleTester().run('no-unescaped-regexp-dot', rule, {
     '/.*/',
     '/.?/',
     '/.{5}/',
-    String.raw`/\\\./`
+    String.raw`/\\\./`,
   ],
   invalid: [
     {
@@ -25,6 +28,6 @@ new RuleTester().run('no-unescaped-regexp-dot', rule, {
     {
       code: String.raw`/\\./`,
       errors: [{ message: 'Unescaped dot character in regular expression' }]
-    }
+    },
   ]
 });
