@@ -1,17 +1,19 @@
 // Flags: --expose-internals --expose-gc
 'use strict';
 require('../common');
+const { gcUntil } = require('../common/gc');
 const assert = require('assert');
-const { internalBinding } = require('internal/test/binding');
-const { WeakReference } = internalBinding('util');
+const { WeakReference } = require('internal/util');
 
 let obj = { hello: 'world' };
 const ref = new WeakReference(obj);
 assert.strictEqual(ref.get(), obj);
 
-setImmediate(() => {
+async function main() {
   obj = null;
-  global.gc();
+  await gcUntil(
+    'Reference is garbage collected',
+    () => ref.get() === undefined);
+}
 
-  assert.strictEqual(ref.get(), undefined);
-});
+main();
