@@ -3,21 +3,20 @@
 const common = require('../common');
 const fs = require('fs');
 const { once } = require('events');
-const { join } = require('path');
 const readline = require('readline');
 const assert = require('assert');
 
 const tmpdir = require('../common/tmpdir');
 tmpdir.refresh();
 
-const filename = join(tmpdir.path, 'test.txt');
+const filename = tmpdir.resolve('test.txt');
 
 const testContents = [
   '',
   '\n',
   'line 1',
   'line 1\nline 2 南越国是前203年至前111年存在于岭南地区的一个国家\nline 3\ntrailing',
-  'line 1\nline 2\nline 3 ends with newline\n'
+  'line 1\nline 2\nline 3 ends with newline\n',
 ];
 
 async function testSimpleDestroy() {
@@ -75,6 +74,7 @@ async function testMutualDestroy() {
         break;
       }
       assert.deepStrictEqual(iteratedLines, expectedLines);
+      break;
     }
 
     assert.deepStrictEqual(iteratedLines, expectedLines);
